// Bintree.cpp : Routines for class Bintree.
//

#include "stdafx.h"

/*
 * Debugging switches
 */

#define INCREMENTAL_RECOMPUTE   TRUE
#define CUBIC_MORPH             TRUE

#if _DEBUG
#define PRINT_ADAPTATION    FALSE
#define PRINT_RECOMPUTATION FALSE
#define TEST_ERRORBOUND     FALSE
#define DEBUG_MONOTONICITY  FALSE
#else
#define PRINT_ADAPTATION    FALSE
#define PRINT_RECOMPUTATION FALSE
#define TEST_ERRORBOUND     FALSE
#define DEBUG_MONOTONICITY  FALSE
#endif

#if DEBUG_MONOTONICITY
int print_error_metric = FALSE;
#endif

// Construct a new Bintree from height data.
Bintree::Bintree(int res, unsigned short *height_data, vec3 base, vec3 extent)
{
  int i;

  generating_type = BINTREE_RAWDATA;

  this->res = res;
  hf_size_x = hf_size_y = (1 << res) + 1;

  x_sections = y_sections = 1;

  mesh_base = base;
  mesh_extent = extent;
  mesh_delta.x = mesh_extent.x / hf_size_x;
  mesh_delta.y = mesh_extent.y / hf_size_y;
  mesh_delta.z = mesh_extent.z / HEIGHT_MAX;
  mesh_delta_sq = SQR(mesh_delta.x);

  vertex_array_size = hf_size_x * hf_size_y;
  vertex_array = (unsigned short *) malloc(2 * vertex_array_size * sizeof(unsigned short));
  if (!vertex_array) mem_fail(2 * vertex_array_size * sizeof(unsigned short),
                              "Bintree vertex array");

  for (i = 0; i < vertex_array_size; i++) {
    vertex_array[2 * i + 0] = height_data[i];
    vertex_array[2 * i + 1] = 0;
  }

  base_triangulation_size = 2;

  // The rest of the initialisation is standard.
  standard_init();
}

// Construct a new Bintree from a Myth II mesh tag.
Bintree::Bintree(MeshTag *mesh)
{
  int x, y;
  unsigned short *heightptr;
  gridval *meshptr;

  generating_type = BINTREE_MYTHMESH;
  this->mesh = mesh;
  x_sections = mesh->mesh_header->x_sections;
  y_sections = mesh->mesh_header->y_sections;

  this->res = 5;
  hf_size_x = 32 * x_sections + 1;
  hf_size_y = 32 * y_sections + 1;

  mesh_delta.x = mesh_delta.y = mesh_delta_sq = 1.0;
  mesh_delta.z = 1.0 / 512;

  mesh_base.zero();  // note: it is important that zero in X and Y falls
                     // at the bottom-left-hand corner of a section, for the
                     // texturing to work.
  mesh_extent.x = hf_size_x * mesh_delta.x;
  mesh_extent.y = hf_size_y * mesh_delta.y;
  mesh_extent.z = 256;

  vertex_array_size = hf_size_x * hf_size_y;
  vertex_array = (unsigned short *) malloc(2 * vertex_array_size * sizeof(unsigned short));
  if (!vertex_array) mem_fail(2 * vertex_array_size * sizeof(unsigned short),
                              "Bintree vertex array");

  heightptr = vertex_array;
  meshptr = mesh->grid;

  for (y = 0; y < hf_size_y - 1; y++) {
    for (x = 0; x < hf_size_x - 1; x++) {
      *heightptr++ = meshptr->height - mesh->min_height;
      *heightptr++ = 0;
      meshptr++;
    }
    // Copy the last in X across one.
    *heightptr++ = heightptr[-2];
    *heightptr++ = 0;
  }
  // Copy the last row.
  memcpy((void *) heightptr, (void *) (heightptr - 2 * hf_size_x), 2 * hf_size_x * sizeof(unsigned short));

  base_triangulation_size = 2 * x_sections * y_sections;

  // The rest of the initialisation is standard.
  standard_init();
}

// Standard Bintree initialisation (called by both constructors).
void Bintree::standard_init(void)
{
  int i, tri_this_level, current_level_index;

  max_subdiv_level = 2 * res;
  if (max_subdiv_level > MAX_SUBDIVISION_LEVEL)
    fprintf(debug_file, "ERROR: subdivision level %d exceeds MAX_SUBDIVISION_LEVEL %d... recompile with new constant.\n",
            max_subdiv_level, MAX_SUBDIVISION_LEVEL);

  // Work out the indices into the precalculated-triangle data array.
  precalc_level_index = (int *) malloc((max_subdiv_level + 1) * sizeof(int));
  if (!precalc_level_index) mem_fail((max_subdiv_level + 1) * sizeof(int),
                                     "Bintree precalc-index array");
  current_level_index = 0;
  tri_this_level = base_triangulation_size << max_subdiv_level;
  for (i = 0; i <= max_subdiv_level; i++) {
    precalc_level_index[i] = current_level_index;
    current_level_index += tri_this_level;
    tri_this_level >>= 1;
  }
  assert(tri_this_level == base_triangulation_size / 2);
  precalc_array_size = current_level_index;

  precalc_array = (precalctri *) malloc(precalc_array_size * sizeof(precalctri));
  if (!precalc_array) mem_fail(precalc_array_size * sizeof(precalctri),
                               "Bintree precalculated data");
  precalculate_data();

  // Calculate the frustum-clipping flags for all section corners. This array is [x|y]_sections
  // in either dimension plus a border of 1 on each side.
  frustumflags = (unsigned char *) malloc((x_sections + 2) * (y_sections + 2) * sizeof(unsigned char));
  if (!frustumflags) mem_fail((x_sections + 2) * (y_sections + 2) * sizeof(unsigned char),
                              "Bintree section-visibility flags");

  total_numtri = 0;
  tribuf_size = target_triangle_count * 2;  // assume that user will never try to increase beyond this
  tribuf = (tri *) malloc(tribuf_size * sizeof(tri));
  if (!tribuf)  mem_fail(tribuf_size * sizeof(tri), "Bintree triangle buffer");

  for (i = 0; i < tribuf_size - 1; i++)
    tribuf[i].next_free = i + 1;
  tribuf[tribuf_size - 1].next_free = NONE_FREE;
  next_free_tri = 0;

  total_diamonds = 0;  
  diamondbuf_size = tribuf_size / 4;
  diamondbuf = NULL;
  diamondbuf = (diamond *) malloc(diamondbuf_size * sizeof(diamond));
  if (!diamondbuf)  mem_fail(diamondbuf_size * sizeof(diamond), "Bintree diamond buffer");

  for (i = 0; i < diamondbuf_size - 1; i++)
    diamondbuf[i].next_free = i + 1;
  diamondbuf[diamondbuf_size - 1].next_free = NONE_FREE;
  next_free_diamond = 0;

  total_morphs = morphbuf_size = 0;
  next_free_morph = NONE_FREE;
  morphbuf = NULL;
  first_morph = last_morph = -1;
  morphed_time = 0;

  tri_queue     = new PriQueue(+1);
  diamond_queue = new PriQueue(-1);

  current_distance_moved = 0;
  threshold_priorities_set = FALSE;

  tri_recompute_queue     = new PriQueue(-1);
  diamond_recompute_queue = new PriQueue(-1);

  error_metric = perpendicular_epsilon_metric;

  is_textured = FALSE;
  use_borders = FALSE;    // Not usually supported. :(

  scenerytypearray = NULL;
  sceneryarray = NULL;
}

// Delete the Bintree.
Bintree::~Bintree(void)
{
  int i;

  if (is_textured) {
    glDeleteTextures(num_textures, texture_handle);
    free(texture_handle);

    for (i = 0; i < num_textures * miplevels; i++)
      free(texturedata[i]);
    free(texturedata);

    for (i = 0; i < num_scenery_types; i++)
      glDeleteTextures(1, &scenerytypearray[i].texture);
  }

  if (scenerytypearray != NULL)
    free(scenerytypearray);

  if (sceneryarray != NULL)
    free(sceneryarray);

  free(precalc_level_index);
  free(precalc_array);

  free(frustumflags);

  free(vertex_array);

  delete tri_queue;
  delete diamond_queue;

  delete tri_recompute_queue;
  delete diamond_recompute_queue;

  if (tribuf != NULL)
    free(tribuf);

  if (diamondbuf != NULL)
    free(diamondbuf);

  if (morphbuf != NULL)
    free(morphbuf);
}

double num_lights;
vec3 light_dir[2], light_color[2];

// Perform a bottom-up precalculation on the bintree's height data.
void Bintree::precalculate_data(void)
{
  int x, y, vertex_offset;
  interpstruct interp_data[MAX_SUBDIVISION_LEVEL + 1];
  unsigned short dummy;

  // Set up the lights.
  num_lights = 2;
  light_dir[0].x = 1.0;
  light_dir[0].y = 1.0;
  light_dir[0].z = 0.4;
  light_color[0].x = 1.0;
  light_color[0].y = 0.95;
  light_color[0].z = 0.8;
  normalize(light_dir[0]);

  light_dir[1].x = -1.0;
  light_dir[1].y = -1.0;
  light_dir[1].z = 0.0;
  light_color[1].x = 0.0;
  light_color[1].y = 0.0;
  light_color[1].z = 0.2;
  normalize(light_dir[1]);

  if (generating_type == BINTREE_RAWDATA) {
    // The mesh is made up of two top-level triangles.
    recursive_precalculate(max_subdiv_level, 0, interp_data,
                           (hf_size_y - 1) * hf_size_x, 0, (hf_size_x - 1) + (hf_size_y - 1) * hf_size_x,
                           &dummy, &dummy);
    recursive_precalculate(max_subdiv_level, 1, interp_data,
                           (hf_size_x - 1), (hf_size_x - 1) + (hf_size_y - 1) * hf_size_x, 0,
                           &dummy, &dummy);
  } else if (generating_type == BINTREE_MYTHMESH) {
    // The mesh is made up of two top-level triangles for every section.
    for (y = 0; y < y_sections; y++) {
      for (x = 0; x < x_sections; x++) {
        vertex_offset = (32 * x) + (32 * y) * hf_size_x;

        if (((x + y) % 2) == 0) {
          // Same orientation as the 0,0 section.
          recursive_precalculate(max_subdiv_level, 2 * (x + y * x_sections) + 0, interp_data,
                                 vertex_offset + 32 * hf_size_x, vertex_offset + 0, vertex_offset + 32 + 32 * hf_size_x,
                                 &dummy, &dummy);

          recursive_precalculate(max_subdiv_level, 2 * (x + y * x_sections) + 1, interp_data,
                                 vertex_offset + 32, vertex_offset + 32 + 32 * hf_size_x, vertex_offset + 0,
                                 &dummy, &dummy);
        } else {
          // Flipped orientation to the 0,0 section.
          recursive_precalculate(max_subdiv_level, 2 * (x + y * x_sections) + 0, interp_data,
                                 vertex_offset + 0, vertex_offset + 32, vertex_offset + 32 * hf_size_x,
                                 &dummy, &dummy);

          recursive_precalculate(max_subdiv_level, 2 * (x + y * x_sections) + 1, interp_data,
                                 vertex_offset + 32 + 32 * hf_size_x, vertex_offset + 32 * hf_size_x, vertex_offset + 32,
                                 &dummy, &dummy);
        }
      }
    }
  } else {
    assert(0);
  }
}

// Set target values.
void Bintree::set_target_priority(double target_pri)
{
  target_priority = target_pri;

  if (!threshold_priorities_set) {
    biggest_split_priority = smallest_diamond_priority = target_priority;
    threshold_priorities_set = TRUE;
  }
}

// Lock the incremental thresholds to the specified priority...
// used during initial adaptation of a base triangulation.
void Bintree::lock_incremental_thresholds(double lock_pri)
{
  lock_thresholds = TRUE;
  threshold_lockval = lock_pri;
}

// Unlock the incremental thresholds.
void Bintree::unlock_incremental_thresholds(void)
{
  lock_thresholds = FALSE;
}

void Bintree::set_target_triangle_count(int new_limit)
{
  tri_limit = new_limit;
}

// Get the worldspace coordinates of a vertex.
vec3 Bintree::vertex_world(int v)
{
  vec3 w;

  w.x = mesh_base.x + mesh_delta.x * (v % hf_size_x);
  w.y = mesh_base.y + mesh_delta.y * (v / hf_size_x);
  w.z = mesh_base.z + mesh_delta.z * get_height(v);

  return w;
}

// Recursively precalculate data about the bintree's possible triangles.
int Bintree::recursive_precalculate(int tri_level, int level_index, interpstruct *interp_data, int vc, int v0, int v1,
                                    unsigned short *minval, unsigned short *maxval)
{
  int tri_index, vb;
  int zb, i, error_bound, child0_thickness, child1_thickness;
  unsigned short child0_min, child0_max, child1_min, child1_max;
  double l;
  vec3 pc, p0, p1, edge0, edge1, n;

  tri_index = precalc_index(tri_level, level_index);

  // Calculate tri_color for this triangle.  
  pc = vertex_world(vc); p0 = vertex_world(v0); p1 = vertex_world(v1);
  edge0 = p0 - pc; edge1 = p1 - pc;
  n = edge0 ^ edge1;
  normalize(n);

  precalc_array[tri_index].tri_color.zero();
  for (i = 0; i < num_lights; i++) {
    l = n * light_dir[i];
    if (l > 0)
      precalc_array[tri_index].tri_color += l * light_color[i];
  }


  // Work out the error thickness for this triangle.
  if (tri_level == 0) {
    // This is a bottom-level triangle - no split vertex to cause approximation error!
    precalc_array[tri_index].minval = MIN(vertex_array[2*vc], MIN(vertex_array[2*v0], vertex_array[2*v1]));
    precalc_array[tri_index].maxval = MAX(vertex_array[2*vc], MAX(vertex_array[2*v0], vertex_array[2*v1]));
    precalc_array[tri_index].error_bound = 0;

    *minval = precalc_array[tri_index].minval;
    *maxval = precalc_array[tri_index].maxval;
    return 0;
  }

  interpstruct child0_interp[MAX_SUBDIVISION_LEVEL + 1], child1_interp[MAX_SUBDIVISION_LEVEL + 1];
  double interp_zb[MAX_SUBDIVISION_LEVEL + 1], this_bound, child0_bound, child1_bound;
  
  // At this level, the triangle approximation has the exact vertex Z values for our triangle.
  interp_data[tri_level].i_zc = vertex_array[2*vc];
  interp_data[tri_level].i_z0 = vertex_array[2*v0];
  interp_data[tri_level].i_z1 = vertex_array[2*v1];

  // What vertex are we splitting at?
  vb = (v0 + v1) / 2;
  zb = vertex_array[2*vb];
  
  // For every level currently being considered (our level and all those above), work out where
  // our sub-triangles' vertices are interpolated to in the approximation plane.
  for (i = tri_level; i <= max_subdiv_level; i++) {
    interp_zb[i] = (interp_data[i].i_z0 + interp_data[i].i_z1) / 2;
    
    child0_interp[i].i_zc = interp_zb[i];
    child0_interp[i].i_z0 = interp_data[i].i_zc;
    child0_interp[i].i_z1 = interp_data[i].i_z0;
    
    child1_interp[i].i_zc = interp_zb[i];
    child1_interp[i].i_z0 = interp_data[i].i_z1;
    child1_interp[i].i_z1 = interp_data[i].i_zc;
  }
  
  // Recursively work out the maximum distance from each approximation plane for all
  // split vertices in our child triangles. NB: if tri_level == 1 then these child
  // triangles DO NOT store anything in child0_interp[k].error_bound - they will all
  // just be zero anyway.
  child0_thickness = recursive_precalculate(tri_level - 1, 2*level_index,     child0_interp,
                                            vb, vc, v0, &child0_min, &child0_max);
  child1_thickness = recursive_precalculate(tri_level - 1, 2*level_index + 1, child1_interp,
                                            vb, v1, vc, &child1_min, &child1_max);

  if (tri_level > 1) {
    // For every level currently being considered, work out the maximum distance from
    // each approximation plane for all split vertices in our triangle.
    for (i = tri_level; i <= max_subdiv_level; i++) {
      this_bound = fabs(zb - interp_zb[i]);
      child0_bound = child0_interp[i].error_bound;
      child1_bound = child0_interp[i].error_bound;
      
      interp_data[i].error_bound = MAX(this_bound, MAX(child0_bound, child1_bound));
    }
  } else {
    // Our child triangles have no split vertices - they are bottom-level triangles.
    // Any error comes purely from our approximations.
    for (i = tri_level; i <= max_subdiv_level; i++)
      interp_data[i].error_bound = fabs(zb - interp_zb[i]);
  }

  error_bound = (int) interp_data[tri_level].error_bound;

  // Make absolutely sure that we aren't violating the monotonicity bound (this could be possible
  // if the children are worse approximations than we are).
  error_bound = MAX(error_bound, MAX(child0_thickness, child1_thickness));

  precalc_array[tri_index].error_bound = error_bound;

  // Compute our min and max bounds.
  precalc_array[tri_index].minval = MIN(child0_min, child1_min);
  precalc_array[tri_index].maxval = MAX(child0_max, child1_max);

  *minval = precalc_array[tri_index].minval;
  *maxval = precalc_array[tri_index].maxval;
  return error_bound;
}

// Revert the Bintree to its base triangulation.
void Bintree::base_triangulation(void)
{
  int t, neighbour_t, x, y, vertex_offset, section_id;

  total_numtri = 0;
  tri_queue->clear();
  diamond_queue->clear();

  tri_recompute_queue->clear();
  diamond_recompute_queue->clear();

  if (generating_type == BINTREE_RAWDATA) {
    // The mesh is made up of two top-level triangles.
    t = new_triangle();
    tribuf[t].n0 = tribuf[t].n1 = -1;
    tribuf[t].nb = 1;
    tribuf[t].vc = (hf_size_y - 1) * hf_size_x;
    tribuf[t].v0 = 0;
    tribuf[t].v1 = (hf_size_x - 1) + (hf_size_y - 1) * hf_size_x;
    tribuf[t].sub_level = max_subdiv_level;
    tribuf[t].level_index = 0;
    tribuf[t].diamond_index = -1;
    setup_triangle_priority(t);
    if (is_textured) {
      first_tex_tri[0] = t;
      tribuf[t].texture_index = 0;
      tribuf[t].prev_tex = -1;
      tribuf[t].next_tex = t + 1;
    }

    t = new_triangle();
    tribuf[t].n0 = tribuf[t].n1 = -1;
    tribuf[t].nb = 0;
    tribuf[t].vc = hf_size_x - 1;
    tribuf[t].v0 = (hf_size_x - 1) + (hf_size_y - 1) * hf_size_x;
    tribuf[t].v1 = 0;
    tribuf[t].sub_level = max_subdiv_level;
    tribuf[t].level_index = 1;
    tribuf[t].diamond_index = -1;
    setup_triangle_priority(t);
    if (is_textured) {
      last_tex_tri[0] = t;
      tribuf[t].texture_index = 0;
      tribuf[t].prev_tex = t - 1;
      tribuf[t].next_tex = -1;
    }

  } else if (generating_type == BINTREE_MYTHMESH) {
    // The mesh is made up of two top-level triangles for every section.
    for (y = 0; y < y_sections; y++) {
      for (x = 0; x < x_sections; x++) {
        vertex_offset = (32 * x) + (32 * y) * hf_size_x;
        section_id = x + y * x_sections;

        if (((x + y) % 2) == 0) {
          // Same orientation as the 0,0 section.
          t = new_triangle();
          tribuf[t].n0 = (x == 0) ? -1 : (2 * (section_id - 1) + 1);
          tribuf[t].n1 = (y == y_sections - 1) ? -1 : (2 * (section_id + x_sections) + 0);
          tribuf[t].nb = 2 * section_id + 1;
          
          tribuf[t].vc = vertex_offset + 32 * hf_size_x;
          tribuf[t].v0 = vertex_offset + 0;
          tribuf[t].v1 = vertex_offset + 32 + 32 * hf_size_x;
          
          tribuf[t].sub_level = max_subdiv_level;
          tribuf[t].level_index = 2 * section_id + 0;
          tribuf[t].diamond_index = -1;
          setup_triangle_priority(t);
          if (is_textured) {
            first_tex_tri[section_id] = t;
            tribuf[t].texture_index = section_id;
            tribuf[t].prev_tex = -1;
            tribuf[t].next_tex = t + 1;
          }
          
          t = new_triangle();
          tribuf[t].n0 = (x == x_sections - 1) ? -1 : (2 * (section_id + 1) + 0);
          tribuf[t].n1 = (y == 0) ? -1 : (2 * (section_id - x_sections) + 1);
          tribuf[t].nb = 2 * section_id + 0;
          
          tribuf[t].vc = vertex_offset + 32;
          tribuf[t].v0 = vertex_offset + 32 + 32 * hf_size_x;
          tribuf[t].v1 = vertex_offset + 0;
          
          tribuf[t].sub_level = max_subdiv_level;
          tribuf[t].level_index = 2 * section_id + 1;
          tribuf[t].diamond_index = -1;
          setup_triangle_priority(t);
          if (is_textured) {
            last_tex_tri[section_id] = t;
            tribuf[t].texture_index = section_id;
            tribuf[t].prev_tex = t - 1;
            tribuf[t].next_tex = -1;
          }
        } else {
          // Flipped orientation to the 0,0 section.
          t = new_triangle();
          tribuf[t].n0 = (y == 0) ? -1 : (2 * (section_id - x_sections) + 0);
          tribuf[t].n1 = (x == 0) ? -1 : (2 * (section_id - 1) + 1);
          tribuf[t].nb = 2 * section_id + 1;
          
          tribuf[t].vc = vertex_offset + 0;
          tribuf[t].v0 = vertex_offset + 32;
          tribuf[t].v1 = vertex_offset + 32 * hf_size_x;
          
          tribuf[t].sub_level = max_subdiv_level;
          tribuf[t].level_index = 2 * section_id + 0;
          tribuf[t].diamond_index = -1;
          setup_triangle_priority(t);
          if (is_textured) {
            first_tex_tri[section_id] = t;
            tribuf[t].texture_index = section_id;
            tribuf[t].prev_tex = -1;
            tribuf[t].next_tex = t + 1;
          }
          
          t = new_triangle();
          tribuf[t].n0 = (y == y_sections - 1) ? -1 : (2 * (section_id + x_sections) + 1);
          tribuf[t].n1 = (x == x_sections - 1) ? -1 : (2 * (section_id + 1) + 0);
          tribuf[t].nb = 2 * section_id + 0;
          
          tribuf[t].vc = vertex_offset + 32 + 32 * hf_size_x;
          tribuf[t].v0 = vertex_offset + 32 * hf_size_x;
          tribuf[t].v1 = vertex_offset + 32;
          
          tribuf[t].sub_level = max_subdiv_level;
          tribuf[t].level_index = 2 * section_id + 1;
          tribuf[t].diamond_index = -1;
          setup_triangle_priority(t);
          if (is_textured) {
            last_tex_tri[section_id] = t;
            tribuf[t].texture_index = section_id;
            tribuf[t].prev_tex = t - 1;
            tribuf[t].next_tex = -1;
          }
        }
      }
    }
  } else {
    assert(0);
  }


  // Check all the adjacencies are right!
  for (t = 0; t < total_numtri; t++) {
    neighbour_t = tribuf[t].n0;
    if (neighbour_t != -1) {
      assert(tribuf[neighbour_t].n1 == t);
      assert(tribuf[t].vc == tribuf[neighbour_t].vc);
      assert(tribuf[t].v0 == tribuf[neighbour_t].v1);
    }

    neighbour_t = tribuf[t].n1;
    if (neighbour_t != -1) {
      assert(tribuf[neighbour_t].n0 == t);
      assert(tribuf[t].vc == tribuf[neighbour_t].vc);
      assert(tribuf[t].v1 == tribuf[neighbour_t].v0);
    }

    neighbour_t = tribuf[t].nb;
    assert(tribuf[neighbour_t].nb == t);
    assert(tribuf[t].v0 == tribuf[neighbour_t].v1);
    assert(tribuf[t].v1 == tribuf[neighbour_t].v0);
  }
}


// Get a new triangle index in the triangle buffer.
int Bintree::new_triangle(void)
{
  int i;

  if (next_free_tri == NONE_FREE) {
    fprintf(debug_file, "ERROR: exceeded hard-coded triangle limit %d.\n", tribuf_size);
    fprintf(debug_file, "\n");
    fprintf(debug_file, "Out of memory.\n");
    fprintf(debug_file, "We wish to hold the whole sky,\n");
    fprintf(debug_file, "But we never will.\n");
    premature_shutdown();
  }

  i = next_free_tri;
  next_free_tri = tribuf[i].next_free;
  tribuf[i].next_free = IN_USE;

  // Set up the indices in the triangle's priq_info.
  tribuf[i].split_queue_info.buf_index = i;
  tribuf[i].recompute_queue_info.buf_index = i;

  total_numtri++;

  return i;
}

// Free a triangle in the triangle buffer.
void Bintree::free_triangle(int tri_index)
{
  assert((tri_index >= 0) && (tri_index < tribuf_size));
  assert(tribuf[tri_index].next_free == IN_USE);
  assert(total_numtri > 0);

  tribuf[tri_index].next_free = next_free_tri;
  next_free_tri = tri_index;

  total_numtri--;
}

// Get a new diamond index in the diamond buffer.
int Bintree::new_diamond(void)
{
  int i;

  if (next_free_diamond == NONE_FREE) {
    fprintf(debug_file, "ERROR: exceeded hard-coded diamond limit %d.\n", diamondbuf_size);
    fprintf(debug_file, "\n");
    fprintf(debug_file, "Out of memory.\n");
    fprintf(debug_file, "We wish to hold the whole sky,\n");
    fprintf(debug_file, "But we never will.\n");
    premature_shutdown();
  }

  i = next_free_diamond;
  next_free_diamond = diamondbuf[i].next_free;
  diamondbuf[i].next_free = IN_USE;

  // Set up the indices in the triangle's priq_info.
  diamondbuf[i].merge_queue_info.buf_index = i;
  diamondbuf[i].recompute_queue_info.buf_index = i;

  total_diamonds++;

  return i;
}

// Free a diamond in the diamond buffer.
void Bintree::free_diamond(int diamond_index)
{
  assert((diamond_index >= 0) && (diamond_index < diamondbuf_size));
  assert(diamondbuf[diamond_index].next_free == IN_USE);
  assert(total_diamonds > 0);

  diamondbuf[diamond_index].next_free = next_free_diamond;
  next_free_diamond = diamond_index;

  total_diamonds--;
}

// Get a new morph index in the morph buffer.
int Bintree::new_morph(void)
{
  int i;

  if (next_free_morph == NONE_FREE) {
    // We must expand the morph buffer.
    next_free_morph = morphbuf_size;

    // We never use morphbuffer index zero.
    if (next_free_morph == 0)
      next_free_morph++;

    morphbuf = (morphvertex *) realloc((void *) morphbuf,
                   (morphbuf_size + morphbuf_growsize) * sizeof(morphvertex));
    if (!morphbuf) mem_fail((morphbuf_size + morphbuf_growsize) * sizeof(morphvertex),
                            "Bintree morphing-vertex-buffer growth");

    for (i = 1; i < morphbuf_growsize - 1; i++, morphbuf_size++)
      morphbuf[morphbuf_size].next_free = morphbuf_size + 1;
    morphbuf[morphbuf_size++].next_free = NONE_FREE;
  }

  i = next_free_morph;
  next_free_morph = morphbuf[i].next_free;
  morphbuf[i].next_free = IN_USE;

  total_morphs++;

  return i;
}

// Free a morph in the morph buffer.
void Bintree::free_morph(int morph_index)
{
  int link_index;

  assert((morph_index > 0) && (morph_index < morphbuf_size));
  assert(morphbuf[morph_index].next_free == IN_USE);
  assert(total_morphs > 0);

  // Maintain the integrity of the previous-link in the morph-list.
  link_index = morphbuf[morph_index].prev_morph;
  if (link_index == -1) {
    first_morph = morphbuf[morph_index].next_morph;
  } else {
    assert((link_index > 0) && (link_index < morphbuf_size));
    assert(morphbuf[link_index].next_free == IN_USE);
    morphbuf[link_index].next_morph = morphbuf[morph_index].next_morph;
  }

  // Maintain the integrity of the next-link in the morph-list.
  link_index = morphbuf[morph_index].next_morph;
  if (link_index == -1) {
    last_morph = morphbuf[morph_index].prev_morph;
  } else {
    assert((link_index > 0) && (link_index < morphbuf_size));
    assert(morphbuf[link_index].next_free == IN_USE);
    morphbuf[link_index].prev_morph = morphbuf[morph_index].prev_morph;
  }

  // Update the freelist.
  morphbuf[morph_index].next_free = next_free_morph;
  next_free_morph = morph_index;

  total_morphs--;
}

// Split a triangle.
void Bintree::split_triangle(int t_a)
{
  int t_b, vb, new_d, m_index;
  int newt_a, newt_b;
  unsigned short interp_zb;
  morphvertex *mptr;

  assert(tribuf[t_a].sub_level > 0);
  t_b = tribuf[t_a].nb;

  // Create a new triangle that is t_a's 'high' child.
  newt_a = new_triangle();

  if (t_b == -1) {
    // This triangle is adjacent to the edge of the mesh. We don't need to do anything
    // about its neighbour.
    newt_b = -1;
  } else {
    // If our neighbour is at a coarser grain than us, then the edge that it is adjacent to
    // us along will not be its base edge, so we will need to split it.
    if (tribuf[t_b].sub_level > tribuf[t_a].sub_level) {
      // It may be at most one level coarser than us.
      assert(tribuf[t_b].sub_level == tribuf[t_a].sub_level + 1);

      // Split this triangle (may cause a recursive split through its base neighbours)
      split_triangle(t_b);

      // t_b may now point to the wrong child of the old t_b. The correct triangle index
      // will have been placed in t_a, though.
      t_b = tribuf[t_a].nb;
    }

    // The triangle should now be at the same level as us. Check that the vertex info
    // matches up, also.
    assert(tribuf[t_b].sub_level == tribuf[t_a].sub_level);
    assert(tribuf[t_b].nb == t_a);
    assert(tribuf[t_b].v0 == tribuf[t_a].v1);
    assert(tribuf[t_b].v1 == tribuf[t_a].v0);

    // Create a new triangle that is t_b's 'high' child.
    newt_b = new_triangle();
  }

  // We are splitting, and so is our base triangle; if they were part of any mergeable diamonds,
  // break them.
  break_diamond(tribuf[t_a].diamond_index);
  if (t_b != -1)
    break_diamond(tribuf[t_b].diamond_index);

  // Get a new diamond to store this pair of triangles that, after this split,
  // will be in a state where they could be re-merged.
  new_d = new_diamond();

  diamondbuf[new_d].t0 = t_a;
  diamondbuf[new_d].t1 = newt_a;
  diamondbuf[new_d].t2 = t_b;
  diamondbuf[new_d].t3 = newt_b;

  // Find the index of the new base vertex.
  vb = (tribuf[t_a].v0 + tribuf[t_a].v1) / 2;
  if (morph_vertices) {
    // Vertex vb is now going to be morphing from its current position
    // to its new position.
    interp_zb = (get_height(tribuf[t_a].v0) + get_height(tribuf[t_a].v1)) / 2;

    // Get a new morph entry.
    m_index = new_morph();
    mptr = morphbuf + m_index;

//    fprintf(debug_file, "splitting at vertex %d, storing in morph vertex %d\n",
//            vb, m_index);
//    fflush(debug_file);

    // Set up the morph's parameters.
    mptr->start_height = mptr->current_height = interp_zb;
    mptr->delta_height = vertex_array[2 * vb + 0] - interp_zb;
    mptr->finish_time = morphed_time + morph_msec;
    mptr->vertex_index = vb;

    // Insert this morph at the end of the currently-morphing chain.
    mptr->prev_morph = last_morph;
    if (last_morph == -1) {
      first_morph = m_index;
    } else {
      assert((last_morph > 0) && (last_morph < morphbuf_size));
      assert(morphbuf[last_morph].next_free == IN_USE);
      morphbuf[last_morph].next_morph = m_index;
    }

    mptr->next_morph = -1;
    last_morph = m_index;

    // Store that this vertex is morphing.
    assert(vertex_array[2 * vb + 1] == 0);
    vertex_array[2 * vb + 1] = m_index;
  }

  perform_split(vb, new_d, t_a, newt_a, t_b, newt_b);
  if (t_b != -1)
    perform_split(vb, new_d, t_b, newt_b, t_a, newt_a);

  assert(tribuf[diamondbuf[new_d].t0].diamond_index == new_d);
  assert(tribuf[diamondbuf[new_d].t1].diamond_index == new_d);
  if (t_b != -1) {
    assert(tribuf[diamondbuf[new_d].t2].diamond_index == new_d);
    assert(tribuf[diamondbuf[new_d].t3].diamond_index == new_d);
  }

  // Set up the priority for the new diamond.
  // NB: must be done AFTER perform_split, where new child triangle vertices
  // are set up.
  setup_diamond_priority(new_d);
}

// Internal routine for split_triangle; performs a single-triangle split and updates
// neighbour information.
void Bintree::perform_split(int vb, int d_index, int t, int t_child, int base_t, int base_t_child)
{
  int neighbour_t, tex_index, link_t;
  double old_priority;

  tribuf[t_child].level_index = 2*tribuf[t].level_index + 1;
  tribuf[t_child].sub_level = tribuf[t].sub_level - 1;
  tribuf[t_child].diamond_index = d_index;
  tribuf[t_child].vc = vb;
  tribuf[t_child].v0 = tribuf[t].v1;
  tribuf[t_child].v1 = tribuf[t].vc;
  tribuf[t_child].recompute_queue_info.priq_location = -1;

  if (is_textured) {
    // We are part of our parent's texture-chain.
    tribuf[t_child].texture_index = tex_index = tribuf[t].texture_index;

    // We are the high child - our previous link is to our parent's low
    // child (which happens to be stored where our parent was).
    tribuf[t_child].prev_tex = t;

    // Our next link is the same as our parent's was.
    tribuf[t_child].next_tex = link_t = tribuf[t].next_tex;
    if (link_t == -1) {
      // Our parent was the last tri in the texture-chain. Now we are.
      last_tex_tri[tex_index] = t_child;
    } else {
      // Link up the next tri to us.
      assert((link_t >= 0) && (link_t < tribuf_size) && (tribuf[link_t].next_free == IN_USE));
      assert(tribuf[link_t].prev_tex == t);
      tribuf[link_t].prev_tex = t_child;
    }
  }

  if (tribuf[t_child].sub_level > 0) {
    // Calculate this triangle's split priority, and insert it into the splittable-triangle queue.
    calculate_tri_priority(t_child);

#if DEBUG_MONOTONICITY
    if (tribuf[t].split_queue_info.priority - tribuf[t_child].split_queue_info.priority 
           < -PRIORITY_TOLERANCE) {
      // The parent's priority was LESS than this child's new priority. This should
      // never happen.
      fprintf(debug_file, "MONOTONICITY VIOLATION on split(%d) child1: child metric %f > parent metric %f\n",
              t, tribuf[t_child].split_queue_info.priority, tribuf[t].split_queue_info.priority);
      print_error_metric = TRUE;

      fprintf(debug_file, "child metric...\n");
      calculate_tri_priority(t_child, TRUE);

      fprintf(debug_file, "parent metric...\n");
      calculate_tri_priority(t, TRUE);

      if (tribuf[t].split_queue_info.priority - tribuf[t_child].split_queue_info.priority
             < -PRIORITY_TOLERANCE) {
        fprintf(debug_file, "VIOLATION CONFIRMED - shutting down\n");
        premature_shutdown();
      } else {
        fprintf(debug_file, "Ah... false alarm?\n");
      }
    }
#endif

    tri_queue->insert_element(&tribuf[t_child].split_queue_info);
  } else {
    // This triangle is not splittable.
    tribuf[t_child].split_queue_info.priority = PRIORITY_NEVER;
    tribuf[t_child].split_queue_info.priq_location = -1;
    tribuf[t_child].recompute_queue_info.priority = DIST_INFINITY;
  }

  // Link up this triangle's neighbours.
  tribuf[t_child].n0 = base_t;
  tribuf[t_child].n1 = t;
  tribuf[t_child].nb = neighbour_t = tribuf[t].n1;
  if (neighbour_t != -1) {
    if (tribuf[neighbour_t].sub_level == tribuf[t].sub_level) {
      // This neighbour triangle is of the same level as t was, so one
      // of its short edges must have been the one abutting t along its vc-v1
      // edge. Looking at the geometry and relying on the fact that two touching
      // triangles of the same level always form a larger whole triangle (and not
      // a parallelogram), the neighbour's vc-v0 edge must have been the one touching.
      assert(tribuf[neighbour_t].n0 == t);
      tribuf[neighbour_t].n0 = t_child;
    } else {
      // This neighbour triangle must be one level finer than t was, and have
      // its long edge along this short edge of ours.
      assert(tribuf[neighbour_t].sub_level == tribuf[t].sub_level - 1);
      assert(tribuf[neighbour_t].nb == t);
      tribuf[neighbour_t].nb = t_child;
    }
  }

  // Write the 'low' child of t over t's information.
  tribuf[t].level_index *= 2;
  tribuf[t].sub_level--;
  tribuf[t].diamond_index = d_index;
  tribuf[t].v1 = tribuf[t].v0;
  tribuf[t].v0 = tribuf[t].vc;
  tribuf[t].vc = vb;

  if (is_textured) {
    // We are part of the same texture-chain as before.

    // Our previous link does not change.

    // Our next link is to our parent's other child.
    tribuf[t].next_tex = t_child;
  }

  if (tribuf[t].sub_level > 0) {
    // Calculate this triangle's new split priority.
    old_priority = tribuf[t].split_queue_info.priority;
    calculate_tri_priority(t);
#if DEBUG_MONOTONICITY
    if (old_priority - tribuf[t].split_queue_info.priority < -PRIORITY_TOLERANCE) {
      // The parent's priority was LESS than this child's new priority. This should
      // never happen.
      fprintf(debug_file, "MONOTONICITY VIOLATION on split(%d) child0: child metric %f > parent metric %f\n",
              t, tribuf[t].split_queue_info.priority, old_priority);

      print_error_metric = TRUE;
      fprintf(debug_file, "child metric...\n");
      calculate_tri_priority(t, TRUE);
      double new_priority = tribuf[t].split_queue_info.priority;


      // Save the parent's child vertices (temporarily)
      int save_vc = tribuf[t].vc;
      int save_v0 = tribuf[t].v0;
      int save_v1 = tribuf[t].v1;

      // Restore the parent's settings to what they were before we wrote the
      // split child into it.
      tribuf[t].level_index /= 2;
      tribuf[t].sub_level++;
      tribuf[t].vc = tribuf[t_child].v1;
      tribuf[t].v0 = tribuf[t].v1;
      tribuf[t].v1 = tribuf[t_child].v0;

      fprintf(debug_file, "parent metric...\n");
      calculate_tri_priority(t, TRUE);
      old_priority = tribuf[t].split_queue_info.priority;

      if (old_priority - new_priority < -PRIORITY_TOLERANCE) {
        fprintf(debug_file, "VIOLATION CONFIRMED - shutting down\n");
        premature_shutdown();
      } else {
        fprintf(debug_file, "Ah... false alarm?\n");

        // Set the parent's child settings back up again.
        tribuf[t].level_index *= 2;
        tribuf[t].sub_level--;
        tribuf[t].vc = save_vc;
        tribuf[t].v0 = save_v0;
        tribuf[t].v1 = save_v1;
      }
    }
#endif

    // It must be lower priority now, as children's priorities are guaranteed to be lower than their parents.
    // Thus we bubble this element DOWN the priority queue's heap.
    tri_queue->bubble_down(&tribuf[t].split_queue_info);
  } else {
    // This triangle is not splittable again; we won't be using the storage in the priority queue
    // for our children, so we must remove it.
    tri_queue->remove_element(&tribuf[t].split_queue_info);
    tribuf[t].split_queue_info.priority = PRIORITY_NEVER;

    tri_recompute_queue->remove_element(&tribuf[t].recompute_queue_info);
    tribuf[t].recompute_queue_info.priority = DIST_INFINITY;
  }

  // Our neighbour along v0-v1 is the old t's neighbour along vc-v0.
  // Our neighbour's adjacency doesn't need to be updated because it now
  // points to us, yet we are stored in the same space as our parent was.
  tribuf[t].nb = tribuf[t].n0;
  tribuf[t].n0 = t_child;
  tribuf[t].n1 = base_t_child;
}

// Make a new mergeable diamond.
void Bintree::make_mergeable_diamond(int t0, int t1, int t2, int t3)
{
  int d;

  d = new_diamond();
  diamondbuf[d].t0 = t0;
  diamondbuf[d].t1 = t1;
  diamondbuf[d].t2 = t2;
  diamondbuf[d].t3 = t3;

  assert((t0 >= 0) && (t0 < tribuf_size) && (tribuf[t0].next_free == IN_USE));
  assert((t1 >= 0) && (t1 < tribuf_size) && (tribuf[t1].next_free == IN_USE));
  assert((tribuf[t0].diamond_index == -1) && (tribuf[t1].diamond_index == -1));
  tribuf[t0].diamond_index = d;
  tribuf[t1].diamond_index = d;
  if (t2 != -1) {
    assert((t2 >= 0) && (t2 < tribuf_size) && (tribuf[t2].next_free == IN_USE));
    assert((t3 >= 0) && (t3 < tribuf_size) && (tribuf[t3].next_free == IN_USE));
    assert((tribuf[t2].diamond_index == -1) && (tribuf[t3].diamond_index == -1));
    tribuf[t2].diamond_index = d;
    tribuf[t3].diamond_index = d;
  }

  // Set up the priority for this new diamond.
  setup_diamond_priority(d);
}

// Break a mergeable diamond (because one of the triangles
// concerned is being split).
void Bintree::break_diamond(int diamond_index)
{
  if (diamond_index == -1)
    return;

  int t0, t1, t2, t3;

  assert((diamond_index >= 0) && (diamond_index < diamondbuf_size));
  assert(diamondbuf[diamond_index].next_free == IN_USE);

  // Remove this diamond from the priority queue.
  assert(diamondbuf[diamond_index].merge_queue_info.priq_location != -1);
  diamond_queue->remove_element(&diamondbuf[diamond_index].merge_queue_info);

  // Remove this diamond from the priority recomputation priority queue.
  assert(diamondbuf[diamond_index].recompute_queue_info.priq_location != -1);
  diamond_recompute_queue->remove_element(&diamondbuf[diamond_index].recompute_queue_info);

  t0 = diamondbuf[diamond_index].t0;  t1 = diamondbuf[diamond_index].t1;
  t2 = diamondbuf[diamond_index].t2;  t3 = diamondbuf[diamond_index].t3;

  assert((t0 >= 0) && (t0 < tribuf_size) && (tribuf[t0].next_free == IN_USE));
  assert(tribuf[t0].diamond_index == diamond_index);
  tribuf[t0].diamond_index = -1;

  assert((t1 >= 0) && (t1 < tribuf_size) && (tribuf[t1].next_free == IN_USE));
  assert(tribuf[t1].diamond_index == diamond_index);
  tribuf[t1].diamond_index = -1;

  if (t2 != -1) {
    assert((t2 >= 0) && (t2 < tribuf_size) && (tribuf[t2].next_free == IN_USE));
    assert(tribuf[t2].diamond_index == diamond_index);
    tribuf[t2].diamond_index = -1;

    assert((t3 >= 0) && (t3 < tribuf_size) && (tribuf[t3].next_free == IN_USE));
    assert(tribuf[t3].diamond_index == diamond_index);
    tribuf[t3].diamond_index = -1;
  } else {
    assert(t3 == -1);
  }

  free_diamond(diamond_index);
}

// Merge a diamond.
void Bintree::merge_diamond(int d)
{
  int sublevel, t0, t1, t2, t3;
  unsigned short m_index;

  assert((d >= 0) && (d < diamondbuf_size) && (diamondbuf[d].next_free == IN_USE));
  t0 = diamondbuf[d].t0; t1 = diamondbuf[d].t1; t2 = diamondbuf[d].t2; t3 = diamondbuf[d].t3;

  // Check that t0 and t1 are both valid, both from the same sublevel, and are the children of the same parent
  // (t0 low, t1 high).
  assert((t0 >= 0) && (t0 < tribuf_size) && (tribuf[t0].next_free == IN_USE) && (tribuf[t0].diamond_index == d));
  assert((t1 >= 0) && (t1 < tribuf_size) && (tribuf[t1].next_free == IN_USE) && (tribuf[t1].diamond_index == d));
  assert(tribuf[t0].sub_level == tribuf[t1].sub_level);
  assert((tribuf[t0].level_index / 2) == (tribuf[t1].level_index / 2));
  assert(tribuf[t0].level_index == (tribuf[t1].level_index - 1));

  sublevel = tribuf[t0].sub_level;
  assert(sublevel < max_subdiv_level);

  // Is there a second half to the diamond? There may not be if we are adjacent to the edge
  // of the triangulation.
  if (t2 == -1) {
    assert(t3 == -1);
  } else {
    // Check that t2 and t3 are both valid, both from the same sublevel as t0, and are the children of the same parent
    // (t2 low, t3 high).
    assert((t2 >= 0) && (t2 < tribuf_size) && (tribuf[t2].next_free == IN_USE) && (tribuf[t2].diamond_index == d));
    assert((t3 >= 0) && (t3 < tribuf_size) && (tribuf[t3].next_free == IN_USE) && (tribuf[t3].diamond_index == d));
    assert(tribuf[t2].sub_level == sublevel);
    assert(tribuf[t3].sub_level == sublevel);
    assert((tribuf[t2].level_index / 2) == (tribuf[t3].level_index / 2));
    assert(tribuf[t2].level_index == (tribuf[t3].level_index - 1));
  }

  // If the center vertex was morphing, remove its entry in the morph chain.
  m_index = vertex_array[2 * tribuf[t0].vc + 1];
  if (m_index != 0) {
//    fprintf(debug_file, "merging at vertex %d, deleting morph vertex %d\n", tribuf[t0].vc, m_index);
//    fflush(debug_file);
    free_morph(m_index);
    vertex_array[2 * tribuf[t0].vc + 1] = 0;
  }

  merge_triangles(t2, t0, t1);
  if (t2 != -1)
    merge_triangles(t0, t2, t3);

  // Remove this diamond from the priority queue.
  assert(diamondbuf[d].merge_queue_info.priq_location != -1);
  diamond_queue->remove_element(&diamondbuf[d].merge_queue_info);

  // Remove this diamond from the priority recomputation priority queue.
  assert(diamondbuf[d].recompute_queue_info.priq_location != -1);
  diamond_recompute_queue->remove_element(&diamondbuf[d].recompute_queue_info);

  free_diamond(d);
}

// Internal routine for merge_diamond; merges a pair of triangles, updates
// neighbour information and checks for newly mergeable diamonds.
void Bintree::merge_triangles(int neighbour_base, int child0, int child1)
{
  int neighbour_t, other_t, farside_t, merge_sublevel, link_t;

  // Write the parent triangle's information into what was our 'low' child.
  tribuf[child0].level_index /= 2;
  merge_sublevel = ++tribuf[child0].sub_level;
  tribuf[child0].diamond_index = -1;     // real val found later on in this routine
  tribuf[child0].vc = tribuf[child0].v0;
  tribuf[child0].v0 = tribuf[child0].v1;
  tribuf[child0].v1 = tribuf[child1].v0;

  if (is_textured) {
    // We are part of the same texture-chain as before - no change there.
    assert(tribuf[child0].texture_index == tribuf[child1].texture_index);

    // Our previous link is our low-child's previous link - since we
    // are stored in the same array element, this does not change.

    // Our children should have been linked together!
    assert(tribuf[child0].next_tex == child1);
    assert(tribuf[child1].prev_tex == child0);

    // Our next link is our high-child's next link.
    tribuf[child0].next_tex = link_t = tribuf[child1].next_tex;
    if (link_t == -1) {
      // Our high child was the last tri in the texture-chain. Now we are.
      last_tex_tri[tribuf[child1].texture_index] = child0;
    } else {
      // Link up the next tri to us.
      assert((link_t >= 0) && (link_t < tribuf_size) && (tribuf[link_t].next_free == IN_USE));
      assert(tribuf[link_t].prev_tex == child1);
      tribuf[link_t].prev_tex = child0;
    }
  }

  // Calculate this triangle's new split priority.
  calculate_tri_priority(child0);

  // Store this triangle's new split priority (taken from out of the diamond),
  // and change its position in the splittable-triangle queue.
  if (tribuf[child0].split_queue_info.priq_location == -1) {
    // Insert this triangle into the splittable-triangle queue (it wasn't before
    // because it was too fine to be split).
    assert(merge_sublevel == 1);
    tri_queue->insert_element(&tribuf[child0].split_queue_info);
  } else {
    // We must be higher priority now, as parents' priorities are guaranteed
    // to be higher than their children's. Thus we bubble this element UP the
    // priority queue's heap.
    tri_queue->bubble_up(&tribuf[child0].split_queue_info);
  }

  // Our neighbour along v0-v1 will be updated from merge_diamond, unless
  // it is the void, in which case nothing needs to be done. Our neighbour
  // along vc-v0 still points to the same storage location; what was our
  // low child is replaced by us. We only need to update our neighbour along
  // vc-v1.
  tribuf[child0].n0 = tribuf[child0].nb;
  tribuf[child0].n1 = neighbour_t = tribuf[child1].nb;
  tribuf[child0].nb = neighbour_base;

  if (neighbour_t != -1) {
    if (tribuf[neighbour_t].sub_level == merge_sublevel) {
      // This neighbour triangle is now the same level as us, so it must have
      // its short edge along this short edge of ours. Looking at the geometry
      // and relying on the fact that two touching triangles of the same level
      // always form a larger whole triangle (and not a parallelogram), the
      // neighbour's vc-v0 edge must have been the one touching.
      assert(tribuf[neighbour_t].n0 == child1);
      tribuf[neighbour_t].n0 = child0;
    } else {
      // This neighbour triangle must have been the same level as our child
      // triangle, and had its long edge touching our child's long edge (which
      // is now our short vc-v1 edge).
      assert(tribuf[neighbour_t].sub_level == merge_sublevel - 1);
      assert(tribuf[neighbour_t].nb == child1);
      tribuf[neighbour_t].nb = child0;
    } 
  }

  // Remove our other child from the split priority queue and recomputation queue
  // (if it was there).
  if (tribuf[child1].split_queue_info.priq_location != -1)
    tri_queue->remove_element(&tribuf[child1].split_queue_info);

  if (tribuf[child1].recompute_queue_info.priq_location != -1)
    tri_recompute_queue->remove_element(&tribuf[child1].recompute_queue_info);

  free_triangle(child1);

  // Check to see if this merging might have formed another mergeable diamond.
  if (merge_sublevel == max_subdiv_level) {
    // We are a top-level triangle and can't be merged.
    return;
  }

  if (tribuf[child0].level_index % 2) {
    // We are the 'high' child of our parent, and so we _must_ have a 'low' sibling
    // adjacent to us along our vc-v1 edge [this happens to already be
    // stored in neighbour_t from the previous calculation].
    assert(neighbour_t != -1);
    if (tribuf[neighbour_t].sub_level != merge_sublevel) {
      // Our sibling has been split. This diamond cannot be merged.
      return;
    }

    // Our sibling is intact; if there isn't another triangle adjacent to our n0 (i.e
    // if our parent was adjacent to the void) then we can form a mergeable half-diamond
    // right now out of [neighbour_t, child0].
    other_t = tribuf[child0].n0;
    if (other_t == -1) {
      make_mergeable_diamond(neighbour_t, child0, -1, -1);
      return;
    }

    // There is a triangle adjacent to us; if it is from a different level then we cannot
    // form a mergeable diamond.
    if (tribuf[other_t].sub_level != merge_sublevel)
      return;

    // The triangle adjacent to us is of the same sublevel as us; by the geometry rules mentioned
    // earlier we know that it must have its vc-v1 edge adjacent to our vc-v0 edge. Thus we only
    // need the triangle on its vc-v0 edge (and there is guaranteed to be _something_ there) in order
    // to complete the diamond.
    farside_t = tribuf[other_t].n0;
    assert(farside_t != -1);
    if (tribuf[farside_t].sub_level != merge_sublevel)
      return;

    // This farside_t is at the same sublevel as other_t and so they must be from the same parent
    // triangle. Because we are the 'high' child, other_t must be the 'low' sibling to farside_t's 'high'.
    assert(tribuf[other_t].level_index + 1 == tribuf[farside_t].level_index);

    // We have all four triangles at the same level - form a mergeable diamond.
    make_mergeable_diamond(neighbour_t, child0, other_t, farside_t);
    return;
  } else {
    // We are the 'low' child of our parent, and so we _must_ have a 'high' sibling
    // adjacent to us along our vc-v0 edge. Find this neighbour_t.
    neighbour_t = tribuf[child0].n0;
    assert(neighbour_t != -1);
    if (tribuf[neighbour_t].sub_level != merge_sublevel) {
      // Our sibling has been split. This diamond cannot be merged.
      return;
    }

    // Our sibling is intact; if there isn't another triangle adjacent to our n1 (i.e
    // if our parent was adjacent to the void) then we can form a mergeable half-diamond
    // right now out of [child0, neighbour_t].
    other_t = tribuf[child0].n1;
    if (other_t == -1) {
      make_mergeable_diamond(child0, neighbour_t, -1, -1);
      return;
    }

    // There is a triangle adjacent to us; if it is from a different level then we cannot
    // form a mergeable diamond.
    if (tribuf[other_t].sub_level != merge_sublevel)
      return;

    // The triangle adjacent to us is of the same sublevel as us; by the geometry rules mentioned
    // earlier we know that it must have its vc-v0 edge adjacent to our vc-v1 edge. Thus we only
    // need the triangle on its vc-v1 edge (and there is guaranteed to be _something_ there) in order
    // to complete the diamond.
    farside_t = tribuf[other_t].n1;
    assert(farside_t != -1);
    if (tribuf[farside_t].sub_level != merge_sublevel)
      return;

    // This farside_t is at the same sublevel as other_t and so they must be from the same parent
    // triangle. Because we are the 'low' child, other_t must be the 'high' sibling to farside_t's 'low'.
    assert(tribuf[other_t].level_index == tribuf[farside_t].level_index + 1);

    // We have all four triangles at the same level - form a mergeable diamond.
    make_mergeable_diamond(child0, neighbour_t, farside_t, other_t);
    return;
  }
}

// Draw the Bintree.
void Bintree::draw(void)
{
  int x, y, s, t, i, v, tri_index, tex_index, frust_index, sec_index;
  int vx, vy;
  GLfloat tx, ty;
  unsigned char thispt_flags;
  vec3 left_clipnorm, right_clipnorm, bottom_clipnorm, top_clipnorm;
  vec3 near_clipnorm, billboard_norm, billboard_side, upvec;
  vec3 worldspace_point, cameraspace_point, distvec;
  scenerytype *stype;

  // Set up the modelling transform... to move and scale the mesh.
  glScaled(mesh_delta.x, mesh_delta.y, mesh_delta.z);
  glTranslated(mesh_base.x, mesh_base.y, mesh_base.z);

  /*
  * Determine which sections are outside the view-frustrum and so
  * don't need to be drawn.
  */
  
  // All points initially start off totally outside the frustum.
  memset((void *) frustumflags, 0x1F, (x_sections + 2) * (y_sections + 2));
  
  // Set up the vectors for the five clipping planes in cameraspace.
  left_clipnorm.x = -camera_clip_near;
  left_clipnorm.y = 0;
  left_clipnorm.z = camera_frustum_size;
  
  right_clipnorm.x = camera_clip_near;
  right_clipnorm.y = 0;
  right_clipnorm.z = camera_frustum_size;
  
  bottom_clipnorm.x = 0;
  bottom_clipnorm.y = -camera_clip_near;
  bottom_clipnorm.z = camera_frustum_size;
  
  top_clipnorm.x = 0;
  top_clipnorm.y = camera_clip_near;
  top_clipnorm.z = camera_frustum_size;
  
  near_clipnorm.x = 0;
  near_clipnorm.y = 0;
  near_clipnorm.z = +1;
  
  // Transform these clipping normals into worldspace.
  left_clipnorm   = left_clipnorm   ^ camera_dir;
  right_clipnorm  = right_clipnorm  ^ camera_dir;
  bottom_clipnorm = bottom_clipnorm ^ camera_dir;
  top_clipnorm    = top_clipnorm    ^ camera_dir;
  near_clipnorm   = near_clipnorm   ^ camera_dir;
  
  // For every corner point of every section, check it
  // against these five clip planes.
  for (y = 0; y <= y_sections; y++) {
    frust_index = 1 + (y + 1) * (x_sections + 2);
    for (x = 0; x <= x_sections; x++) {
      thispt_flags = 0x00;
      
      // Set up the point's location.
      distvec.x = 32 * x * mesh_delta.x;
      distvec.y = 32 * y * mesh_delta.y;
      distvec.z = vertex_array[2 * ((32 * x) + (32 * y) * hf_size_x)] * mesh_delta.z;
      distvec += (mesh_base - camera_pos);
      
      if (distvec * left_clipnorm > FRUSTUMCLIPPING_SAFETY) {
        // This point lies outside the left clipping plane.
        thispt_flags |= 0x01;
      }
      if (distvec * right_clipnorm > FRUSTUMCLIPPING_SAFETY) {
        // This point lies outside the right clipping plane.
        thispt_flags |= 0x02;
      }
      if (distvec * bottom_clipnorm > FRUSTUMCLIPPING_SAFETY) {
        // This point lies outside the bottom clipping plane.
        thispt_flags |= 0x04;
      }
      if (distvec * top_clipnorm > FRUSTUMCLIPPING_SAFETY) {
        // This point lies outside the top clipping plane.
        thispt_flags |= 0x08;
      }
      if (distvec * near_clipnorm > FRUSTUMCLIPPING_SAFETY - camera_clip_near) {
        // This point lies behind the near clipping plane.
        thispt_flags |= 0x10;
      }
      
      // thispt_flags now holds five bits indicating which planes it lies outside. We bitwise
      // AND it with the four sections which it borders... if it lies inside any planes, it will
      // clear their bits and so indicate that they have to be drawn this frame.
      frustumflags[frust_index                       ] &= thispt_flags;
      frustumflags[frust_index - 1                   ] &= thispt_flags;
      frustumflags[frust_index     - (x_sections + 2)] &= thispt_flags;
      frustumflags[frust_index - 1 - (x_sections + 2)] &= thispt_flags;
      
      frust_index++;
    }
  }
  
  if (is_textured && use_texturing) {

    // Draw the texture-mapped mesh. Go to texture-transform mode and set up
    // the base color.
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    glColor3f(1.0, 1.0, 1.0);

    // Eventually there will be triangle-stripping code in here. For now we just draw every
    // triangle individually, organised by sections for texturing purposes.
    tex_index = 0;
    for (y = 0; y < y_sections; y++) {
      frust_index = 1 + (y + 1) * (x_sections + 2);
      for (x = 0; x < x_sections; x++, tex_index++) {
        // Don't draw any sections for which all of their points are outside the
        // view frustum in some firection.
        if (frustumflags[frust_index++])
          continue;

        if (use_texgen) {
          // Texgen gives us coordinates in terms of sections, so we must
          // translate these coordinates by -(x, y) to get coordinates in [0, 1].
          // Note that texgen assumes that mesh_base is 0, 0.
          glLoadIdentity();
          glTranslatef((GLfloat) -x, (GLfloat) -y, 0);
        }

        // Set up the texture.
        glBindTexture(GL_TEXTURE_2D, texture_handle[tex_index]);
        glBegin(GL_TRIANGLES);
        
        // We traverse along the next_tex and prev_tex indices stored in every
        // triangle.

        if (use_texgen) {
          // We use automatic texture-coordinate generation.
          for (t = first_tex_tri[tex_index]; t != -1; t = tribuf[t].next_tex) {
            assert((t >= 0) && (t < tribuf_size) && (tribuf[t].next_free == IN_USE));
            
            v = tribuf[t].vc;
            glVertex3f((GLfloat) (v % hf_size_x),
              (GLfloat) (v / hf_size_x),
              (GLfloat) (get_height(v)));
            
            v = tribuf[t].v0;
            glVertex3f((GLfloat) (v % hf_size_x),
              (GLfloat) (v / hf_size_x),
              (GLfloat) (get_height(v)));
            
            v = tribuf[t].v1;
            glVertex3f((GLfloat) (v % hf_size_x),
              (GLfloat) (v / hf_size_x),
              (GLfloat) (get_height(v)));
          }
        } else {
          // We must generate texture coordinates manually.
          for (t = first_tex_tri[tex_index]; t != -1; t = tribuf[t].next_tex) {
            assert((t >= 0) && (t < tribuf_size) && (tribuf[t].next_free == IN_USE));
            
            v = tribuf[t].vc;
            vx = v % hf_size_x;
            vy = v / hf_size_x;
            tx = vx / 32.0f - x;
            ty = vy / 32.0f - y;
            glTexCoord2f(tx, ty);
            glVertex3f((GLfloat) vx, (GLfloat) vy, (GLfloat) (get_height(v)));
            
            v = tribuf[t].v0;
            vx = v % hf_size_x;
            vy = v / hf_size_x;
            tx = vx / 32.0f - x;
            ty = vy / 32.0f - y;
            glTexCoord2f(tx, ty);
            glVertex3f((GLfloat) vx, (GLfloat) vy, (GLfloat) (get_height(v)));
            
            v = tribuf[t].v1;
            vx = v % hf_size_x;
            vy = v / hf_size_x;
            tx = vx / 32.0f - x;
            ty = vy / 32.0f - y;
            glTexCoord2f(tx, ty);
            glVertex3f((GLfloat) vx, (GLfloat) vy, (GLfloat) (get_height(v)));
          }
        }
        glEnd();
      }
    }

    /*
     * Draw the scenery.
     */

    // Billboarding... all scenery objects face the camera (+Z). Construct
    // an orthonormal coordinate system from the vector towards the camera
    // and +Z.
    upvec.x = 0;
    upvec.y = 0;
    upvec.z = 1;

    billboard_norm = upvec ^ camera_dir;

    // Turn off texture generation for scenery. Also set the texture transform
    // to the identity. Enable the alpha test.
    if (use_texgen) {
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
    }
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_LESS, 0.8f);
    glLoadIdentity();
    for (i = 0, stype = scenerytypearray; i < num_scenery_types; i++, stype++) {
      // Bind the scenery's texture.
      glBindTexture(GL_TEXTURE_2D, stype->texture);

      glBegin(GL_QUADS);
      for (s = stype->first_sceneryobj; s != -1; s = sceneryarray[s].next_obj) {
        // Don't draw any scenery that lies in sections that aren't being drawn.
        if (frustumflags[sceneryarray[s].frustum_index])
          continue;

#if 0
        billboard_side = upvec ^ (camera_pos - sceneryarray[s].world_location);
#else
        billboard_side = upvec ^ billboard_norm;
#endif
        normalize(billboard_side);

        glTexCoord2f(0.0f, 0.0f);
        glVertex3f((GLfloat) (sceneryarray[s].location.x - stype->x_radius * billboard_side.x),
                   (GLfloat) (sceneryarray[s].location.y - stype->x_radius * billboard_side.y),
                   (GLfloat) sceneryarray[s].location.z);

        glTexCoord2f(1.0f, 0.0f);
        glVertex3f((GLfloat) (sceneryarray[s].location.x + stype->x_radius * billboard_side.x),
                   (GLfloat) (sceneryarray[s].location.y + stype->x_radius * billboard_side.y),
                   (GLfloat) sceneryarray[s].location.z);

        glTexCoord2f(1.0f, 1.0f);
        glVertex3f((GLfloat) (sceneryarray[s].location.x + stype->x_radius * billboard_side.x),
                   (GLfloat) (sceneryarray[s].location.y + stype->x_radius * billboard_side.y),
                   (GLfloat) (sceneryarray[s].location.z + stype->y_height));

        glTexCoord2f(0.0f, 1.0f);
        glVertex3f((GLfloat) (sceneryarray[s].location.x - stype->x_radius * billboard_side.x),
                   (GLfloat) (sceneryarray[s].location.y - stype->x_radius * billboard_side.y),
                   (GLfloat) (sceneryarray[s].location.z + stype->y_height));
      }
      glEnd();
    }
    // Restore settings for the terrain display.
    if (use_texgen) {
      glEnable(GL_TEXTURE_GEN_S);
      glEnable(GL_TEXTURE_GEN_T);
    }
    glDisable(GL_ALPHA_TEST);

    // Go back to modelview matrix mode.
    glMatrixMode(GL_MODELVIEW);
  } else {
    // Eventually there will be triangle-stripping code in here. For now we just draw every
    // triangle individually.
    glBegin(GL_TRIANGLES);
    for (t = 0; t < tribuf_size; t++) {
      if (tribuf[t].next_free != IN_USE)
        continue;

      // Which section are we in?
      sec_index = tribuf[t].level_index >> (1 + max_subdiv_level - tribuf[t].sub_level);
      frust_index = ((sec_index % x_sections) + 1) + ((sec_index / x_sections) + 1) * (x_sections + 2);

      // Don't draw this section if it is outside the view frustum.
      if (frustumflags[frust_index])
        continue;


      tri_index = precalc_index(tribuf[t].sub_level, tribuf[t].level_index);
      glColor3f((GLfloat) precalc_array[tri_index].tri_color.x, 
                (GLfloat) precalc_array[tri_index].tri_color.y, 
                (GLfloat) precalc_array[tri_index].tri_color.z);

        v = tribuf[t].vc;
        glVertex3f((GLfloat) (v % hf_size_x),
                   (GLfloat) (v / hf_size_x),
                   (GLfloat) (get_height(v)));

        v = tribuf[t].v0;
        glVertex3f((GLfloat) (v % hf_size_x),
                   (GLfloat) (v / hf_size_x),
                   (GLfloat) (get_height(v)));

        v = tribuf[t].v1;
        glVertex3f((GLfloat) (v % hf_size_x),
                   (GLfloat) (v / hf_size_x),
                   (GLfloat) (get_height(v)));
    }
    glEnd();
  }
}

// Set up the split priority of a triangle. Takes a newly created triangle with
// vertices etc. set up, calculates its priority and inserts it into both
// tri_queue and the triangle recomputation queue.
void Bintree::setup_triangle_priority(int tri_index)
{
  // This triangle is not yet in the recomputation queue.
  tribuf[tri_index].recompute_queue_info.priq_location = -1;

  // Calculate the triangle's priority and put it into the recomputation queue.
  calculate_tri_priority(tri_index);

  // Insert the triangle into the split priority queue (it won't already
  // be there because it was just created).
  tri_queue->insert_element(&tribuf[tri_index].split_queue_info);
}

// Calculate the split priority of a triangle, and its distance before it must
// be recomputed. Update the triangle's position in the priority recomputation priority queue.
// DO NOT insert the triangle into the split priority queue... other code must do that.
//
// Called from: setup_triangle_priority   recompute_priorities
void Bintree::calculate_tri_priority(int tri_index, int just_do_calculation)
{
  double old_recompute_dist, move_dist, threshold;
  double new_recompute_dist;

  // Call the error metric. We want to know the distance until the split priority would
  // be greater than the current biggest_split_priority.
  assert(threshold_priorities_set);
  threshold = (lock_thresholds) ? threshold_lockval : biggest_split_priority;
  tribuf[tri_index].split_queue_info.priority = (this->*error_metric)
      (tribuf[tri_index].vc, tribuf[tri_index].v0, tribuf[tri_index].v1,
       tribuf[tri_index].sub_level, tribuf[tri_index].level_index,
       TRUE, threshold, &move_dist);

  // Debugging bailout.
  if (just_do_calculation)
    return;

  if (tribuf[tri_index].split_queue_info.priority > biggest_split_priority)
    biggest_split_priority = tribuf[tri_index].split_queue_info.priority;

  // We have the distance that we must move before this triangle's priority hits
  // the split threshold - in theory. We need to work out what value of
  // current_distance_moved it would be safe to perform the recomputation at.
  old_recompute_dist = tribuf[tri_index].recompute_queue_info.priority;
  if (tribuf[tri_index].split_queue_info.priority == PRIORITY_NEVER)
    new_recompute_dist = current_distance_moved + DIST_MAX_DEFERRAL;
  else
    new_recompute_dist = current_distance_moved + RECOMPUTE_SAFETY_FACTOR * move_dist;
  assert(new_recompute_dist >= 0);
  tribuf[tri_index].recompute_queue_info.priority = new_recompute_dist;

  // We now need to insert the triangle into the recomputation priority queue.
  if (tribuf[tri_index].recompute_queue_info.priq_location == -1) {
    // It wasn't before. Insert it fresh.
    tri_recompute_queue->insert_element(&tribuf[tri_index].recompute_queue_info);
  } else {
    // The triangle already exists in the recomputation priority queue.
    if (new_recompute_dist > old_recompute_dist) {
      // The new recomputation distance is greater than the old one; hardly
      // a big surprise. Bubble the triangle down the queue - it is now lower priority.
      tri_recompute_queue->bubble_down(&tribuf[tri_index].recompute_queue_info);
    } else {
      // The new recomputation distance is SMALLER than the old one - should this ever happen?
      // Bubble the triangle up the queue, it is now very high priority.
      tri_recompute_queue->bubble_up(&tribuf[tri_index].recompute_queue_info);
    }
  }
}

// Set up the merge priority of a diamond. Takes a newly created diamond with
// t0, t1, t2, t3 set up, calculates its priority and inserts it into
// diamond_queue and the diamond recomputation queue.
void Bintree::setup_diamond_priority(int d)
{
  // This diamond is not yet in the recomputation queue.
  diamondbuf[d].recompute_queue_info.priq_location = -1;

  // Calculate the diamond's priority and put it into the recomputation queue.
  calculate_diamond_priority(d);

  // Insert the diamond into the merge priority queue (it won't already
  // be there because it was just created).
  diamond_queue->insert_element(&diamondbuf[d].merge_queue_info);
}

// Calculate the merge priority of a diamond, and its distance before it must
// be recomputed. Update the diamond's position in the priority recomputation priority queue.
// DO NOT insert the diamond into the merge priority queue... other code must do that.
//
// Called from: setup_diamond_priority   recompute_priorities
void Bintree::calculate_diamond_priority(int d)
{
  double p0, p1, move_dist0, move_dist1, old_recompute_dist, new_recompute_dist;

  // Calculate the priority that the t0-t1 parent triangle would have.
  diamondbuf[d].parent0_priority = p0 = calculate_diamondside_priority(diamondbuf[d].t0, diamondbuf[d].t1, &move_dist0);

  if (diamondbuf[d].t2 == -1) {
    // This is a half-diamond.
    diamondbuf[d].parent1_priority = PRIORITY_NEVER;
  } else {
    // Calculate the priority that the t2-t3 parent triangle would have.
    // The diamond's priority is the maximum of this and the t0-t1 triangle.
    diamondbuf[d].parent1_priority = p1 = calculate_diamondside_priority(diamondbuf[d].t2, diamondbuf[d].t3, &move_dist1);

    p0 = MAX(p0, p1);
    move_dist0 = MAX(move_dist0, move_dist1);   // because they both have to be below threshold in order for it to be of interest
  }

  // Store this diamond's priority.
  diamondbuf[d].merge_queue_info.priority = p0;
  if (diamondbuf[d].merge_queue_info.priority < smallest_diamond_priority)
    smallest_diamond_priority = diamondbuf[d].merge_queue_info.priority;

  // We have the distance that we must move before this diamond's priority falls below
  // the merge threshold - in theory. We need to work out what value of
  // current_distance_moved it would be safe to perform the recomputation at.
  old_recompute_dist = diamondbuf[d].recompute_queue_info.priority;
  new_recompute_dist = current_distance_moved + RECOMPUTE_SAFETY_FACTOR * move_dist0;
  diamondbuf[d].recompute_queue_info.priority = new_recompute_dist;

  // We now need to insert the diamond into the recomputation priority queue.
  if (diamondbuf[d].recompute_queue_info.priq_location == -1) {
    // It wasn't before. Insert it fresh.
    diamond_recompute_queue->insert_element(&diamondbuf[d].recompute_queue_info);
  } else {
    // The diamond already exists in the recomputation priority queue.
    if (new_recompute_dist > old_recompute_dist) {
      // The new recomputation distance is greater than the old one; hardly
      // a big surprise. Bubble the diamond down the queue - it is now lower priority.
      diamond_recompute_queue->bubble_down(&diamondbuf[d].recompute_queue_info);
    } else {
      // The new recomputation distance is SMALLER than the old one - should this ever happen?
      // Bubble the diamond up the queue, it is now very high priority.
      diamond_recompute_queue->bubble_up(&diamondbuf[d].recompute_queue_info);
    }
  }
}

// Calculate the split priority that would be found if two children
// were merged.
double Bintree::calculate_diamondside_priority(int child0, int child1, double *move_dist)
{
  double returnval, threshold;

  assert(tribuf[child0].sub_level == tribuf[child1].sub_level);
  assert(tribuf[child0].level_index + 1 == tribuf[child1].level_index);

  assert(threshold_priorities_set);
  threshold = (lock_thresholds) ? threshold_lockval : biggest_split_priority;
  returnval = (this->*error_metric)(tribuf[child0].v0, tribuf[child0].v1, tribuf[child1].v0,
                                    tribuf[child0].sub_level + 1, tribuf[child0].level_index / 2,
                                    FALSE, threshold, move_dist);

#if DEBUG_MONOTONICITY
  if (returnval - tribuf[child0].split_queue_info.priority < -PRIORITY_TOLERANCE) {
    // The calculated diamond-side parent priority is LESS than the child's priority. This should
    // never happen.
    fprintf(debug_file, "MONOTONICITY VIOLATION on diamondside_priority[child0]: child metric %f > parent metric %f\n",
            tribuf[child0].split_queue_info.priority, returnval);
    print_error_metric = TRUE;
    fprintf(debug_file, "child metric...\n");
    calculate_tri_priority(child0, TRUE);
    
    fprintf(debug_file, "parent metric...\n");
    returnval = (this->*error_metric)(tribuf[child0].v0, tribuf[child0].v1, tribuf[child1].v0,
                                      tribuf[child0].sub_level + 1, tribuf[child0].level_index / 2,
                                      FALSE, threshold, move_dist);

    if (returnval - tribuf[child0].split_queue_info.priority < -PRIORITY_TOLERANCE) {
      fprintf(debug_file, "VIOLATION CONFIRMED - shutting down\n");
      premature_shutdown();
    } else {
      fprintf(debug_file, "Ah... false alarm?\n");
    }
  }

  if (returnval - tribuf[child1].split_queue_info.priority < -PRIORITY_TOLERANCE) {
    // The calculated diamond-side parent priority is LESS than the child's priority. This should
    // never happen.
    fprintf(debug_file, "MONOTONICITY VIOLATION on diamondside_priority[child1]: child metric %f > parent metric %f\n",
            tribuf[child1].split_queue_info.priority, returnval);
    print_error_metric = TRUE;
    fprintf(debug_file, "child metric...\n");
    calculate_tri_priority(child1, TRUE);
    
    fprintf(debug_file, "parent metric...\n");
    returnval = (this->*error_metric)(tribuf[child0].v0, tribuf[child0].v1, tribuf[child1].v0,
                                      tribuf[child0].sub_level + 1, tribuf[child0].level_index / 2,
                                      FALSE, threshold, move_dist);

    if (returnval - tribuf[child1].split_queue_info.priority < -PRIORITY_TOLERANCE) {
      fprintf(debug_file, "VIOLATION CONFIRMED - shutting down\n");
      premature_shutdown();
    } else {
      fprintf(debug_file, "Ah... false alarm?\n");
    }
  }
#endif

  return returnval;
}

// Simple error-metric (debugging routine)
double Bintree::debug_metric(int vc, int v0, int v1, int sub_level, int level_index,
                             int dist_until_above_threshold, double threshold, double *dist_until)
{
  vec3 pc, p0, p1, minv, maxv, dist_v;
  double dist_from_box, error_dist;
  int tri_precalc_index;

  tri_precalc_index = precalc_index(sub_level, level_index);

  pc = vertex_world(vc) - camera_pos;
  p0 = vertex_world(v0) - camera_pos;
  p1 = vertex_world(v1) - camera_pos;

  error_dist = precalc_array[tri_precalc_index].error_bound * mesh_delta.z;
  pc.z += error_dist;

  minv = vmin(pc, vmin(p0, p1));
  maxv = vmax(pc, vmax(p0, p1));

  if (minv.x * maxv.x < 0)
    dist_v.x = 0;
  else if (minv.x > 0)
    dist_v.x = minv.x;
  else
    dist_v.x = -maxv.x;

  if (minv.y * maxv.y < 0)
    dist_v.y = 0;
  else if (minv.y > 0)
    dist_v.y = minv.y;
  else
    dist_v.y = -maxv.y;

  if (minv.z * maxv.z < 0)
    dist_v.z = 0;
  else if (minv.z > 0)
    dist_v.z = minv.z;
  else
    dist_v.z = -maxv.z;

  // Does not handle deferred recomputation at all.
  *dist_until = 0;

  dist_from_box = mag(dist_v);
  if (dist_from_box < 1e-03)
    return PRIORITY_MAX;
  else
    return sqrt(SQR(p1.x - p0.x) + SQR(p1.y - p0.y)) / dist_from_box;
}

// Screen-space distortion error-metric
double Bintree::screenspace_errormetric(int vc, int v0, int v1, int sub_level, int level_index,
                                        int dist_until_above_threshold, double threshold, double *dist_until)
{
  vec3 pc, p0, p1, errorvec;
  double disc_0, disc_1, disc_c, max_disc, min_r, screen_error;
  int tri_precalc_index;

  tri_precalc_index = precalc_index(sub_level, level_index);

  // Calculate the vertices' coordinates in camera-space.
  pc = cspace_meshbase + (vc % hf_size_x)   * cspace_xstep
                       + (vc / hf_size_x)   * cspace_ystep
                       + vertex_array[2*vc] * cspace_zstep;

  p0 = cspace_meshbase + (v0 % hf_size_x)   * cspace_xstep
                       + (v0 / hf_size_x)   * cspace_ystep
                       + vertex_array[2*v0] * cspace_zstep;

  p1 = cspace_meshbase + (v1 % hf_size_x)   * cspace_xstep
                       + (v1 / hf_size_x)   * cspace_ystep
                       + vertex_array[2*v1] * cspace_zstep;

  errorvec = precalc_array[tri_precalc_index].error_bound * cspace_zstep;

  // Does not handle deferred recomputation at all.
  *dist_until = 0;

  // NB: This next section causes massive trouble if we actually use it, without the
  // frustum-clipping.
#if 0
  // If the wedgie goes through z = 0, set its priority to maximum.
  if ((fabs(pc.z) <= errorvec.z) || (fabs(p0.z) <= errorvec.z) || (fabs(p1.z) <= errorvec.z))
    return PRIORITY_MAX;

  if ((pc.z * p0.z < 0) || (pc.z * p1.z < 0))
    return PRIORITY_MAX;

  // The wedgie lies completely on one side of z = 0. If it's behind z = 0, set its priority to zero.
  if (pc.z < 0)
    return PRIORITY_NEVER;
#else
  if ((p0.z < 0) || (p1.z < 0) || (pc.z < 0))
    return PRIORITY_NEVER;
#endif

  // For every point (p, q, r) and error vector (a, b, c) in cameraspace, we need to find the
  // maximum screen-space distortion dist = | (p+a)/(r+c) - (p-a)/(r-c), (q+b)/(r+c) - (q-b)/(r-c) |
  // which is equivalent to dist = 2/(r^2 - c^2) sqrt((ar - cp)^2 + (br - cq)^2)
  //
  // This value may have a maximum anywhere over the triangle. However the two components r
  // and disc = (ar - pc)^2 + (br - qc)^2 must have maxima / minima at the triangles' vertices.
  // Hence by taking the minimum r and the maximum of disc we may find a distance bound
  // dist_bound = 2/(min_r^2 - c^2) sqrt(max_disc) ... we use this as the error metric.

  min_r = MIN(pc.z, MIN(p0.z, p1.z));
  assert(min_r > errorvec.z);  // we know this because fabs(p.z) < errorvec.z causes early exit for any p

  disc_0 = SQR(errorvec.x * p0.z - errorvec.z * p0.x) + SQR(errorvec.y * p0.z - errorvec.z * p0.y);
  disc_1 = SQR(errorvec.x * p1.z - errorvec.z * p1.x) + SQR(errorvec.y * p1.z - errorvec.z * p1.y);
  disc_c = SQR(errorvec.x * pc.z - errorvec.z * pc.x) + SQR(errorvec.y * pc.z - errorvec.z * pc.y);
  max_disc = MAX(disc_c, MAX(disc_0, disc_1));

  screen_error = 2 * sqrt(max_disc) / (SQR(min_r) - SQR(errorvec.z));

  // Screenspace error is now equal to distance on the z=1 plane, effectively. To convert it to pixels, we must
  // multiply by window_radius * near_clip / camera_frustum_size.
  return screen_error * (viewresolution_x / 2) * camera_clip_near / camera_frustum_size;
}

// Distortion-angle error-metric
double Bintree::perpendicular_epsilon_metric(int vc, int v0, int v1, int sub_level, int level_index,
                                             int dist_until_above_threshold, double threshold, double *dist_until)
{
  int vcx, vcy, v0x, v0y, v1x, v1y, vcv0x, vcv0y;
  unsigned short vcz, v0z, v1z, tri_minz, tri_maxz;
  double epsilon, camera_u, camera_v, max_dx, min_dxsquared, min_dzsquared, priority;
  double tri_hf_cells, tri_hf_cells_sq, current_min_d, threshold_min_d;
  int tri_precalc_index, error_bound;
#if TEST_ERRORBOUND
  double err_c, err_0, err_1, dx, dx_sq, dz, dz_sq;
#endif

  // Work out the accuracy of this triangle's approximation.
  tri_precalc_index = precalc_index(sub_level, level_index);
  error_bound = precalc_array[tri_precalc_index].error_bound;

  // If this triangle is exactly accurate (presumably quite rare) then that's fine!
  if (error_bound == 0) {
    // Never needs to be recomputed; can be put off indefinitely.
    *dist_until = DIST_INFINITY;
    return 0;
  }

  // For a point p on the triangle, we look at the vector d from the camera c to this point.
  // Let's say the camera was looking straight at this point. Then suppose this point was shifted
  // by epsilon along Z; the distortion in screen-space would be equal to pd / |d| where pd is the
  // distance moved by the point in the plane perpendicular to the line from c to p.
  //
  // Constructing a few geometric diagrams, we may see that pd : epsilon as dx : |d| where dx is the
  // length of the horizontal component of d. Then the distortion in screen-space is proportional to
  // to epsilon * dx / |d|^2.
  // 
  // We must find the maximum value of this function over the surface of the triangle. Unfortunately
  // this is tricky; we can find an upper bound to the function by finding the maximum of dx
  // (occurs at a vertex of the triangle) and a lower bound to the minimum of |d|^2. epsilon is constant.
  //
  // The minimum of |d|^2 is found by examining the camera's barycentric coordinates w.r.t the triangle
  // and determining which point in the triangle is closest in 2D - we then take as the distance in Z
  // the minimum of |dz| for all three vertices. This can produce a lower bound for |d|^2 which is zero
  // - this means that the camera is both inside the triangle in 2D and also within its bounds in Z. We
  // set the triangle's priority to maximum in this case - in future maybe some more tests could go here.
  // 
  // dist_bound = epsilon * max_dx / (min_dxsquared + min_dzsquared) ... we use this as the error metric.

#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "computing error for [%d %d %d] (level %d / index %d @ %d).\n", 
            vc, v0, v1, sub_level, level_index, tri_precalc_index);
#endif

  // Calculate the vertices' coordinates in heightfield steps.
  vcx = vc % hf_size_x;  vcy = vc / hf_size_x;  vcz = vertex_array[2*vc];
  v0x = v0 % hf_size_x;  v0y = v0 / hf_size_x;  v0z = vertex_array[2*v0];
  v1x = v1 % hf_size_x;  v1y = v1 / hf_size_x;  v1z = vertex_array[2*v1];

  // We know that this is a right-isoceles triangle, and in one of eight orientations... we need
  // to work out the camera's barycentric coordinates w.r.t. the triangle. This will let us calculate
  // dx and |d|^2.

  // Work out the length of the vc-v0 side squared, in grid cells.
  vcv0x = v0x - vcx;
  vcv0y = v0y - vcy;
  tri_hf_cells_sq = SQR(vcv0x) + SQR(vcv0y);

  // camera u = (c - vc) . (v0 - vc) / (v0 - vc) . (v0 - vc)
  // camera v = (c - vc) . (v1 - vc) / (v1 - vc) . (v1 - vc)
  //
  // We know that v1 - vc is rotated 90 degrees counter-clockwise from v0 - vc
  // and has the same length, i.e. is [-vcv0y, vcv0x] - use this to cut down on costs.
#if DEBUG_MONOTONICITY
  if (print_error_metric) {
    fprintf(debug_file, "  triangle is [%d %d] [%d %d] [%d %d]\n", vcx, vcy, v0x, v0y, v1x, v1y);
    fprintf(debug_file, "  camera is [%f %f]\n", camera_hf_x, camera_hf_y);
  }
#endif
  camera_u = ((camera_hf_x - vcx) *  vcv0x + (camera_hf_y - vcy) * vcv0y) / tri_hf_cells_sq;
  camera_v = ((camera_hf_x - vcx) * -vcv0y + (camera_hf_y - vcy) * vcv0x) / tri_hf_cells_sq;
#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  camera at u, v %f, %f\n", camera_u, camera_v);
#endif

  // Take the square root to work out the real number of cells.
  tri_hf_cells = sqrt(tri_hf_cells_sq);

  // Determine which point on this triangle the camera is closest to, and
  // calculate the distance dx squared from the camera to this point, in barycentric space.
  // Multiply this by mesh_delta to get the distance in XY worldspace.
  assert(mesh_delta.x == mesh_delta.y); // this trick doesn't work otherwise
  min_dxsquared = get_closest_barycentricsq(camera_u, camera_v) * tri_hf_cells_sq * mesh_delta_sq;

#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  min_dxsquared %f\n", min_dxsquared);
#endif

  // Get the smallest Z distance from the camera to the triangle's bounds in Z.
  // This will be zero if the camera is within the triangle's bounds in Z.
#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  tri vertices [%d %d %d]\n", vcz, v0z, v1z);
#endif


  tri_minz = precalc_array[tri_precalc_index].minval;
  tri_maxz = precalc_array[tri_precalc_index].maxval;

#if DEBUG_MONOTONICITY
  if (print_error_metric) {
    fprintf(debug_file, "  tri_bounds %d, %d\n", tri_minz, tri_maxz);
    fprintf(debug_file, "  camera at z %f\n", camera_hf_z);
  }
#endif
  if (camera_hf_z < tri_minz) {
    min_dzsquared = SQR(mesh_delta.z * (tri_minz - camera_hf_z));
  } else if (camera_hf_z > tri_maxz) {
    min_dzsquared = SQR(mesh_delta.z * (camera_hf_z - tri_maxz));
  } else {
    min_dzsquared = 0;
  }
#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  -> min_dzsquared %f\n", min_dzsquared);
#endif

  if ((min_dxsquared == 0) && (min_dzsquared == 0)) {
    // The camera lies within the 2D extent of the triangle and also
    // lies within the Z bounds of the triangle (plus / minus errors). We give it maximum priority.
#if DEBUG_MONOTONICITY
    if (print_error_metric) {
      fprintf(debug_file, "  ... camera lies within bounds, returning PRIORITY_MAX\n");
      fflush(debug_file);
    }
#endif
    *dist_until = 0;
    return PRIORITY_MAX;
  }

  // Determine which vertex on this triangle the camera is furthest away from,
  // and calculate the square of the barycentric XY distance to this vertex.
  switch(get_furthest_barycentric(camera_u, camera_v)) {
  case 0: // v0
#if DEBUG_MONOTONICITY
    if (print_error_metric)
      fprintf(debug_file, "  furthest vertex v0\n");
#endif
    max_dx = SQR(camera_u - 1) + SQR(camera_v - 0);
    break;
  case 1: // v1
#if DEBUG_MONOTONICITY
    if (print_error_metric)
      fprintf(debug_file, "  furthest vertex v1\n");
#endif
    max_dx = SQR(camera_u - 0) + SQR(camera_v - 1);
    break;
  case 2: // vc
#if DEBUG_MONOTONICITY
    if (print_error_metric)
      fprintf(debug_file, "  furthest vertex vc\n");
#endif
    max_dx = SQR(camera_u - 0) + SQR(camera_v - 0);
    break;
  default:
    assert(0);
  }

  // Convert this to worldspace XT distance.
#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  furthest distance (barycentric_sq) %f\n", max_dx);
#endif
  max_dx = tri_hf_cells * mesh_delta.x * sqrt(max_dx);
#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  furthest distance (world XY) %f\n", max_dx);
#endif
  
  // Work out what the error distance is for this wedgie.
  epsilon = mesh_delta.z * error_bound;
#if DEBUG_MONOTONICITY
  if (print_error_metric)
    fprintf(debug_file, "  error_bound %d -> epsilon %f\n", error_bound, epsilon);
#endif


  // We now have all the ingredients we need to compute the priority. Note that the dist_bound from the
  // equation at the top is equal to distance on the z=1 plane. We move this to the near clip plane, and
  // work out its value as a proportion of camera_frustum_size (the X radius of the frustum). Finally convert the
  // error to pixels of distortion by multiplying by the X radius of the screen, viewresolution_x / 2.
  priority = ((viewresolution_x / 2) * camera_clip_near / camera_frustum_size)
    * epsilon * max_dx / (min_dxsquared + min_dzsquared);

  if (priority > PRIORITY_MAX) {
#if DEBUG_MONOTONICITY
    if (print_error_metric) {
      fprintf(debug_file, "  ... priority %f, exceeds PRIORITY_MAX, returning max\n", priority);
      fflush(debug_file);
    }
#endif
    // Will need to be recomputed immediately (if it isn't immediately split / merged).
    *dist_until = 0;
    return PRIORITY_MAX;
  } else {
#if DEBUG_MONOTONICITY
    if (print_error_metric) {
      fprintf(debug_file, "  ... priority %f\n", priority);
      fflush(debug_file);
    }
#endif

#if TEST_ERRORBOUND
    // Calculate the actual error metric as computed at the three corners of the triangle.
    dx_sq = mesh_delta_sq * (SQR(camera_hf_x - vcx) + SQR(camera_hf_y - vcy));
    dx = sqrt(dx_sq);

    dz = mesh_delta.z * fabs(vcz - camera_hf_z);
    dz_sq = SQR(dz);

    err_c = ((viewresolution_x / 2) * camera_clip_near / camera_frustum_size) * epsilon * dx / (dx_sq + dz_sq);

    fprintf(debug_file, "vc dx %f / (d^2 %f = dx^2 %f + dz^2 %f) -> %f\n",
            dx, dx_sq + dz_sq, dx_sq, dz_sq, err_c);



    dx_sq = mesh_delta_sq * (SQR(camera_hf_x - v0x) + SQR(camera_hf_y - v0y));
    dx = sqrt(dx_sq);

    dz = mesh_delta.z * fabs(v0z - camera_hf_z);
    dz_sq = SQR(dz);

    err_0 = ((viewresolution_x / 2) * camera_clip_near / camera_frustum_size) * epsilon * dx / (dx_sq + dz_sq);

    fprintf(debug_file, "v0 dx %f / (d^2 %f = dx^2 %f + dz^2 %f) -> %f\n",
            dx, dx_sq + dz_sq, dx_sq, dz_sq, err_0);




    dx_sq = mesh_delta_sq * (SQR(camera_hf_x - v1x) + SQR(camera_hf_y - v1y));
    dx = sqrt(dx_sq);

    dz = mesh_delta.z * fabs(v1z - camera_hf_z);
    dz_sq = SQR(dz);

    err_1 = ((viewresolution_x / 2) * camera_clip_near / camera_frustum_size) * epsilon * dx / (dx_sq + dz_sq);

    fprintf(debug_file, "v1 dx %f / (d^2 %f = dx^2 %f + dz^2 %f) -> %f\n",
            dx, dx_sq + dz_sq, dx_sq, dz_sq, err_1);




    fprintf(debug_file, "triangle %d %d %d [zvals: %d %d %d]\n", vc, v0, v1, vcz, v0z, v1z);
    fprintf(debug_file, "dx / |d| is %f. error is %f (%d units) at distance %f, scaled gives %f on viewplane\n",
            max_dx / sqrt(min_dxsquared + min_dzsquared), epsilon, error_bound, sqrt(min_dxsquared + min_dzsquared),
            epsilon * max_dx / (min_dxsquared + min_dzsquared));
    fprintf(debug_file, "this is %f at the near clip distance, x_frustum %f so = %f of X width (%d)\n",
            camera_clip_near * epsilon * max_dx / (min_dxsquared + min_dzsquared), camera_frustum_size,
            (camera_clip_near / camera_frustum_size) * epsilon * max_dx / (min_dxsquared + min_dzsquared), viewresolution_x / 2);
    fprintf(debug_file, "therefore pixel error = %f\n",
            ((viewresolution_x / 2) * camera_clip_near / camera_frustum_size) * epsilon * max_dx / (min_dxsquared + min_dzsquared));


    fprintf(debug_file, "bound dx %f / (d^2 %f = dx^2 %f + dz^2 %f) -> %f\n",
            max_dx, min_dxsquared + min_dzsquared, min_dxsquared, min_dzsquared, priority);
    fprintf(debug_file, "vertex errors %f %f %f | bound %f\n", err_c, err_0, err_1, priority);
#endif
    
    /*
     * Priority deferral computation
     */

    // We need to work out sort of how long we can safely leave this triangle / diamond
    // without recomputing its priority. A big assumption is involved here;
    // we assume that dz / d is roughly constant (the viewer is moving roughly
    // towards or away from the triangle).
    //
    // If this is true, then priority * d is also roughly constant. Thus the priority
    // increases proportional to 1/d... and we can work out the value of d that would
    // be required in order for the priority to hit the given value of threshold.

    current_min_d = sqrt(min_dxsquared + min_dzsquared);
    threshold_min_d = current_min_d * priority / threshold;

    if (dist_until_above_threshold)
      *dist_until = current_min_d - threshold_min_d;  // How far do we have to move towards the triangle until pri > thresh?
    else
      *dist_until = threshold_min_d - current_min_d;  // How far do we have to move away from the triangle until pri < thresh?

    if (*dist_until < 0) {
      // If we have already crossed the threshold (i.e. the current priority is on the trigger side of
      // threshold) then we will end up needing to be recomputed immediately.
      *dist_until = 0;
    } else if (*dist_until > DIST_MAX_DEFERRAL) {
      // Don't defer recomputation for more than some fixed distance... this stops
      // triangles being set to never recompute.
      *dist_until = DIST_MAX_DEFERRAL;
    }

    assert(*dist_until >= 0);

    return priority;
  }
}

// Internal routine used by perpendicular_epsilon_metric.
double Bintree::get_closest_barycentricsq(double camera_u, double camera_v)
{
  // Given barycentric coordinates u and v inside a triangle, finds the
  // closest point in the triangle to this pair of coordinates. Then
  // returns the SQUARE of the distance from the point to the triangle.

  if (camera_u < 0) {
    if (camera_v < 0) {
      // We are in the bottom-left corner; the closest point is vc.
      // Return the euclidean distance to this point (0, 0).
      return SQR(camera_u - 0) + SQR(camera_v - 0);
    } else if (camera_v < 1) {
      // We are along the left edge; the closest point is between vc and v1.
      // Return the parallel distance to the edge (u == 0).
      return SQR(camera_u);
    }
  } else if ((camera_v < 0) && (camera_u < 1)) {
    // We are along the bottom edge; the closest point is between vc and v0.
    // Return the parallel distance to the edge (v == 0).
    return SQR(camera_v);
  }

  // We have now cleared away the bottom-left, and both square sections on the bottom and the left.
  if (camera_u - camera_v > 1) {
    // We are in the bottom-right corner; the closest point is v0.
    // Return the euclidean distance to this point (1, 0).
    return SQR(camera_u - 1) + SQR(camera_v - 0);

  } else if (camera_u - camera_v < -1) {
    // We are in the top-left corner; the closest point is v1.
    // Return the euclidean distance to this point (0, 1).
    return SQR(camera_u - 0) + SQR(camera_v - 1);

  } else if (camera_u + camera_v < 1) {
    // We are inside the triangle. The closest point is right here.
    return 0;

  } else {
    // We are in the top-right edge; the closest point is between v0 and v1.
    // Return the parallel distance to the edge (u + v == 1).
    return SQR(camera_u + camera_v - 1) / 2;
  }
}

// Internal routine used by perpendicular_epsilon_metric. Returns 0 = v0, 1 = v1, 2 = vc
int Bintree::get_furthest_barycentric(double camera_u, double camera_v)
{
  if (camera_u > 0.5) {
    if (camera_v > 0.5) {
      // vc is the furthest vertex from us.
      return 2;
    } else {
      // v1 is the furthest vertex from us.
      return 1;
    }
  } else {
    if (camera_v > 0.5) {
      // v0 is the furthest vertex from us.
      return 0;
    } else {
      if (camera_u > camera_v) {
        // v1 is the furthest vertex from us.
        return 1;
      } else {
        // v0 is the furthest vertex from us.
        return 0;
      }
    }
  }
}
        
// Update the current camera position.
void Bintree::set_view(vec3 c_pos, quaternion c_dir)
{
  camera_pos = c_pos;
  camera_dir = c_dir;

  // Pre-calculate some parameters for transforming vertices into camera-space.
  if (error_metric == screenspace_errormetric) {
    camera_dir_inv = ~camera_dir;
    cspace_meshbase = mesh_base - camera_pos;
    cspace_xstep.zero(); cspace_ystep.zero(); cspace_zstep.zero();
    cspace_xstep.x = mesh_delta.x;
    cspace_ystep.y = mesh_delta.y;
    cspace_zstep.z = mesh_delta.z;
    
    cspace_meshbase = cspace_meshbase ^ camera_dir_inv;
    cspace_xstep = cspace_xstep ^ camera_dir_inv;
    cspace_ystep = cspace_ystep ^ camera_dir_inv;
    cspace_zstep = cspace_zstep ^ camera_dir_inv;
    
    // NB! These coordinates are now all in openGL camera-space, which has the screen in the XY plane
    // and NEGATIVE Z values in front of the camera. Flip the Z coordinates so that they are positive
    // when we come to calculate triangle Z values.
    cspace_meshbase.z *= -1; cspace_xstep.z *= -1; cspace_ystep.z *= -1; cspace_zstep.z *= -1;

  } else if (error_metric == perpendicular_epsilon_metric) {
    camera_hf_x = (camera_pos.x - mesh_base.x) / mesh_delta.x;
    camera_hf_y = (camera_pos.y - mesh_base.y) / mesh_delta.y;
    camera_hf_z = (camera_pos.z - mesh_base.z) / mesh_delta.z;
  }
}

// Recompute all priorities due to a view change.
void Bintree::recompute_priorities(double total_distance_moved, int force_recompute)
{
  int i;
  priq_info *next_element;
  double old_priority;

#if INCREMENTAL_RECOMPUTE
  if (!force_recompute) {
    double old_recompute_dist;
#if PRINT_RECOMPUTATION
    int num_tri_recomputed = 0, num_diamonds_recomputed = 0;

    fprintf(debug_file, "incremental_recomputation(%f) - current_distance_moved %f\n",
            total_distance_moved, current_distance_moved);
#endif

    if (total_distance_moved - current_distance_moved < DIST_TOLERANCE) {
      // We haven't moved. Don't bother with an incremental update.
      return;
    }

    current_distance_moved = total_distance_moved;

    while(1) {
      // Get the next triangle off the front of the triangle recomputation priority queue.
      next_element = tri_recompute_queue->get_element();
      if (next_element == NULL) {
#if PRINT_RECOMPUTATION
        fprintf(debug_file, "  no more triangles in tri_recompute_queue!\n");
#endif
        break;
      }

      // If this triangle doesn't need to be recomputed this frame (because we haven't moved far enough)
      // then don't do so.
      i = next_element->buf_index;
      assert((i >= 0) && (i < tribuf_size) && (tribuf[i].next_free == IN_USE));
#if PRINT_RECOMPUTATION
      fprintf(debug_file, "  tri %d (pri %f) current_dist %f recompute_dist %f...",
              i, tribuf[i].split_queue_info.priority, current_distance_moved,
              tribuf[i].recompute_queue_info.priority);
#endif
      if (total_distance_moved - next_element->priority < DIST_TOLERANCE) {
#if PRINT_RECOMPUTATION
        fprintf(debug_file, "  ...doesn't need recomputation (dist %f)\n", total_distance_moved);
#endif
        break;
      }

      // Calculate the new priority of the triangle.
      old_priority = tribuf[i].split_queue_info.priority;
      old_recompute_dist = next_element->priority;
      calculate_tri_priority(i);
#if PRINT_RECOMPUTATION
      num_tri_recomputed++;
      fprintf(debug_file, "  ...recomputed (pri %f), new dist %f\n",
              tribuf[i].split_queue_info.priority, tribuf[i].recompute_queue_info.priority);
#endif

      if (tribuf[i].recompute_queue_info.priority <= old_recompute_dist) {
        fprintf(debug_file, "INCREMENTAL RECOMPUTE: computation of tri %d moves computation distance back %f -> %f!\n",
                i, old_recompute_dist, tribuf[i].recompute_queue_info.priority);
        premature_shutdown();
      }

      // Move the triangle up or down the split priority queue accordingly.
      assert(tribuf[i].split_queue_info.priq_location != -1);
      if (tribuf[i].split_queue_info.priority > old_priority)
        tri_queue->bubble_up(&tribuf[i].split_queue_info);
      else
        tri_queue->bubble_down(&tribuf[i].split_queue_info);
    }

    while(1) {
      // Get the next diamond off the front of the diamond recomputation priority queue.
      next_element = diamond_recompute_queue->get_element();
      if (next_element == NULL) {
#if PRINT_RECOMPUTATION
        fprintf(debug_file, "  no more diamonds in diamond_recompute_queue!\n");
#endif
        break;
      }

      // If this diamond doesn't need to be recomputed this frame (because we haven't moved far enough)
      // then don't do so.
      i = next_element->buf_index;
      assert((i >= 0) && (i < diamondbuf_size) && (diamondbuf[i].next_free == IN_USE));
#if PRINT_RECOMPUTATION
      fprintf(debug_file, "  dia %d (pri %f) current_dist %f recompute_dist %f...",
              i, diamondbuf[i].merge_queue_info.priority, current_distance_moved,
              diamondbuf[i].recompute_queue_info.priority);
#endif
      if (total_distance_moved - next_element->priority < DIST_TOLERANCE) {
#if PRINT_RECOMPUTATION
        fprintf(debug_file, "  ...doesn't need recomputation (dist %f)\n", total_distance_moved);
#endif
        break;
      }

      // Calculate the new priority of the triangle.
      old_priority = diamondbuf[i].merge_queue_info.priority;
      old_recompute_dist = next_element->priority;
      calculate_diamond_priority(i);
#if PRINT_RECOMPUTATION
      num_diamonds_recomputed++;
      fprintf(debug_file, "  ...recomputed (pri %f), new dist %f\n",
              diamondbuf[i].merge_queue_info.priority, diamondbuf[i].recompute_queue_info.priority);
#endif

      if (diamondbuf[i].recompute_queue_info.priority <= old_recompute_dist) {
        fprintf(debug_file, "INCREMENTAL RECOMPUTE: computation of diamond %d moves computation distance back %f -> %f!\n",
                i, old_recompute_dist, diamondbuf[i].recompute_queue_info.priority);
        premature_shutdown();
      }

      // Move the diamond up or down the merge priority queue accordingly.
      assert(diamondbuf[i].merge_queue_info.priq_location != -1);
      if (diamondbuf[i].merge_queue_info.priority < old_priority)
        diamond_queue->bubble_up(&diamondbuf[i].merge_queue_info);
      else
        diamond_queue->bubble_down(&diamondbuf[i].merge_queue_info);
    }

#if PRINT_RECOMPUTATION
    fprintf(debug_file, "done recomputation, updated %d triangles and %d diamonds\n", num_tri_recomputed, num_diamonds_recomputed);
#endif
    return;
  }

#endif

  // Either incremental recomputation has been turned off by a compile switch, or
  // it has been disabled for this update. Recompute *every* triangle and diamond.

  current_distance_moved = total_distance_moved;

  for (i = 0; i < tribuf_size; i++) {
    if ((tribuf[i].next_free != IN_USE) || (tribuf[i].sub_level == 0))
      continue;

    // Calculate a new priority for this triangle, and bubble it up or down
    // the priority queue as a result.
    old_priority = tribuf[i].split_queue_info.priority;
    calculate_tri_priority(i);

    assert(tribuf[i].split_queue_info.priq_location != -1);
    if (tribuf[i].split_queue_info.priority > old_priority)
      tri_queue->bubble_up(&tribuf[i].split_queue_info);
    else
      tri_queue->bubble_down(&tribuf[i].split_queue_info);
  }

  for (i = 0; i < diamondbuf_size; i++) {
    if (diamondbuf[i].next_free != IN_USE)
      continue;

    // Calculate a new priority for this diamond, and bubble it up or down
    // the priority queue as a result..
    old_priority = diamondbuf[i].merge_queue_info.priority;
    calculate_diamond_priority(i);

    // Bubble it up or down the priority queue as a result.
    assert(diamondbuf[i].merge_queue_info.priq_location != -1);
    if (diamondbuf[i].merge_queue_info.priority < old_priority)
      diamond_queue->bubble_up(&diamondbuf[i].merge_queue_info);
    else
      diamond_queue->bubble_down(&diamondbuf[i].merge_queue_info);
  }
}

// Adapt the triangulation to the target priority, while keeping under the triangle limit.
void Bintree::adapt_triangulation(void)
{
#if PRINT_ADAPTATION
  int done_splits, done_merges;
#endif
  priq_info *biggest_split, *smallest_diamond;

#if PRINT_ADAPTATION
  done_splits = done_merges = 0;
#endif
  while (1) {

#if PRINT_ADAPTATION
    fprintf(debug_file, "current(%d) ", total_numtri);
#endif

    biggest_split = tri_queue->get_element();
    if (biggest_split == NULL)
      biggest_split_priority = PRIORITY_NEVER;
    else
      biggest_split_priority = biggest_split->priority;
#if PRINT_ADAPTATION
    fprintf(debug_file, "split(tri %6d, pri %f) ",
            (biggest_split == NULL) ? -1 : biggest_split->buf_index, biggest_split_priority);
#endif

    smallest_diamond = diamond_queue->get_element();
    if (smallest_diamond == NULL)
      smallest_diamond_priority = PRIORITY_MAX;   // high-priority diamonds are those that never get merged
    else
      smallest_diamond_priority = smallest_diamond->priority;
#if PRINT_ADAPTATION
    fprintf(debug_file, "merge(diamond %6d, pri %f) ",
            (smallest_diamond == NULL) ? -1 : smallest_diamond->buf_index, smallest_diamond_priority);
#endif

    // If we have room to grow by splitting and we are too inaccurate, then split a triangle.
    if ((total_numtri < tri_limit) && (biggest_split_priority > target_priority)) {
#if PRINT_ADAPTATION
      fprintf(debug_file, "splitting\n");
#endif
      split_triangle(biggest_split->buf_index);

#if PRINT_ADAPTATION
      done_splits++;
#endif
    } else {
      // We are either at the triangle limit, or we are accurate enough. Are there any triangles
      // that we can remove without them immediately being split?
      if (biggest_split_priority - smallest_diamond_priority > PRIORITY_TOLERANCE) {
#if PRINT_ADAPTATION
        fprintf(debug_file, "merging\n");
#endif
        merge_diamond(smallest_diamond->buf_index);

#if PRINT_ADAPTATION
        done_merges++;
#endif
      } else {
#if PRINT_ADAPTATION
        fprintf(debug_file, "stop (%d merge, %d split)\n", done_merges, done_splits);
#endif
        break;
      }
    }
  }
}

// Check that all the mergeable diamonds have their triangles' diamond_index fields
// set correctly.
void Bintree::assert_tri_diamond_indices(void)
{
  int i, t0, t1, t2, t3;

  for (i = 0; i < diamondbuf_size; i++) {
    if (diamondbuf[i].next_free != IN_USE)
      continue;

    t0 = diamondbuf[i].t0;
    t1 = diamondbuf[i].t1;
    t2 = diamondbuf[i].t2;
    t3 = diamondbuf[i].t3;

    assert((t0 >= 0) && (t0 < tribuf_size) && (tribuf[t0].next_free == IN_USE));
    assert(tribuf[t0].diamond_index == i);

    assert((t1 >= 0) && (t1 < tribuf_size) && (tribuf[t1].next_free == IN_USE));
    assert(tribuf[t1].diamond_index == i);

    if (t2 == -1) {
      assert(t3 == -1);
    } else {
      assert((t2 >= 0) && (t2 < tribuf_size) && (tribuf[t2].next_free == IN_USE));
      assert(tribuf[t2].diamond_index == i);
      
      assert((t3 >= 0) && (t3 < tribuf_size) && (tribuf[t3].next_free == IN_USE));
      assert(tribuf[t3].diamond_index == i);
    }
  }
}

// Set up the texture arrays
void Bintree::init_texture(int num_tex, int tex_dim)
{
  int temp;

  is_textured = TRUE;
  num_textures = num_tex;
  texture_dim = tex_dim;

  assert(num_textures == base_triangulation_size / 2);

  // Work out how many mip-mapping levels will be required.
  miplevels = 1;
  for (temp = 1; temp < texture_dim; temp <<= 1)
    miplevels++;
  if (temp != texture_dim) {
    fprintf(debug_file, "ERROR: mip-mapping requires power of 2 texture dimension (was %d)\n",
            texture_dim);
    premature_shutdown();
  }

  // Allocate the texture arrays.
  texturedata = (unsigned char **) malloc(num_textures * miplevels
                                          * sizeof(unsigned char *));
  if (!texturedata) mem_fail(num_textures * miplevels * sizeof(unsigned char *),
                             "Bintree texture-pointer array");

  texture_handle = (GLuint *) malloc(num_textures * sizeof(GLuint));
  if (!texture_handle) mem_fail(num_textures * sizeof(GLuint),
                                "Bintree texture-handle array");

  // Allocate the index arrays for finding the first and last triangle
  // in a texture chain.
  first_tex_tri = (int *) malloc(num_textures * sizeof(int));
  last_tex_tri = (int *) malloc(num_textures * sizeof(int));
  if (!first_tex_tri || !last_tex_tri) mem_fail(2 * num_textures * sizeof(int),
                                                "Bintree texture-tri-index arrays");
}

// Build the mipmap levels for a texture.
void Bintree::build_miplevels(int section_index, unsigned char *tex_data)
{
  if (use_borders) {
    /*
     * NOTA BENE!!!
     *
     * Border textures are the ideal way of solving the problems that occur
     * when you try to split a large texture up into several smaller sub-textures,
     * and use bilinear filtering. Because the texture can't filter from one
     * sub-texture to the other, there are visible discontinuities.
     *
     * Border textures basically place a small strip of the surrounding texture
     * around every sub-texture, to produce the correct result when filtering.
     * However quite a few consumer 3D chipsets do not support border texturing.
     * For example, my Viper V550 (Riva TNT-based) does not.
     *
     * So this is optional. It works much better, if it works at all... and
     * mostly it doesn't. :(
     */

    int level, old_levelsize, new_levelsize, x, y, i, left_section, below_section, belowleft_section;
    int *new_row, far_right, far_top;
    unsigned char *oldtex, *newtex, *oldptr, *newptr;
    
    assert((section_index >= 0) && (section_index < num_textures));
    texturedata[section_index * miplevels + 0] = tex_data;
    
    if ((section_index % x_sections) == 0)
      left_section = -1;
    else
      left_section = section_index - 1;
    
    if ((section_index / x_sections) == 0)
      below_section = -1;
    else
      below_section = section_index - x_sections;
    
    if ((left_section == -1) || (below_section == -1))
      belowleft_section = -1;
    else
      belowleft_section = section_index - x_sections - 1;
    
    far_right = ((section_index % x_sections) == (x_sections - 1));
    far_top   = ((section_index % y_sections) == (y_sections - 1));
    
    /*
     * NB: Every miplevel is in fact (texture_dim >> miplevel) + 2 across
     * because we must store a border of 1 pixel on every side in order to
     * perform linear filtering correctly from one section to another.
     */
    
    // Allocate space for the first miplevel, and copy the texture data across.
    texturedata[section_index * miplevels] = (unsigned char *) malloc(3 * SQR(2 + texture_dim));
    if (!texturedata[section_index * miplevels]) mem_fail(3 * SQR(2 + texture_dim),
                                                          "Bintree colormap texture");
    for (i = 0; i < texture_dim; i++)
      memcpy(&texturedata[section_index * miplevels][3 * (1 + (i + 1) * (2 + texture_dim))], &tex_data[3 * (i * texture_dim)], 3 * texture_dim);
    
    // Copy the neighbouring toplevel textures into our borders.
    copy_borders(section_index, left_section, below_section, belowleft_section, 0, texture_dim);
    extend_borders(section_index, 0, texture_dim,
      (left_section == -1), far_right, (below_section == -1), far_top);
    
    // Construct the miplevels.
    oldtex = texturedata[section_index * miplevels];
    old_levelsize = texture_dim;
    new_row = (int *) malloc(sizeof(int) * 3 * texture_dim / 2);
    if (!new_row) mem_fail(sizeof(int) * 3 * texture_dim / 2, "Bintree mipmap generation array");
    
    for (level = 0; level < miplevels - 1; level++) {
      
      // Allocate space for the next miplevel.
      new_levelsize = old_levelsize / 2;
      newtex = (unsigned char *) malloc(3 * SQR(2 + new_levelsize));
      if (!newtex) mem_fail(3 * SQR(2 + new_levelsize), "Bintree mipmap texture");
      texturedata[section_index * miplevels + level + 1] = newtex;
      
      // Set the pointers up to start at row 1 (row 0 is a border).
      newptr = &newtex[3*(1 * (2 + new_levelsize))];
      oldptr  = &oldtex[3*(1 * (2 + old_levelsize))];
      for (y = 0; y < old_levelsize; y += 2) {
        assert(oldptr - oldtex == 3 * ((y + 1) * (2 + old_levelsize)));
        
        // Skip the left-hand border pixel.
        oldptr += 3;
        
        // Run along an even y row in the image, setting the new miplevel's
        // row of pixels to be half the average of every pair of pixels.
        for (x = 0, i = 0; x < old_levelsize; x += 2) {
          assert(oldptr - oldtex == 3 * ((1 + x) + (y + 1) * (2 + old_levelsize)));
          assert(i == 3 * (x / 2));
          new_row[i++] = *oldptr++;
          new_row[i++] = *oldptr++;
          new_row[i++] = *oldptr++;
          
          // Go back a pixel in the new row and add in the contributions
          // from the next pixel.
          i -= 3;
          
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
        }
        
        // Skip the right-hand border pixel and the left-hand border pixel on the next row.
        oldptr += 6;
        
        // We need to run along the next row as well, adding in its contributions.
        for (x = 0, i = 0; x < old_levelsize; x += 2) {
          assert(oldptr - oldtex == 3 * ((1 + x) + (y + 2) * (2 + old_levelsize)));
          assert(i == 3 * (x / 2));
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          
          // Go back a pixel in the new row and add in the contributions
          // from the next pixel.
          i -= 3;
          
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
        }
        
        // Skip the right-hand border pixel.
        oldptr += 3;
        
        // Copy the information from the new row into the new texture. Remember to,
        // once again, skip the border pixels.
        assert(newptr - newtex == 3 * (y/2 + 1) * (2 + new_levelsize));
        newptr += 3;
        for (x = 0, i = 0; x < 3 * new_levelsize; x++)
          *newptr++ = (new_row[i++] + 2) / 4;
        newptr += 3;
      }
      
      // Copy the neighbouring mipmap textures into our borders.
      copy_borders(section_index, left_section, below_section, belowleft_section, level + 1, new_levelsize);
      extend_borders(section_index, level + 1, new_levelsize,
        (left_section == -1), far_right, (below_section == -1), far_top);
      
      // Set up for the next miplevel.
      old_levelsize = new_levelsize;
      oldtex = newtex;
    }
    
    free(new_row);

  } else {
    // Border textures are disabled.

    int level, old_levelsize, new_levelsize, x, y, i;
    int *new_row;
    unsigned char *oldtex, *newtex, *oldptr, *newptr;
    
    assert((section_index >= 0) && (section_index < num_textures));
    texturedata[section_index * miplevels + 0] = tex_data;
    
    // Allocate space for the first miplevel, and copy the texture data across.
    texturedata[section_index * miplevels] = (unsigned char *) malloc(3 * SQR(texture_dim));
    if (!texturedata[section_index * miplevels]) mem_fail(3 * SQR(texture_dim),
                                                          "Bintree colormap texture");
    memcpy(texturedata[section_index * miplevels], tex_data, 3 * SQR(texture_dim));
    
    // Construct the miplevels.
    oldtex = texturedata[section_index * miplevels];
    old_levelsize = texture_dim;
    new_row = (int *) malloc(sizeof(int) * 3 * texture_dim / 2);
    if (!new_row) mem_fail(sizeof(int) * 3 * texture_dim / 2, "Bintree mipmap generation array");
    
    for (level = 0; level < miplevels - 1; level++) {
      
      // Allocate space for the next miplevel.
      new_levelsize = old_levelsize / 2;
      newtex = (unsigned char *) malloc(3 * SQR(new_levelsize));
      if (!newtex) mem_fail(3 * SQR(new_levelsize), "Bintree mipmap texture");
      texturedata[section_index * miplevels + level + 1] = newtex;
      
      newptr = newtex;
      oldptr  = oldtex;
      for (y = 0; y < old_levelsize; y += 2) {
        assert(oldptr - oldtex == 3 * (y * old_levelsize));
        
        // Run along an even y row in the image, setting the new miplevel's
        // row of pixels to be half the average of every pair of pixels.
        for (x = 0, i = 0; x < old_levelsize; x += 2) {
          assert(oldptr - oldtex == 3 * (x + y * old_levelsize));
          assert(i == 3 * (x / 2));
          new_row[i++] = *oldptr++;
          new_row[i++] = *oldptr++;
          new_row[i++] = *oldptr++;
          
          // Go back a pixel in the new row and add in the contributions
          // from the next pixel.
          i -= 3;
          
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
        }
        
        // We need to run along the next row as well, adding in its contributions.
        for (x = 0, i = 0; x < old_levelsize; x += 2) {
          assert(oldptr - oldtex == 3 * (x + (y + 1) * old_levelsize));
          assert(i == 3 * (x / 2));
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          
          // Go back a pixel in the new row and add in the contributions
          // from the next pixel.
          i -= 3;
          
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
          new_row[i++] += *oldptr++;
        }
        
        // Copy the information from the new row into the new texture.
        assert(newptr - newtex == 3 * (y/2) * new_levelsize);
        for (x = 0, i = 0; x < 3 * new_levelsize; x++)
          *newptr++ = (new_row[i++] + 2) / 4;
      }
      
      // Set up for the next miplevel.
      old_levelsize = new_levelsize;
      oldtex = newtex;
    }
    
    free(new_row);
  }
}

// Copy the borders of a miplevel.
void Bintree::copy_borders(int section, int l_section, int b_section, int bl_section,
                           int level, int dim)
{
  unsigned char *s_array, *l_array, *b_array, *bl_array;

  s_array = texturedata[section * miplevels + level];

  // Copy the left border.
  if (l_section != -1) {
    l_array = texturedata[l_section * miplevels + level];

    // We copy our border from the section to the left, and vice-versa.
    copy_vertical_border(s_array + 3*((0      ) + 1 * (2 + dim)),
                         l_array + 3*((0 + dim) + 1 * (2 + dim)), dim, (2 + dim));
      
    copy_vertical_border(l_array + 3*((1 + dim) + 1 * (2 + dim)),
                         s_array + 3*((1      ) + 1 * (2 + dim)), dim, (2 + dim));
  }
    
  // Copy the below border.
  if (b_section != -1) {
    b_array = texturedata[b_section * miplevels + level];

    // We copy our border from the section below, and vice-versa.
    copy_horizontal_border(s_array + 3*(1 + (0      ) * (2 + dim)),
                           b_array + 3*(1 + (0 + dim) * (2 + dim)), dim);
      
    copy_horizontal_border(b_array + 3*(1 + (1 + dim) * (2 + dim)),
                           s_array + 3*(1 + (1      ) * (2 + dim)), dim);
  }
    
  // Copy the below-left pixel.
  if (bl_section != -1) {
    assert((b_section != -1) && (l_section != -1));
    bl_array = texturedata[bl_section * miplevels + level];

    // We copy our diagonal pixel from the section below and to the left, and vice-versa.
    copy_diagonal_pixel(s_array  + 3*((0      ) + (0      ) * (2 + dim)),
                        bl_array + 3*((0 + dim) + (0 + dim) * (2 + dim)));
      
    copy_diagonal_pixel(bl_array + 3*((1 + dim) + (1 + dim) * (2 + dim)),
                        s_array  + 3*((1      ) + (1      ) * (2 + dim)));    
      
    // Since we (as the above-right section abutting it) are responsible for this diagonal overlap area, we must
    // also copy pixels between our below and left sections.
    copy_diagonal_pixel(l_array + 3*((1 + dim) + (0      ) * (2 + dim)),
                        b_array + 3*((1      ) + (0 + dim) * (2 + dim)));
      
    copy_diagonal_pixel(b_array + 3*((0      ) + (1 + dim) * (2 + dim)),
                        l_array + 3*((0 + dim) + (0      ) * (2 + dim)));
  }
}

// Extend the borders of a miplevel (because there is no neighbour there to
// provide the values).
void Bintree::extend_borders(int section, int level, int dim,
                             int extend_left, int extend_right, int extend_bottom, int extend_top)
{
  unsigned char *s_array = texturedata[section * miplevels + level];

  // Extend left and right first, if needed.
  if (extend_left)
    copy_vertical_border(s_array + 3*(0 + 1 * (2 + dim)),
                         s_array + 3*(1 + 1 * (2 + dim)), dim, (2 + dim));
  
  if (extend_right)
    copy_vertical_border(s_array + 3*(dim + 1 + 1 * (2 + dim)),
                         s_array + 3*(dim     + 1 * (2 + dim)), dim, (2 + dim));

  // Extend up and down now, if needed. Note that since we've just extended left and right
  // we are guaranteed something immediately above/below the corners to extend.
  if (extend_bottom) {
    // Extend down. Updates diagonal pixels as required.
    copy_horizontal_border(s_array + 3*(0 * (2 + dim)),
                           s_array + 3*(1 * (2 + dim)), (2 + dim));
  } else {
    // We have to extend diagonal pixels manually, if they need to be extended.
    if (extend_left)
      copy_diagonal_pixel(s_array + 3*(          0 * (2 + dim)),
                          s_array + 3*(          1 * (2 + dim)));

    if (extend_right)
      copy_diagonal_pixel(s_array + 3*(dim + 1 + 0 * (2 + dim)),
                          s_array + 3*(dim + 1 + 1 * (2 + dim)));
  }
  
  if (extend_top) {
    // Extend up. Updates diagonal pixels as required.
    copy_horizontal_border(s_array + 3*((dim + 1) * (2 + dim)),
                           s_array + 3*((dim    ) * (2 + dim)), (2 + dim));
  } else {
    // We have to extend diagonal pixels manually, if they need to be extended.
    if (extend_left)
      copy_diagonal_pixel(s_array + 3*(          (dim + 1) * (2 + dim)),
                          s_array + 3*(          (dim    ) * (2 + dim)));

    if (extend_right)
      copy_diagonal_pixel(s_array + 3*(dim + 1 + (dim + 1) * (2 + dim)),
                          s_array + 3*(dim + 1 + (dim    ) * (2 + dim)));
  }
}

// Copy a vertical mipmap border.
void Bintree::copy_vertical_border(unsigned char *dest, unsigned char *src, int num_elems, int stride)
{
  for (int i = 0; i < num_elems; i++) {
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;

    src  += 3 * (stride - 1);
    dest += 3 * (stride - 1);
  }
}

// All of the textures have been stored - do any processing necessary.
void Bintree::textures_done(void)
{
  int i, j;
  GLfloat object_plane[2][4];

  // Allocate texture handles.
  glGenTextures(num_textures, texture_handle);

  glEnable(GL_TEXTURE_2D);

  if (use_texgen) {
    // We calculate the texture coordinates automatically from a
    // vertex's position by dividing by the width of the texture.
    object_plane[0][0] = (GLfloat) (1.0 / (32 * mesh_delta.x));
    object_plane[0][1] = 0;
    object_plane[0][2] = 0;
    object_plane[0][3] = 0;
    glEnable(GL_TEXTURE_GEN_S);
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGenfv(GL_S, GL_OBJECT_PLANE, object_plane[0]);

    object_plane[1][0] = 0;
    object_plane[1][1] = (GLfloat) (1.0 / (32 * mesh_delta.y));
    object_plane[1][2] = 0;
    object_plane[1][3] = 0;
    glEnable(GL_TEXTURE_GEN_T);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGenfv(GL_T, GL_OBJECT_PLANE, object_plane[1]);
  }

  // Set up each texture and copy it to video memory.
  // FIXME: at present we try to load every texture into memory...
  // ... this could be better.
  for (i = 0; i < num_textures; i++) {
    glBindTexture(GL_TEXTURE_2D, texture_handle[i]);

    // The texture is clamped at its edges..
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // Trilinear filtering, using mipmapping.
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

    // Store the texture's miplevels.
    if (use_borders) {
      for (j = 0; j < miplevels; j++) {
        glTexImage2D(GL_TEXTURE_2D, j, GL_RGB,
                     2 + (texture_dim >> j), 2 + (texture_dim >> j), 1,
                     GL_RGB, GL_UNSIGNED_BYTE, texturedata[i * miplevels + j]);
      }
    } else {
      for (j = 0; j < miplevels; j++) {
        glTexImage2D(GL_TEXTURE_2D, j, GL_RGB,
                     texture_dim >> j, texture_dim >> j, 0,
                     GL_RGB, GL_UNSIGNED_BYTE, texturedata[i * miplevels + j]);
      }
    }
  }
}

// Enable/disable texturing (just a rendering flag, doesn't change
// the fact that textures are loaded).
void Bintree::set_texturing(int use_texturing)
{
  this->use_texturing = use_texturing;

  if (use_texturing)
    glEnable(GL_TEXTURE_2D);
  else
    glDisable(GL_TEXTURE_2D);
}


// Set up the scenery-type and scenery-object arrays.
void Bintree::alloc_scenery(int numtypes, int num_obj)
{
  int i;

  scenerytypearraysize = numtypes;
  sceneryarraysize = num_obj;

  scenerytypearray = (scenerytype *) malloc(scenerytypearraysize * sizeof(scenerytype));
  sceneryarray     = (sceneryobj  *) malloc(sceneryarraysize     * sizeof(sceneryobj));
  if (!scenerytypearray) mem_fail(scenerytypearraysize * sizeof(scenerytype),
                                  "Bintree scenery-type array");
  if (!sceneryarray)     mem_fail(sceneryarraysize * sizeof(sceneryobj),
                                  "Bintree scenery-item array");
  num_scenery_types = numtypes;
  num_scenery_obj = 0;

  // Initialise the scenery types.
  for (i = 0; i < num_scenery_types; i++) {
    scenerytypearray[i].first_sceneryobj = -1;
    scenerytypearray[i].last_sceneryobj  = -1;

    glGenTextures(1, &scenerytypearray[i].texture);
  }
}

// Create a scenery object.
void Bintree::create_scenery_object(int scenery_type, vec3 location)
{
  int x, y, obj_index;

  assert((scenery_type >= 0) && (scenery_type < num_scenery_types));
  assert((num_scenery_obj >= 0) && (num_scenery_obj < sceneryarraysize));
  obj_index = num_scenery_obj++;

  // Link up the previously last object to us.
  sceneryarray[obj_index].prev_obj = scenerytypearray[scenery_type].last_sceneryobj;
  if (sceneryarray[obj_index].prev_obj == -1)
    scenerytypearray[scenery_type].first_sceneryobj = obj_index;
  else
    sceneryarray[sceneryarray[obj_index].prev_obj].next_obj = obj_index;

  // Add this object at the end of the scenery chain.
  sceneryarray[obj_index].next_obj = -1;
  scenerytypearray[scenery_type].last_sceneryobj = obj_index;

  // Set up the scenery object.
  sceneryarray[obj_index].location = location;
  
  sceneryarray[obj_index].world_location.x = mesh_base.x + mesh_delta.x * location.x;
  sceneryarray[obj_index].world_location.y = mesh_base.y + mesh_delta.y * location.y;
  sceneryarray[obj_index].world_location.z = mesh_base.z + mesh_delta.z * location.z;


  // Work out which frustum_index we use.
  x = (int) (location.x / 32);
  y = (int) (location.y / 32);

  if ((x < 0) || (x >= x_sections) || (y < 0) || (y >= y_sections)) {
    fprintf(debug_file, "WARNING: scenery item off mesh, section %d %d (loc %f %f)\n",
            x, y, location.x, location.y);
    sceneryarray[obj_index].frustum_index = 0;
    return;
  } else {
    sceneryarray[obj_index].frustum_index = (x + 1) + (y + 1) * (x_sections + 2);
  }
}


// Set up a scenery texture (also initialises object size).
void Bintree::scenery_texture(int scenery_type, short scale_factor, int w, int h,
                              unsigned char *rgba_texture)
{
  int scenery_miplevels, level, old_w, old_h, new_w, new_h, x, y, i;
  int *new_row, *mip_dim;
  unsigned char *oldtex, *newtex, *oldptr, *newptr;
  unsigned char **miptex;

  // We say 32 pixels to the mesh cell. Since there are 512 height
  // steps to the mesh cell in Z, this is one pixel = 16 height steps.
  //
  // We also apply the object's scale factor... 255 = no change.
  scenerytypearray[scenery_type].x_radius = (scale_factor / 255.0) * (w / 64.0);
  scenerytypearray[scenery_type].y_height = (scale_factor / 255.0) * (h * 16);
  glBindTexture(GL_TEXTURE_2D, scenerytypearray[scenery_type].texture);

  // The texture is clamped at its edges..
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

  // Trilinear filtering, using mipmapping.
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

  // Compute the number of miplevels required.
  new_w = w; new_h = h;
  scenery_miplevels = 0;
  while (1) {
    scenery_miplevels++;
    if ((new_w == 1) && (new_h == 1))
      break;

    if (new_w > 1)
      new_w >>= 1;

    if (new_h > 1)
      new_h >>= 1;
  }

  // Construct the miplevels. This is a little tricky because of the way that alpha
  // works, but it's possible. :)
  mip_dim = (int *) malloc(2 * scenery_miplevels * sizeof(int));
  if (!mip_dim) mem_fail(2 * scenery_miplevels * sizeof(int),
                         "Bintree scenery-texture mipmap-dimension array");

  miptex = (unsigned char **) malloc(scenery_miplevels * sizeof(unsigned char *));
  if (!miptex) mem_fail(scenery_miplevels * sizeof(unsigned char *),
                        "Bintree scenery-texture mipmap-pointer array");

  oldtex = miptex[0] = rgba_texture;
  old_w = mip_dim[2*0 + 0] = w;
  old_h = mip_dim[2*0 + 1] = h;
  new_row = (int *) malloc(sizeof(int) * 4 * w / 2);
  if (!new_row) mem_fail(sizeof(int) * 4 * w / 2,
                         "Bintree scenery-texture new-mipmap-row array");
    
  for (level = 0; level < scenery_miplevels - 1; level++) {
    // Determine the size of the next miplevel.
    if (old_w > 1)
      new_w = old_w / 2;
    else
      new_w = old_w;

    if (old_h > 1)
      new_h = old_h / 2;
    else
      new_h = old_h;
    mip_dim[2 * (level + 1) + 0] = new_w;
    mip_dim[2 * (level + 1) + 1] = new_h;

    // Allocate space for the next miplevel.
    newtex = (unsigned char *) malloc(4 * new_w * new_h);
    if (!newtex) mem_fail(4 * new_w * new_h,
                          "Bintree scenery-texture mipmap texture");
    miptex[level + 1] = newtex;
      
    newptr = newtex;
    oldptr = oldtex;
    for (y = 0; y < old_h; y += 2) {
      assert(oldptr - oldtex == 4 * (y * old_w));
        
      // Run along an even y row in the image, setting the new miplevel's
      // row of pixels to be half the average of every pair of pixels.
      if (old_w == 1) {
        // There isn't a pair of pixels, just one.
        assert(oldptr - oldtex == 4 * (0 + y * old_w));
        new_row[0] = 2 * oldptr[0] * (255 - oldptr[3]);
        new_row[1] = 2 * oldptr[1] * (255 - oldptr[3]);
        new_row[2] = 2 * oldptr[2] * (255 - oldptr[3]);
        new_row[3] = 2 * (255 - oldptr[3]);
        oldptr += 4;
      } else {
        for (x = 0, i = 0; x < old_w; x += 2) {
          assert(oldptr - oldtex == 4 * (x + y * old_w));
          assert(i == 4 * (x / 2));
          new_row[i + 0] = 2 * oldptr[0] * (255 - oldptr[3]);
          new_row[i + 1] = 2 * oldptr[1] * (255 - oldptr[3]);
          new_row[i + 2] = 2 * oldptr[2] * (255 - oldptr[3]);
          new_row[i + 3] = 2 * (255 - oldptr[3]);
          oldptr += 4;
          
          // Add in the contributions from the next pixel to the
          // same pixel in the new row.
          new_row[i + 0] += 2 * oldptr[0] * (255 - oldptr[3]);
          new_row[i + 1] += 2 * oldptr[1] * (255 - oldptr[3]);
          new_row[i + 2] += 2 * oldptr[2] * (255 - oldptr[3]);
          new_row[i + 3] += 2 * (255 - oldptr[3]);
          oldptr += 4;
          i += 4;
        }
      }
        
      if (old_h == 1) {
        // There is no next row. Just use this row's contributions.
        assert(newptr - newtex == 4 * (y/2) * new_w);
        for (x = 0, i = 0; x < new_w; x++) {
          if (new_row[i + 3] == 0) {
            newptr[0] = 0;
            newptr[1] = 0;
            newptr[2] = 0;
            newptr[3] = 255;
          } else {
            newptr[0] = new_row[i + 0] / new_row[i + 3];
            newptr[1] = new_row[i + 1] / new_row[i + 3];
            newptr[2] = new_row[i + 2] / new_row[i + 3];
            newptr[3] = 255 - new_row[i + 3] / 2;
          }
          newptr += 4;
          i += 4;
        }

        continue;
      }

      // We need to run along the next row as well, adding in its contributions.
      if (old_w == 1) {
        // There isn't a pair of pixels, just one.
        assert(oldptr - oldtex == 4 * (0 + (y + 1) * old_w));
        new_row[0] = 2 * oldptr[0] * (255 - oldptr[3]);
        new_row[1] = 2 * oldptr[1] * (255 - oldptr[3]);
        new_row[2] = 2 * oldptr[2] * (255 - oldptr[3]);
        new_row[3] = 2 * (255 - oldptr[3]);
        oldptr += 4;
      } else {
        for (x = 0, i = 0; x < old_w; x += 2) {
          assert(oldptr - oldtex == 4 * (x + (y + 1) * old_w));
          assert(i == 4 * (x / 2));
          new_row[i + 0] = 2 * oldptr[0] * (255 - oldptr[3]);
          new_row[i + 1] = 2 * oldptr[1] * (255 - oldptr[3]);
          new_row[i + 2] = 2 * oldptr[2] * (255 - oldptr[3]);
          new_row[i + 3] = 2 * (255 - oldptr[3]);
          oldptr += 4;
          
          // Add in the contributions from the next pixel to the
          // same pixel in the new row.
          new_row[i + 0] += 2 * oldptr[0] * (255 - oldptr[3]);
          new_row[i + 1] += 2 * oldptr[1] * (255 - oldptr[3]);
          new_row[i + 2] += 2 * oldptr[2] * (255 - oldptr[3]);
          new_row[i + 3] += 2 * (255 - oldptr[3]);
          oldptr += 4;
          i += 4;
        }
      }
      
      // Copy the information from the new row into the new texture.
      assert(newptr - newtex == 4 * (y/2) * new_w);
      for (x = 0, i = 0; x < new_w; x++) {
        if (new_row[i + 3] == 0) {
          newptr[0] = 0;
          newptr[1] = 0;
          newptr[2] = 0;
          newptr[3] = 255;
        } else {
          newptr[0] = new_row[i + 0] / new_row[i + 3];
          newptr[1] = new_row[i + 1] / new_row[i + 3];
          newptr[2] = new_row[i + 2] / new_row[i + 3];
          newptr[3] = 255 - new_row[i + 3] / 4;
        }
        newptr += 4;
        i += 4;
      }
    }
    
    // Set up for the next miplevel.
    old_w = new_w;
    old_h = new_h;
    oldtex = newtex;
  }

  // Perform a post-processing operation on the mipmap; for any
  // pixel that is totally transparent, set its RGB channel to be equal
  // to that of one of its neighbours. Helps stop bilinear interpolation
  // from darkening edge pixels too much.
  //
  // We do this once in X and once in Y, to get all of the diagonal neighbours too.
  for (level = 0; level < scenery_miplevels; level++) {
    new_w = mip_dim[2*level + 0];
    new_h = mip_dim[2*level + 1];

    // Perform the copy in X.
    newptr = miptex[level];
    for (y = 0; y < new_h; y++) {
      for (x = 0; x < new_w; x++) {
        assert(newptr - miptex[level] == 4 * (x + y * new_w));

        if (newptr[3] == 0xFF) {
          if ((x > 0) && (newptr[3 - 4] < 0xFF)) {
            // Copy the left-hand pixel's RGB values across.
            newptr[0] = newptr[0 - 4];
            newptr[1] = newptr[1 - 4];
            newptr[2] = newptr[2 - 4];
          } else if ((x < new_w - 1) && (newptr[3 + 4] < 0xFF)) {
            // Copy the right-hand pixel's RGB values across.
            newptr[0] = newptr[0 + 4];
            newptr[1] = newptr[1 + 4];
            newptr[2] = newptr[2 + 4];
          }
        }
        newptr += 4;
      }
    }

    // Perform the copy in Y.
    newptr = miptex[level];
    for (y = 0; y < new_h; y++) {
      for (x = 0; x < new_w; x++) {
        assert(newptr - miptex[level] == 4 * (x + y * new_w));

        if (newptr[3] == 0xFF) {
          if ((y > 0) && (newptr[3 - 4 * new_w] < 0xFF)) {
            // Copy the left-hand pixel's RGB values across.
            newptr[0] = newptr[0 - 4 * new_w];
            newptr[1] = newptr[1 - 4 * new_w];
            newptr[2] = newptr[2 - 4 * new_w];
          } else if ((y < new_h - 1) && (newptr[3 + 4 * new_w] < 0xFF)) {
            // Copy the right-hand pixel's RGB values across.
            newptr[0] = newptr[0 + 4 * new_w];
            newptr[1] = newptr[1 + 4 * new_w];
            newptr[2] = newptr[2 + 4 * new_w];
          }
        }
        newptr += 4;
      }
    }
  }

  for (i = 0; i < scenery_miplevels; i++) {
    assert((mip_dim[2*i + 0] > 0) && (mip_dim[2*i + 1] > 0));

    glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA, mip_dim[2*i + 0], mip_dim[2*i + 1], 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, miptex[i]);
    free(miptex[i]);
  }
  free(miptex);

  free(new_row);
}

// Morph all vertices to the current time.
void Bintree::do_morph(long current_time)
{
  morphvertex *mptr;
  int m_index, next_index;
  long time_until_completion;
  double interp, cube_val;

//  fprintf(debug_file, "MORPHING %d vertices from time %d -> time %d\n",
//          total_morphs, morphed_time, current_time);

  for (m_index = first_morph; m_index != -1; m_index = next_index) {
    assert((m_index > 0) && (m_index < morphbuf_size));
    mptr = morphbuf + m_index;
    assert(mptr->next_free == IN_USE);

    next_index = mptr->next_morph;

    // Work out how long until completion.
    time_until_completion = mptr->finish_time - current_time;

    // Are we finished?
    if (time_until_completion > 0) {
      // Work out our new location using a cubic interpolation filter.
      interp = ((double) time_until_completion) / morph_msec;
      cube_val = 1 - interp;
#if CUBIC_MORPH
      cube_val *= SQR(cube_val);
#endif
      mptr->current_height = mptr->start_height +
          (unsigned short) (mptr->delta_height * cube_val);

//      fprintf(debug_file, "%d: time until completion %d -> interp %f -> cube_val %f -> lerp %d (%d) -> %d\n",
//              m_index, time_until_completion, interp, cube_val, mptr->start_height,
//              mptr->delta_height, mptr->current_height);
    } else {
      // Tell our vertex that it has reached its desired location.
      assert(vertex_array[2 * mptr->vertex_index + 0] == mptr->start_height + mptr->delta_height);
      assert(vertex_array[2 * mptr->vertex_index + 1] == m_index);
      vertex_array[2 * mptr->vertex_index + 1] = 0;

//      fprintf(debug_file, "%d: time until completion %d -> finish\n",
//              m_index, time_until_completion);

      // Delete this entry in the morphlist.
      free_morph(m_index);
    }
  }

  morphed_time = current_time;
//  fflush(debug_file);
}

// Get the current height of a vertex (taking into account any morph)
unsigned short Bintree::get_height(int vertex_index)
{
  if (morph_vertices) {
    int m_index = vertex_array[2 * vertex_index + 1];

    if (m_index != 0) {
      assert((m_index > 0) && (m_index < morphbuf_size));
      assert(morphbuf[m_index].next_free == IN_USE);
      return morphbuf[m_index].current_height;
    }
  }

  return vertex_array[2 * vertex_index + 0];
}
