// Bintree.h : Definitions for class Bintree.
//

#define HEIGHT_MAX  65535

#define PRIORITY_NEVER  -1.0e9
#define PRIORITY_MAX    +1.0e9

#define PRIORITY_TOLERANCE 1.0e-02

#define DIST_TOLERANCE     1.0e-02
#define DIST_MAX_DEFERRAL  1.0e+03
#define DIST_INFINITY     +1.0e9

#define RECOMPUTE_SAFETY_FACTOR 0.8

#define FRUSTUMCLIPPING_SAFETY  4.0

// Sentinel values for freelists.
#define NONE_FREE   -1
#define IN_USE      -2

#define MAX_SUBDIVISION_LEVEL   32


const int morphbuf_growsize   = 200;

typedef unsigned short vertexdata;

// Precalculated information for every possible triangle in the bintree.
typedef struct precalctri_ {
  unsigned short error_bound, minval, maxval;
  vec3 tri_color;
} precalctri;

// A structure used for recursively precalculating the error_bound.
typedef struct interpstruct_ {
  double i_zc, i_z0, i_z1;
  double error_bound;
} interpstruct;

// Types of Bintrees.
enum bintreetype {BINTREE_RAWDATA, BINTREE_MYTHMESH};

// A triangle in the bintree.
typedef struct tri_ {
  // IN_USE for a triangle currently in use, otherwise the index
  // of the next free triangle. NONE_FREE indicates that there are no more
  // free triangles after this one in the freelist.
  int next_free;

  // The three vertices of the triangle.
  int vc, v0, v1;

  // The triangle's three neighbours. Indices into tribuf.
  int n0, n1, nb;

  // The level of the triangle - 0 is the most finely subdivided.
  int sub_level;

  // The index of the triangle within its level - used to access precalctri
  // along with sub_level by the formula
  // index = precalc_array_size - (precalc_array_size >> sub_level) + level_index
  int level_index;

  // If this triangle forms part of a mergeable diamond, this is its index. If not,
  // this holds -1.
  int diamond_index;

  // The texture index of this triangle, and two other indices to the previous
  // and next triangle sharing the same texture index.
  int texture_index, prev_tex, next_tex;

  // The triangle's information for the split priority queue.
  //
  // priority = a measure of the error in the triangulation caused by
  // this triangle's approximation.
  priq_info split_queue_info;

  // The triangle's information for the recomputation priority queue.
  //
  // priority = the value that total_distance_moved must reach before
  // this triangle's priority must be recomputed.
  priq_info recompute_queue_info;
} tri;

// A mergeable diamond in the bintree (four child triangles that may be merged by
// removing a vertex to form two parent triangles).
typedef struct diamond_ {
  // as for tri::next_free
  int next_free;

  // The four child triangles; t0 and t1 are children of the same parent, as are t2 and t3
  int t0, t1, t2, t3;

  // The priorities that would be assigned to this diamond's two child triangles
  // if it was merged.
  double parent0_priority, parent1_priority;

  // The diamond's information for the merge priority queue.
  //
  // priority = a measure of the error in the triangulation that would be caused by
  // merging this diamond into two rather than four triangles.
  priq_info merge_queue_info;

  // The triangle's information for the recomputation priority queue.
  //
  // priority = the value that total_distance_moved must reach before
  // this diamond's priority must be recomputed.
  priq_info recompute_queue_info;
} diamond;

// A type of scenery to be placed on the mesh.
typedef struct scenerytype_ {
  double x_radius, y_height;               // in terms of mesh points
  GLuint texture;                          // texture to place on the scenery object

  int first_sceneryobj, last_sceneryobj;   // indices into sceneryarray
} scenerytype;

// A scenery item to be placed on the mesh.
typedef struct sceneryobj_ {
  int frustum_index;                       // index into frustumflags that tells us whether to draw
  vec3 location;                           // location - x and y in mesh cells, z in units
  vec3 world_location;                     // world location - for billboarding
  int next_obj, prev_obj;                  // indices to prev and next scenery obj of this type
} sceneryobj;

// A vertex that is being morphed. We morph newly-created vertices
// to their new position over some number of milliseconds. (Merged vertices
// are still immediately snapped)
typedef struct morphvertex_ {
  int next_free;
  int vertex_index;
  unsigned short start_height, current_height;
  short delta_height;
  long finish_time;
  short next_morph, prev_morph;
} morphvertex;

class Bintree {
public:
  // Resolution of the bintree... from 0 to infinity
  int res;
  
  // The number of triangles in the base triangulation.
  int base_triangulation_size;

  // Maximum number of splits of a top-level triangle
  int max_subdiv_level;

  // Side length of the height data array in X and Y
  int hf_size_x, hf_size_y;

  // The triangulation's target accuracies.
  double target_priority;
  int tri_limit;

  // The vertex data array.
  int vertex_array_size;
  vertexdata *vertex_array;

  // The precalculated-triangle-data array.
  int precalc_array_size, *precalc_level_index;
  precalctri *precalc_array;

  // The texture for the Bintree. 24-bit arrays, one set of mipmaps per section of the base
  // triangulation.
  int is_textured, num_textures, texture_dim, miplevels, use_borders, use_texturing;
  unsigned char **texturedata;
  GLuint *texture_handle;

  // Frustum in/out flags for every section corner.
  unsigned char *frustumflags;

  // Objects to be placed on the mesh.
  int num_scenery_types, num_scenery_obj;
  int scenerytypearraysize, sceneryarraysize;
  scenerytype *scenerytypearray;
  sceneryobj *sceneryarray;

  // The mesh's base (zero corner) and extent in X, Y and Z.
  // mesh_delta is the X / Y / Z step for one increment along the heightfield or up.
  // mesh_delta_sq = SQR(mesh_delta.x).
  vec3 mesh_base, mesh_extent, mesh_delta;
  double mesh_delta_sq;

  // The triangles in the current triangulation.
  int total_numtri, next_free_tri, tribuf_size;
  tri *tribuf;

  // Indices to the first and last triangles with any given texture index.
  int *first_tex_tri, *last_tex_tri;

  // The mergeable diamonds in the current triangulation.
  int total_diamonds, next_free_diamond, diamondbuf_size;
  diamond *diamondbuf;

  // The vertices being morphed at the current time.
  long morphed_time;
  int total_morphs, morphbuf_size, next_free_morph;
  int first_morph, last_morph;
  morphvertex *morphbuf;

  // Priority queues for splittable triangles and mergeable diamonds,
  // and the current priorities at the heads of each.
  int threshold_priorities_set, lock_thresholds;
  double threshold_lockval, biggest_split_priority, smallest_diamond_priority;
  class PriQueue *tri_queue, *diamond_queue;

  // Priority queues for recomputing the priority of triangles and diamonds,
  // and the current distance-moved accumulator.
  double current_distance_moved;
  class PriQueue *tri_recompute_queue, *diamond_recompute_queue;

  // The current camera position and viewing parameters.
  vec3 camera_pos;
  quaternion camera_dir;

  // The error metric that we're using.
  double (Bintree::* error_metric)(int vc, int v0, int v1, int sub_level, int level_index,
                                   int dist_until_above_threshold, double threshold, double *dist_until);

  // Precalculated values from the camera's position... used for screenspace_errormetric
  quaternion camera_dir_inv;
  vec3 cspace_meshbase, cspace_xstep, cspace_ystep, cspace_zstep;

  // Precalculated values from the camera's position... used for perpendicular_epsilon_metric
  double camera_hf_x, camera_hf_y, camera_hf_z;

  // Parameters that are needed to get back to the base triangulation (basically
  // our defining parameters).
  bintreetype generating_type;
  class MeshTag *mesh;  // generating_type == BINTREE_MYTHMESH
  int x_sections, y_sections;
  
  // Construct a new Bintree from height data.
  Bintree(int res, unsigned short *height_data, vec3 base, vec3 extent);

  // Construct a new Bintree from a Myth II mesh tag.
  Bintree(class MeshTag *mesh);

  // Standard Bintree initialisation (called by both constructors).
  void standard_init(void);

  // Draw the Bintree.
  void draw(void);

  // Delete the Bintree.
  ~Bintree(void);

  /*
   * Pre-calculation routines
   */

  // Perform a bottom-up precalculation on the bintree's height data.
  void precalculate_data(void);

  // Recursively precalculate data about the bintree's possible triangles.
  int recursive_precalculate(int tri_level, int level_index, interpstruct *interp_data,
                             int vc, int v0, int v1, unsigned short *minval, unsigned short *maxval);

  // Get the precalc-array index for a triangle.
  inline int precalc_index(int tri_level, int level_index) {
    return precalc_level_index[tri_level] + level_index;
  }

  // Get the worldspace coordinates of a vertex.
  vec3 vertex_world(int v);



  /*
   * Split routines
   */

  // Split a triangle.
  void split_triangle(int t);

  // Internal routine for split_triangle; performs a single-triangle split and updates
  // neighbour information.
  void perform_split(int vb, int d_index, int t, int t_child, int base_t, int base_t_child);

  // Make a new mergeable diamond.
  void make_mergeable_diamond(int t0, int t1, int t2, int t3);

  // Break a mergeable diamond (because one of the triangles
  // concerned is being split).
  void break_diamond(int diamond_index);

  /*
   * Merge routines
   */

  // Merge a diamond.
  void merge_diamond(int d);

  // Internal routine for merge_diamond; merges a pair of triangles, updates
  // neighbour information and checks for newly mergeable diamonds.
  void merge_triangles(int neighbour_base, int child0, int child1);



  /*
   * Priority-queue management
   */

  // Revert the Bintree to its base triangulation.
  void base_triangulation(void);

  // Callbacks for the triangle priority queue.
  int tri_priority_compare(int tri1, int tri2);
  void tri_location_change(int tri_index, int new_location);

  // Callbacks for the triangle priority recomputation priority queue.
  int tri_recompute_compare(int tri1, int tri2);
  void tri_recompute_change(int tri_index, int new_location);

  // Callbacks for the diamond priority queue.
  int diamond_priority_compare(int diamond1, int diamond2);
  void diamond_location_change(int diamond_index, int new_location);

  // Callbacks for the diamond priority recomputation priority queue.
  int diamond_recompute_compare(int diamond1, int diamond2);
  void diamond_recompute_change(int diamond_index, int new_location);

  /*
   * Memory management
   */

  // Get a new triangle index in the triangle buffer.
  int new_triangle(void);

  // Free a triangle in the triangle buffer.
  void free_triangle(int tri_index);

  // Get a new diamond index in the diamond buffer.
  int new_diamond(void);

  // Free a diamond in the diamond buffer.
  void free_diamond(int diamond_index);

  // Get a new morph index in the morph buffer.
  int new_morph(void);

  // Free a morph in the morph buffer.
  void free_morph(int morph_index);

  /*
   * Error-metric calculation
   */

  // Update the current view parameters.
  void set_view(vec3 camera_pos, quaternion camera_dir);

  // Set up the split priority of a triangle. Takes a newly created triangle with
  // vertices etc. set up, calculates its priority and inserts it into both
  // tri_queue and the triangle recomputation queue.
  void setup_triangle_priority(int tri_index);
  
  // Calculate the split priority of a triangle, and its distance before it must
  // be recomputed. Update the triangle's position in the priority recomputation priority queue.
  // DO NOT insert the triangle into the split priority queue... other code must do that.
  //
  // Called from: setup_tri_priority   recompute_priorities
  void calculate_tri_priority(int tri_index, int just_do_calculation = FALSE);
  
  // Set up the merge priority of a diamond. Takes a newly created diamond with
  // t0, t1, t2, t3 set up, calculates its priority and inserts it into
  // diamond_queue and the diamond recomputation queue.
  void setup_diamond_priority(int d);
  
  // Calculate the merge priority of a diamond, and its distance before it must
  // be recomputed. Update the diamond's position in the priority recomputation priority queue.
  // DO NOT insert the diamond into the merge priority queue... other code must do that.
  //
  // Called from: setup_diamond_priority   recompute_priorities
  void calculate_diamond_priority(int d);

  // Calculate the split priority that would be found if two children
  // were merged.
  double calculate_diamondside_priority(int child0, int child1, double *move_dist);

  // Error metrics.
  double debug_metric(int vc, int v0, int v1, int sub_level, int level_index,
                      int dist_until_above_threshold, double threshold, double *dist_until);
  double screenspace_errormetric(int vc, int v0, int v1, int sub_level, int level_index,
                                 int dist_until_above_threshold, double threshold, double *dist_until);
  double perpendicular_epsilon_metric(int vc, int v0, int v1, int sub_level, int level_index,
                                      int dist_until_above_threshold, double threshold, double *dist_until);

  // Internal routines used by perpendicular_epsilon_metric.
  double get_closest_barycentricsq(double camera_u, double camera_v);
  int get_furthest_barycentric(double camera_u, double camera_v);

  // Lock the incremental thresholds to the specified priority...
  // used during initial adaptation of a base triangulation.
  void lock_incremental_thresholds(double lock_pri);

  // Unlock the incremental thresholds.
  void unlock_incremental_thresholds(void);

  // Recompute all priorities due to a view change.
  void recompute_priorities(double total_distance_moved, int force_recompute = FALSE);


  /*
   * Adaption routines
   */

  // Set target values.
  void set_target_priority(double target_priority);
  void set_target_triangle_count(int tri_limit);

  // Check that all the mergeable diamonds have their triangles' diamond_index fields
  // set correctly.
  void assert_tri_diamond_indices(void);

  // Adapt the triangulation to the target priority, while keeping under the triangle limit.
  void adapt_triangulation(void);


  /*
   * Texture-mapping
   */

  // Enable/disable texturing (just a rendering flag, doesn't change
  // the fact that textures are loaded).
  void set_texturing(int use_texturing);

  // Set up the texture arrays
  void init_texture(int num_textures, int texture_dim);

  // Build the mipmap levels for a texture.
  void build_miplevels(int section_index, unsigned char *tex_data);

  // Extend the borders of a miplevel (because there is no neighbour there to
  // provide the values).
  void extend_borders(int section, int level, int dim,
                      int extend_left, int extend_right, int extend_bottom, int extend_top);

  // Copy the borders of a miplevel.
  void copy_borders(int section, int l_section, int b_section, int bl_section,
                    int level, int dim);

  // Copy a vertical mipmap border.
  void copy_vertical_border(unsigned char *dest, unsigned char *src, int num_elems, int stride);

  // Copy a horizontal mipmap border.
  inline void copy_horizontal_border(unsigned char *dest, unsigned char *src, int num_elems) {
    memcpy(dest, src, 3 * num_elems);
  }

  // Copy a diagonal mipmap pixel.
  inline void copy_diagonal_pixel(unsigned char *dest, unsigned char *src) {
    dest[0] = src[0];
    dest[1] = src[1];
    dest[2] = src[2];
  }

  // All of the textures have been stored - do any processing necessary.
  void textures_done(void);

  /*
   * Scenery
   */

  // Set up the scenery-type and scenery-object arrays.
  void alloc_scenery(int numtypes, int num_obj);

  // Create a scenery object at a given location.
  void create_scenery_object(int scenery_type, vec3 location);

  // Set up a scenery texture (also initialises object size).
  void scenery_texture(int scenery_type, short scale_factor, int w, int h,
                       unsigned char *rgba_texture);

  /*
   * Vertex-morphing
   */

  // Morph all vertices to the current time.
  void do_morph(long current_time);

  // Get the current height of a vertex (taking into account any morph)
  unsigned short get_height(int vertex_index);
};
