// Main.cpp : Main event loop for the application.
//

#include "stdafx.h"


/*
 * User-settable parameters.
 */

double key_sensitivity, mouse_sensitivity;
double camera_acceleration, camera_max_velocity;
double camera_clip_near, camera_clip_far, camera_frustum_size;

char mythpath[256];

int target_triangle_count;
double target_pixel_size;

int use_texture_borders, use_texgen;
int draw_scenery;
int morph_vertices, morph_msec;

int skycol[3];

/*
 * Global variables.
 */

// Camera and GL-view variables.
vec3 camera_pos;
quaternion camera_dir;
double camera_vel;
GLfloat camera_modelmatrix[16];


// The Myth tag database.
TagDatabase *myth_tags;

// The list of maps that can be selected in the Map Dialog.
int mapsel_num_entrypoints, mapsel_choice;
entrypoint **mapsel_entrypoint;


// Data pertaining to the current mesh.
datamode data_mode = DATA_MYTHPLUGIN;

unsigned short *rawdata_heightarray;  // if DATA_FUNCTION
MeshTag *view_mesh;                   // if DATA_MYTHPLUGIN




// Input
DWORD last_update_time;
long mouse_coords[2], lastmouse_coords[2];
unsigned char mouse_buttons[2];


// Bintree stuff.
Bintree *bintree = NULL;
int must_adapt_triangulation, require_full_recompute;
double total_distance_moved;


/*
 * Entry points from the Windows wrapper.
 */

// Work out what we will be displaying. Can interact
// with the user via windows etc. because we haven't popped
// up the main GL window yet.
void user_init(void)
{
  double distsq;
  unsigned short *hptr;
  char *strptr, plugindir[256];
  FILE *tempfile;
  int x, y, i;
  iteratestruct *iterator;
  TagHashStructure *hash;

  // Initialise default settings.
  default_settings();

  // Read the user's settings.
  read_settingsfile();

  srand(7);

  switch(data_mode) {
  case DATA_FUNCTION:
    // Build the array of height data for the bintree.
    rawdata_heightarray = (unsigned short *) malloc(SQR((1 << 8) + 1) * sizeof(unsigned short));
    if (!rawdata_heightarray) mem_fail(SQR((1 << 8) + 1) * sizeof(unsigned short),
                                       "raw height-value array");
    hptr = rawdata_heightarray;
    for (y = 0; y <= 256; y++) {
      for (x = 0; x <= 256; x++) {
        distsq = SQR((x - 128) / 128.0) + SQR((y - 128) / 128.0);
        if (distsq >= 1.0)
          *hptr++ = 0;
        else
          *hptr++ = (unsigned short) (HEIGHT_MAX * pow(1 - distsq, 3));
      }
    }
    break;

  case DATA_MYTHPLUGIN:
    // Check to see if we already know where Myth II is.
    sprintf(plugindir, "%sMyth II.exe", mythpath);
    if ((tempfile = fopen(plugindir, "r")) == NULL) {
      // We can't find Myth II - need to ask the user to locate it.
      // This will let us find where all of the plugins are stored.
      if (!choose_file(256, mythpath, "Please locate Myth II.exe")) {
        end_program();
        return;
      }

      // Strip the last 'Myth II.exe' off the path.
      strptr = strrchr(mythpath, '\\');
      assert(strptr != NULL);
      strptr++;
      *strptr = '\0';

    } else {
      // We can find Myth II - don't need to ask the user again.
      fclose(tempfile);
    }

    // Read the tags.
    myth_tags = new TagDatabase;

    sprintf(plugindir, "%stags", mythpath);
    myth_tags->read_plugins(plugindir);

    sprintf(plugindir, "%splugins", mythpath);
    myth_tags->read_plugins(plugindir);

    // Iterate over all the entrypoints, so that we can build a dialog box.
    mapsel_num_entrypoints = myth_tags->entrypoint_hash->num_tags;
    if (mapsel_num_entrypoints == 0) {
      fprintf(debug_file, "ERROR: Unable to find any entrypoints for maps in any of the directories scanned.\n");
      premature_shutdown();
    }

    mapsel_entrypoint = (entrypoint **) malloc(mapsel_num_entrypoints * sizeof(entrypoint *));
    if (!mapsel_entrypoint) mem_fail(mapsel_num_entrypoints * sizeof(entrypoint *),
                                     "map selection array");

    i = 0; iterator = NULL;
    while (myth_tags->entrypoint_hash->iterate_tags(&iterator)) {
      mapsel_entrypoint[i++] = (entrypoint *) iterator->tagentry->tagdata;
    }
    assert(i == mapsel_num_entrypoints);

    // Sort the entrypoints by name, to present them in the dialog box.
    qsort((void *) mapsel_entrypoint, mapsel_num_entrypoints, sizeof(entrypoint *), entrypoint_sort_comparison);

    // Raise the map selection dialog box.
    map_selection_dialog();
    if (mapsel_choice == -1) {
      fprintf(debug_file, "User cancelled map selection. Exit.\n");
      end_program();
      return;
    }

    // Activate the plugin corresponding to this entrypoint.
    assert((mapsel_choice >= 0) && (mapsel_choice < mapsel_num_entrypoints));
    mapsel_entrypoint[mapsel_choice]->our_plugin->plugin_activated = TRUE;

    // Get this mesh's tag.
    hash = myth_tags->hash_structure(string_to_tagid("mesh"));
    assert(hash != NULL);
    view_mesh = (MeshTag *) hash->retrieve_tag(mapsel_entrypoint[mapsel_choice]->mesh);

    if (view_mesh == NULL) {
      fprintf(debug_file, "ERROR: Cannot find mesh '%s', yet it had an entrypoint!\n", tagid_to_string(mapsel_entrypoint[mapsel_choice]->mesh));
      premature_shutdown();
    }

    free(mapsel_entrypoint);
  }

  // Write the settings that the user has given us.
  write_settingsfile();
}

// Entrypoint comparison for building a plugin selection dialogue.
int entrypoint_sort_comparison(const void *ep1, const void *ep2)
{
  entrypoint *e1 = *((entrypoint **) ep1), *e2 = *((entrypoint **) ep2);

  return strcmp(e1->meshname, e2->meshname);
}

void user_setup(void)
{
  // Set up the GL rendering engine.
  setup_view();

  /*
   * Construct the Bintree.
   */

  switch(data_mode) {
  case DATA_FUNCTION:
    bintree = new Bintree(8, rawdata_heightarray, vec3(-50.01, -50.02, -30.03), vec3(100, 100, 40));
    break;

  case DATA_MYTHPLUGIN:
    // Construct the bintree from the mesh that we are viewing.
    bintree = view_mesh->make_bintree(use_texture_borders);
  }

  // Set the initial state of the system.
  poll_mouse(mouse_coords, mouse_buttons);
  must_adapt_triangulation = TRUE;
  require_full_recompute = FALSE;
  last_update_time = timeGetTime();

  // The camera's zero position is looking along -Z with upvector +Y.
  // We want it to initially be looking along along +X with upvector +Z.
  //
  // Compose this from two rotations - the first, about the -Y axis by PI/2 (90 degrees) which gets us
  // looking along +X with upvector +Y. The second, about the +X axis by PI/2 which gets our upvector to +Z.
  quaternion r1(cos(PI/4), 0, -sin(PI/4), 0), r2(cos(PI/4), sin(PI/4), 0, 0);
  camera_dir = r2 ^ r1;

  total_distance_moved = 0;
  camera_pos.zero();
  camera_vel = 0;
  switch(data_mode) {
  case DATA_FUNCTION:
    camera_pos.x = -30.0;
    camera_pos.z = 0;
    break;

  case DATA_MYTHPLUGIN:
    camera_pos.x = bintree->mesh_base.x + 0.2 * bintree->mesh_extent.x;
    camera_pos.y = bintree->mesh_base.y + 0.5 * bintree->mesh_extent.y;
    camera_pos.z = bintree->mesh_base.z + 
       (bintree->mesh->max_height - bintree->mesh->min_height) * bintree->mesh_delta.z;
    break;
  }

  calculate_camera_matrix();
  bintree->set_view(camera_pos, camera_dir);
  bintree->set_target_priority(target_pixel_size);
  bintree->set_target_triangle_count(target_triangle_count);

  // Perform an initial calculation on the bintree... straighten out
  // all the kinks, basically. The lock() / unlock() pair are to make sure
  // that rapidly changing incremental thresholds don't exceed our
  // assumed safety factors.
  bintree->lock_incremental_thresholds(target_pixel_size);
  bintree->base_triangulation();
  bintree->adapt_triangulation();
  bintree->unlock_incremental_thresholds();
}

void user_redraw(void)
{
  // Clear the scene.
  glClearColor(skycol[0] / 255.0f, skycol[1] / 255.0f, skycol[2] / 255.0f, 0.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadMatrixf(camera_modelmatrix);

  if (bintree) {
    bintree->do_morph(timeGetTime());
    bintree->draw();
  }
}

int user_idle(void)
{
  DWORD current_time = timeGetTime();

  if (current_time - last_update_time < 10)
    return 0;

  unsigned char keyboardbuf[256];
  double update_seconds = (current_time - last_update_time) / 1000.0;
  double constant_x, constant_y, constant_roll, move_x, move_y, move_mag, roll_mag, sintheta;
  quaternion rot_q;
  int incr_amount;

  /*
   * Input-handling
   */

  // Poll the keyboard.
  poll_keyboard(keyboardbuf);
  if (keyboardbuf[DIK_ESCAPE]) {
    end_program();
    return 0;
  }

  // Poll the mouse.
  lastmouse_coords[0] = mouse_coords[0]; lastmouse_coords[1] = mouse_coords[1];
  poll_mouse(mouse_coords, mouse_buttons);

  /*
   * Camera control
   */

  // Find out which direction we are moving in continuously (from keys).
  constant_x = constant_y = 0;
  if (keyboardbuf[DIK_LEFT])      constant_x    -= key_sensitivity;
  if (keyboardbuf[DIK_RIGHT])     constant_x    += key_sensitivity;
  if (keyboardbuf[DIK_UP])        constant_y    += key_sensitivity;
  if (keyboardbuf[DIK_DOWN])      constant_y    -= key_sensitivity;

  // Movement of 1.0 is PI radians, i.e. 180 degrees.
  move_x =  mouse_sensitivity * (mouse_coords[0] - lastmouse_coords[0]) + constant_x * update_seconds;
  move_y = -mouse_sensitivity * (mouse_coords[1] - lastmouse_coords[1]) + constant_y * update_seconds;

  // Construct the rotation quaternion in the relative XY plane.
  move_mag = sqrt(SQR(move_x) + SQR(move_y));
  if (move_mag > 0) {
    sintheta = sin(PI * move_mag / 2);
    rot_q.w   =   cos(PI * move_mag / 2);
    rot_q.v.x =   sintheta * (move_y / move_mag);
    rot_q.v.y = - sintheta * (move_x / move_mag);
    rot_q.v.z = 0;
  
    camera_dir = camera_dir ^ rot_q;
  }

  constant_roll = 0;
  if (keyboardbuf[DIK_Q])  constant_roll += key_sensitivity;
  if (keyboardbuf[DIK_E])  constant_roll -= key_sensitivity;

  // Same units as movement
  roll_mag = constant_roll * update_seconds;

  if (roll_mag != 0) {
    // Construct the rotation quaternion in the relative XY plane.
    rot_q.w = cos(PI * roll_mag / 2);
    rot_q.v.x = 0;
    rot_q.v.y = 0;
    rot_q.v.z = sin(PI * roll_mag / 2);

    camera_dir = camera_dir ^ rot_q;
  }

  calculate_camera_matrix();

  // Calculate the camera's velocity.
  if (keyboardbuf[DIK_A])  camera_vel += camera_acceleration * update_seconds;
  if (keyboardbuf[DIK_Z])  camera_vel -= camera_acceleration * update_seconds;
  if (camera_vel < -camera_max_velocity)
    camera_vel = -camera_max_velocity;
  else if (camera_vel > camera_max_velocity)
    camera_vel = camera_max_velocity;

  if (keyboardbuf[DIK_X])  camera_vel = 0;

  if (camera_vel != 0) {
    // Move the camera.
    vec3 dir_vector(0, 0, -1);
    dir_vector = dir_vector ^ camera_dir;

    camera_pos += camera_vel * update_seconds * dir_vector;

    total_distance_moved += camera_vel * update_seconds;
    must_adapt_triangulation = TRUE;
  }
  last_update_time = timeGetTime();

  /*
   * Mesh adaption
   */

  // Do we want to increase or decrease the number of triangles being drawn?
  incr_amount = 10;
  if ((keyboardbuf[DIK_LCONTROL]) || (keyboardbuf[DIK_RCONTROL]))
    incr_amount *= 5;

  if (keyboardbuf[DIK_MINUS]) {
    must_adapt_triangulation = TRUE;
    target_triangle_count -= incr_amount;
    if (target_triangle_count < 0)
      target_triangle_count = 0;

    bintree->set_target_triangle_count(target_triangle_count);
  } else if (keyboardbuf[DIK_EQUALS]) {
    must_adapt_triangulation = TRUE;
    target_triangle_count += incr_amount;
    bintree->set_target_triangle_count(target_triangle_count);
  }

  bintree->set_view(camera_pos, camera_dir);

  // Adapt the triangulation (if the user isn't holding down the shift key).
  if (keyboardbuf[DIK_LSHIFT] || keyboardbuf[DIK_RSHIFT]) {
    // If the user starts forcing the program not to recompute, then our distance
    // bounds will become invalid - they will need to be totally recomputed.
    require_full_recompute = TRUE;
  } else {
    if (must_adapt_triangulation || require_full_recompute) {
      bintree->recompute_priorities(total_distance_moved, require_full_recompute);
      bintree->adapt_triangulation();
      require_full_recompute = FALSE;
      must_adapt_triangulation = FALSE;
    }
  }

  // The tab key disables texturing.
  bintree->set_texturing(keyboardbuf[DIK_TAB] == 0);

  return 1;
}

void user_destroy(void)
{
  delete bintree;
}

/*
 * Internal routines
 */

// Set up GL rendering parameters.
void setup_view(void)
{
  GLfloat frust_x, frust_y;

  frust_x = (GLfloat) camera_frustum_size;
  frust_y = (GLfloat) (viewresolution_y * camera_frustum_size / viewresolution_x);

	glViewport(0, 0, viewresolution_x, viewresolution_y);

  // Set up the viewing transform.
  glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-frust_x, frust_x, -frust_y, frust_y,
            (GLfloat) camera_clip_near, (GLfloat) camera_clip_far);
	glMatrixMode(GL_MODELVIEW);

  // Enable the Z-buffer.
  glEnable(GL_DEPTH_TEST);

#if 0
  // Haven't decided whether to enable back-face culling yet.
	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);
#endif
}

// Set up the camera modelmatrix.
void calculate_camera_matrix(void)
{
  // Calculate the modelmatrix that is the _inverse_ of the camera's transformation, which gives us the
  // transformation from modelspace to cameraspace. This is what we need for the GL modelmatrix.
  quaternion q_inv = ~camera_dir;
  vec3 rotated_trans;
  double rotationmatrix[9];

  q_inv.make_matrix(rotationmatrix);

  camera_modelmatrix[0]  = (GLfloat) rotationmatrix[0];
  camera_modelmatrix[1]  = (GLfloat) rotationmatrix[1];
  camera_modelmatrix[2]  = (GLfloat) rotationmatrix[2];
  camera_modelmatrix[3]  = 0;

  camera_modelmatrix[4]  = (GLfloat) rotationmatrix[3];
  camera_modelmatrix[5]  = (GLfloat) rotationmatrix[4];
  camera_modelmatrix[6]  = (GLfloat) rotationmatrix[5];
  camera_modelmatrix[7]  = 0;

  camera_modelmatrix[8]  = (GLfloat) rotationmatrix[6];
  camera_modelmatrix[9]  = (GLfloat) rotationmatrix[7];
  camera_modelmatrix[10] = (GLfloat) rotationmatrix[8];
  camera_modelmatrix[11] = 0;

  // This inverse transformation is M = R-1T-1, i.e. the inverse translation is postmultiplied against
  // the inverse rotation. Thus the translation amount should be rotated by the inverse quaternion.
  rotated_trans = (-camera_pos) ^ q_inv;
  camera_modelmatrix[12] = (GLfloat) rotated_trans.x;
  camera_modelmatrix[13] = (GLfloat) rotated_trans.y;
  camera_modelmatrix[14] = (GLfloat) rotated_trans.z;
  camera_modelmatrix[15] = 1;
}

// Set up default settings.
void default_settings(void)
{
  key_sensitivity = 0.5;
  mouse_sensitivity = 0.001;

  camera_acceleration = 100;
  camera_max_velocity = 100;

  strcpy(mythpath, "");

  target_triangle_count = 20000;
  target_pixel_size = 2.0;

  camera_clip_near = 1.0;
  camera_clip_far = 384.0;
  camera_frustum_size = 1.0;

  viewresolution_x = 640;
  viewresolution_y = 480;
  view_fullscreen = TRUE;

  use_texture_borders = FALSE;   // not supported on many consumer 3D chipsets :(

  draw_scenery = TRUE;
  use_texgen = TRUE;

  skycol[0] = 70;
  skycol[1] = 70;
  skycol[2] = 130;

  morph_vertices = TRUE;
  morph_msec = 600;
}

// Read our settings from a file.
void read_settingsfile(void)
{
  char inbuf[200], *strptr, tempstr[20];
  FILE *in_file = fopen("settings.txt", "r");

  if (!in_file) {
    // The settings file doesn't exist. This is the first time that
    // the user has run this program.
    show_hello_box();
    return;
  }

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Keyboard Sensitivity: %f\n", &key_sensitivity);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Mouse Sensitivity: %f\n", &mouse_sensitivity);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Camera Acceleration: %f\n", &camera_acceleration);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Camera Max Velocity: %f\n", &camera_max_velocity);

  fgets(inbuf, 200, in_file);
  strptr = strchr(inbuf, ':');
  if (strptr != NULL) {
    // Copy mythpath from after the first colon.
    strptr++;
    while (isspace(*strptr))
      strptr++;

    strcpy(mythpath, strptr);
    
    if (mythpath[0] != '\0') {
      // Strip trailing whitespace from the path.
      strptr = &mythpath[strlen(mythpath) - 1];
      while (isspace(*strptr) && (strptr >= mythpath))
        strptr--;
      strptr[1] = '\0';
    }
  }

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Target Triangle Count: %d\n", &target_triangle_count);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Error Bound (Pixels): %f\n", &target_pixel_size);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Camera Near Clip: %f\n", &camera_clip_near);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Camera Far Clip: %f\n", &camera_clip_far);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Camera Frustum Radius: %f\n", &camera_frustum_size);

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Resolution: %d x %d\n", &viewresolution_x, &viewresolution_y);

  fgets(inbuf, 200, in_file);
  if (sscanf(inbuf, "Fullscreen: %s\n", tempstr) == 1) {
    if (tolower(tempstr[0]) == 'y')
      view_fullscreen = TRUE;
    else if (tolower(tempstr[0]) == 'n')
      view_fullscreen = FALSE;
  }

  fgets(inbuf, 200, in_file);
  if (sscanf(inbuf, "Use Texture Borders: %s\n", tempstr) == 1) {
    if (tolower(tempstr[0]) == 'y')
      use_texture_borders = TRUE;
    else if (tolower(tempstr[0]) == 'n')
      use_texture_borders = FALSE;
  }

  fgets(inbuf, 200, in_file);
  if (sscanf(inbuf, "Draw Scenery: %s\n", tempstr) == 1) {
    if (tolower(tempstr[0]) == 'y')
      draw_scenery = TRUE;
    else if (tolower(tempstr[0]) == 'n')
      draw_scenery = FALSE;
  }

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Sky Colour: %d %d %d\n", &skycol[0], &skycol[1], &skycol[2]);

  fgets(inbuf, 200, in_file);
  if (sscanf(inbuf, "Morph Vertices: %s\n", tempstr) == 1) {
    if (tolower(tempstr[0]) == 'y')
      morph_vertices = TRUE;
    else if (tolower(tempstr[0]) == 'n')
      morph_vertices = FALSE;
  }

  fgets(inbuf, 200, in_file);
  sscanf(inbuf, "Morph Milliseconds: %d\n", &morph_msec);

  fgets(inbuf, 200, in_file);
  if (sscanf(inbuf, "Use TexGen: %s\n", tempstr) == 1) {
    if (tolower(tempstr[0]) == 'y')
      use_texgen = TRUE;
    else if (tolower(tempstr[0]) == 'n')
      use_texgen = FALSE;
  }

  fclose(in_file);
}

// Write our settings out to a file.
void write_settingsfile(void)
{
  FILE *out_file = fopen("settings.txt", "w");

  if (!out_file) {
    fprintf(debug_file, "WARNING: Could not write settings file\n");
    return;
  }

  fprintf(out_file, "Keyboard Sensitivity: %f\n", key_sensitivity);
  fprintf(out_file, "Mouse Sensitivity: %f\n", mouse_sensitivity);
  fprintf(out_file, "Camera Acceleration: %f\n", camera_acceleration);
  fprintf(out_file, "Camera Max Velocity: %f\n", camera_max_velocity);
  fprintf(out_file, "Myth Path: %s\n", mythpath);
  fprintf(out_file, "Target Triangle Count: %d\n", target_triangle_count);
  fprintf(out_file, "Error Bound (Pixels): %f\n", target_pixel_size);
  fprintf(out_file, "Camera Near Clip: %f\n", camera_clip_near);
  fprintf(out_file, "Camera Far Clip: %f\n", camera_clip_far);
  fprintf(out_file, "Camera Frustum Radius: %f\n", camera_frustum_size);
  fprintf(out_file, "Resolution: %d x %d\n", viewresolution_x, viewresolution_y);
  fprintf(out_file, "Fullscreen: %s\n", view_fullscreen ? "Yes" : "No");
  fprintf(out_file, "Use Texture Borders: %s\n", use_texture_borders ? "Yes" : "No");
  fprintf(out_file, "Draw Scenery: %s\n", draw_scenery ? "Yes" : "No");
  fprintf(out_file, "Sky Colour: %d %d %d\n", skycol[0], skycol[1], skycol[2]);
  fprintf(out_file, "Morph Vertices: %s\n", morph_vertices ? "Yes" : "No");
  fprintf(out_file, "Morph Milliseconds: %d\n", morph_msec);
  fprintf(out_file, "Use TexGen: %s\n", use_texgen ? "Yes" : "No");

  fclose(out_file);
}
