// MeshTag.h : Definitions for class MeshTag and supporting code.
//

// The 1024-byte header for the mesh data.
typedef struct meshheader_ {
  tag_id colormap;      // colormap        (of type '.256')
  tag_id watertype;     // water type      (of type 'medi')

  short x_sections;     // X resolution of mesh in sections
  short y_sections;     // Y resolution of mesh in sections

  long unknown1;
  long height_length;   // length of the height array
  long unknown2;
  long bulk_offset;     // offset from the start of the tag
  long bulk_datasize;   // total length of all bulk arrays

  long unknown3;
  long num_itemtypes;   // number of 32-byte item types
  long itemtype_offset; // offset of itemtype (from bulk_offset)
  long itemtype_length; // length of itemtype

  long unknown4;
  long num_items;       // number of 64-byte items
  long item_offset;     // offset of item (from bulk_offset)
  long item_length;     // length of item

  long unknown5;
  tag_id meshlighting;  // mesh lighting   (of type 'meli')
  tag_id watertype2;    // water type 2    (of type 'medi')

  short mesh_flags;     // single / multiplayer, amongst other things
  short netgame_flags;  // 0x0001 - 0x0400 = individual game types from BC to Rugby

  tag_id particle;      // particle effect (of type 'part')
  long num_teams;       // number of teams

  long unknown6;
  long unknown7;
  long unknown8;
  long unknown9;
  long unknown10;
  long unknown11;
  long unknown12;
  short unknown13;
  short unknown14;
  short unknown15;
  short unknown16;

  tag_id ambientsound;  // ambient sound   (of type 'amso')

  long num_actions;     // number of map actions
  long action_offset;   // offset of actions (from bulk_offset)
  long action_length;   // length of action

  tag_id mapname;       // map name        (of type 'stli')
  tag_id postgame;      // postgame screen (of type '.256')
  tag_id pregame;       // pregame screen  (of type '.256')
  tag_id overhead;      // overhead map    (of type '.256')
  tag_id nextlevel;     // next level      (of type 'mesh')

  long unknown17;
  tag_id objectives;    // hints and object(of type 'stli')
  long unknown18;
  long unknown19;
  tag_id narration;     // narration text  (of type 'stli')
  long unknown20;
  long unknown21;
  long unknown22;

  long array1_offset;   // offset of array1 (from bulk_offset)
  long array1_length;   // length of array1

  long unknown23;
  long array2_offset;   // offset of array2 (from bulk_offset)
  long array2_length;   // length of array2

  long unknown24;
  long unknown25;
  long unknown26;
  long unknown27;

  tag_id wind;          // wind effects    (of type 'wind')

  long unknown28;
  long unknown29;
  long unknown30;
  long unknown31;
  long unknown32;

  tag_id subtitle;      // subtitles       (of type 'stli')
  tag_id narrationsound;// narration sound (of type 'soun')

  long unknown33[191];

} meshheader;

// The values stored for a gridpoint on the mesh.
typedef struct gridval_ {
  short height;               // elevation of the grid point
  char grad_x, grad_y;        // gradient at this point (?)
  unsigned char water_flags;  // something to do with water
  unsigned char passability;  // each nibble stores passability for one of the
                              // sub-triangles in this mesh cell.
  unsigned short unused[3];
} gridval;

// Not an exhaustive listing of itemtype types. :(
#define ITEMTYPE_CAMERA           0
#define ITEMTYPE_SCENERY          1
#define ITEMTYPE_UNIT             3
#define ITEMTYPE_SOUND            5
#define ITEMTYPE_MODEL            6
#define ITEMTYPE_FX               10

// A type of item on the mesh.
typedef struct itemtype_ {
  short type;                 // ITEMTYPE_whatever
  short flags;
  tag_id item_tag;
  short team;
  short unknown[5];
  class ScenTag *scenerytag;  // nb: am assuming this is unused on disk.
                              // it appears to be.
  short bintree_scenery_index;// also appears to be unused
  short item_count;           // number of items present currently... zero on disk
  short item_max;             // max number of item ever allowed on level
  short itemtype_index;       // reference number used in the item section
} itemtype;

// An item placed on the mesh.
typedef struct itempos_ {
  short unknown1;
  short flags;                // bit 0 = not initially present
  short type;                 // ITEMTYPE_whatever
  short itemtype_index;       // reference number back into itemtype
  short item_id;              // unique ID for this instance
  short unknown2;
  long item_pos[3];           // positions in 1024ths of a mesh point, X Y Z
  short unknown3[4];
  short orientation;          // 0 - 65536 around circle
  short unknown4[13];
  short item_id2;             // appears to be another unique ID
  short item_id3;             // ditto
} itempos;

// A mesh - a map to be played on.
class MeshTag : public Tag {
public:
  meshheader *mesh_header;
  int num_scenery_types, num_scenery_items;
  int num_scenery_textures;

  // Bulk arrays
  short min_height, max_height;
  gridval *grid;
  int *scenery_lookup;
  itemtype *itemtypearray;
  itempos *itemarray;

  // Read the mesh's static data from disk.
  virtual void read_data(void);

  // Read the big mesh arrays from disk, or dump them after use.
  virtual void load_bulk_data(void);
  virtual void free_bulk_data(void);

  // Create a new tag.
  MeshTag(class MythPlugin *plugin, tagheader *header);

  // Turn the mesh into a Bintree.
  Bintree *make_bintree(int use_texture_borders);

  // Do we draw this kind of scenery?
  int draw_scenery_tag(tag_id scenery_tag);

  // Delete the tag.
  ~MeshTag();
};
