// MythPlugin.cpp : Routines for class MythPlugin.
//

#include "stdafx.h"

// Read a plugin from a file. First reads the plugin header, and then
// constructs the full plugin data type if it's valid.
MythPlugin *read_plugin(char *pluginname)
{
  FILE *tagfile;
  pluginheader header;

  // Open the plugin file.
  if ((tagfile = fopen(pluginname, "rb")) == NULL) {
    fprintf(debug_file, "ERROR: Cannot open plugin file '%s'\n", pluginname);
    return NULL;
  }

  /*
   * Read the plugin header.
   */

  header.plugin_type = myth_read_short(tagfile);
  header.version     = myth_read_short(tagfile);
  
  myth_read_bytes(tagfile, sizeof(header.name), (unsigned char *) header.name);
  myth_read_bytes(tagfile, sizeof(header.url),  (unsigned char *) header.url);

  header.num_entrypoints = myth_read_short(tagfile);
  header.num_tags = myth_read_short(tagfile);
  header.checksum = myth_read_long(tagfile);
  header.flags = myth_read_long(tagfile);
  header.size = myth_read_long(tagfile);

  header.header_checksum = myth_read_long(tagfile);

  header.unused[0] = myth_read_long(tagfile);

  header.plugin_signature = myth_read_tag(tagfile);

  /*
   * Check that the plugin header is valid.
   */

  if ((header.plugin_type < 0) || (header.plugin_type > PLUGINTYPE_MAX) ||
      (header.num_tags    < 0) || (header.num_tags    > MAX_PLUGIN_TAGS)) {
    fprintf(debug_file, "ERROR: Plugin file '%s' does not have a valid header (type %d, tags %d)\n",
            pluginname, header.plugin_type, header.num_tags);
    return NULL;
  }

  return new MythPlugin(tagfile, &header);
}

// Read the plugin from a file... called once the header has been
// read and verified.
MythPlugin::MythPlugin(FILE *tagfile, pluginheader *copy_header)
{
  int i;
  tagheader *t;

  this->tagfile = tagfile;
  memcpy(&header, copy_header, sizeof(pluginheader));

  // Plugins are initially not activated.
  if (header.plugin_type == PLUGINTYPE_PLUGIN)
    plugin_activated = FALSE;
  else
    plugin_activated = TRUE;

  /*
   * Read the entry points.
   */

  if (header.num_entrypoints == 0) {
    entrypoint_array = NULL;
  } else {
    entrypoint_array = (entrypoint *) malloc(header.num_entrypoints
                                             * sizeof(entrypoint));
    if (!entrypoint_array) mem_fail(header.num_entrypoints * sizeof(entrypoint),
                                    "MythPlugin entrypoint array");
    
    for (i = 0; i < header.num_entrypoints; i++) {
      
      entrypoint_array[i].mesh = myth_read_tag(tagfile);
      
      myth_read_longs(tagfile, 2, entrypoint_array[i].unused);

      // Skip a long value and write a pointer to our plugin into the entrypoint structure.
      myth_read_long(tagfile);
      entrypoint_array[i].our_plugin = this;
      
      myth_read_bytes(tagfile, sizeof(entrypoint_array[i].meshtagname),
        (unsigned char *) entrypoint_array[i].meshtagname);
      myth_read_bytes(tagfile, sizeof(entrypoint_array[i].meshname),
        (unsigned char *) entrypoint_array[i].meshname);
    }
  }

  /*
   * Set up the tags. Don't actually read any of their data from the file yet though.
   */

  if (header.num_tags == 0) {
    tags = NULL;
    return;
  }

  tags = (Tag **) malloc(header.num_tags * sizeof(Tag *));
  if (!tags) mem_fail(header.num_tags * sizeof(Tag *),
                      "MythPlugin tag array");

  for (i = 0; i < header.num_tags; i++) {
    t = (tagheader *) malloc(sizeof(tagheader));
    if (!t) mem_fail(sizeof(tagheader),
                     "MythPlugin tag header");
    
    t->datum_header = myth_read_short(tagfile);
    t->flags = myth_read_byte(tagfile);
    t->type = myth_read_byte(tagfile);

    myth_read_bytes(tagfile, sizeof(t->name), (unsigned char *) t->name);

    t->tag_type = myth_read_tag(tagfile);
    t->our_id = myth_read_tag(tagfile);

    t->offset = myth_read_long(tagfile);
    t->size = myth_read_long(tagfile);

    t->userdata = myth_read_long(tagfile);
    t->version = myth_read_short(tagfile);

    t->foundation_id = myth_read_byte(tagfile);
    t->priority = myth_read_byte(tagfile);

    t->telltale_signature = myth_read_tag(tagfile);

    if (t->telltale_signature != string_to_tagid("mth2")) {
      // This tag should always read 'mth2' - it doesn't, something is badly wrong.
      fprintf(debug_file, "ERROR: corrupt tag #%d (%s/%s) has signature at location %8X of '%s' != 'mth2'\n",
              i, tagid_to_string(t->tag_type), tagid_to_string(t->our_id),
              ftell(tagfile) - 4, tagid_to_string(t->telltale_signature));
      premature_shutdown();
      return;
    }

    tags[i] = new_tag(this, t);
  }
}

MythPlugin::~MythPlugin(void)
{
  fclose(tagfile);

  for (int i = 0; i < header.num_tags; i++)
    delete tags[i];

  if (header.num_tags > 0)
    free(tags);

  if (header.num_entrypoints > 0)
    free(entrypoint_array);
}
