// MythTag.cpp : Routines for class MythTag and supporting code.
//

#include "stdafx.h"

// Global null tag.
tag_id null_tag_id = -1L;

// Routines for reading Myth's platform-independent tag files.
unsigned char myth_read_byte(FILE *read_file)
{
  int readval;

  if ((readval = fgetc(read_file)) == EOF) {
    fprintf(debug_file, "Premature end of file reading Myth tags!\n");
    premature_shutdown();
  }

  return readval;
}

short myth_read_short(FILE *read_file)
{
  // Data is stored in big-endian format in Myth's tag files.
  short accum;

  accum = myth_read_byte(read_file);
  accum <<= 8;
  accum += myth_read_byte(read_file);

  return accum;
}

long myth_read_long(FILE *read_file)
{
  // Data is stored in big-endian format in Myth's tag files.
  long accum;

  accum = myth_read_byte(read_file);
  accum <<= 8;
  accum += myth_read_byte(read_file);
  accum <<= 8;
  accum += myth_read_byte(read_file);
  accum <<= 8;
  accum += myth_read_byte(read_file);

  return accum;
}

void myth_read_bytes(FILE *read_file, int n_bytes, unsigned char *dest)
{
  for (int i = 0; i < n_bytes; i++)
    *dest++ = myth_read_byte(read_file);
}

void myth_read_shorts(FILE *read_file, int n_shorts, short *dest)
{
  for (int i = 0; i < n_shorts; i++)
    *dest++ = myth_read_short(read_file);
}

void myth_read_longs(FILE *read_file, int n_longs, long *dest)
{
  for (int i = 0; i < n_longs; i++)
    *dest++ = myth_read_long(read_file);
}

// Print out a tag-id (character by character). NB: uses a static buffer of 10 tags.
// Expressions using more than 10 tag IDs will have problems - also, save the 
// results if you want to keep them.
char *tagid_to_string(tag_id write_id)
{
  static char tagstring[10][5];
  static int bufindex = 0;

  if (write_id == null_tag_id) {
    return "NULL";
  } else {
    bufindex = (bufindex + 1) % 10;

    tagstring[bufindex][4] = '\0';
    tagstring[bufindex][3] = write_id & 0xFF;
    write_id >>= 8;
    tagstring[bufindex][2] = write_id & 0xFF;
    write_id >>= 8;
    tagstring[bufindex][1] = write_id & 0xFF;
    write_id >>= 8;
    tagstring[bufindex][0] = write_id & 0xFF;

    return tagstring[bufindex];
  }
}

// Convert a string into a tag-id.
tag_id string_to_tagid(char *tagstr)
{
  tag_id out_tag;

  out_tag = *tagstr++;
  out_tag <<= 8;
  out_tag += *tagstr++;
  out_tag <<= 8;
  out_tag += *tagstr++;
  out_tag <<= 8;
  out_tag += *tagstr++;

  return out_tag;
}

/*
 * Tag - routines for all tag types.
 */

// Create a new tag from a header that we read from disk.
Tag *new_tag(MythPlugin *plugin, tagheader *header)
{
  if (header->tag_type == string_to_tagid("mesh")) {
    return new MeshTag(plugin, header);

  } else if (header->tag_type == string_to_tagid(".256")) {
    return new X256Tag(plugin, header);

  } else if (header->tag_type == string_to_tagid("scen")) {
    return new ScenTag(plugin, header);

  } else if (header->tag_type == string_to_tagid("core")) {
    return new CoreTag(plugin, header);

  } else if (header->tag_type == string_to_tagid("obje")) {
    return new ObjeTag(plugin, header);

  } else {
    return new UnimplementedTag(plugin, header);
  }
}

Tag::Tag(MythPlugin *my_plugin, tagheader *header_struct)
{
  plugin = my_plugin;
  header = header_struct;

  loaded_data = FALSE;
  loaded_bulk_data = FALSE;
}

Tag::~Tag(void)
{
  free(header);
}

/*
 * UnimplementedTag - any tags whose format isn't understood.
 */

// Create a new tag.
UnimplementedTag::UnimplementedTag(MythPlugin *my_plugin, tagheader *header_struct)
   : Tag(my_plugin, header_struct)
{
}

// Read the tag data from disk.
void UnimplementedTag::read_data(void)
{
  // We have no idea how to read this kind of tag.
  loaded_data = TRUE;
}

// Read the bulk tag data (any big arrays) from disk.
void UnimplementedTag::load_bulk_data(void)
{
  if (!loaded_bulk_data) {
    loaded_bulk_data = TRUE;

    // No bulk data.
    bulkdata_size = 0;
  }
}

// Free the bulk tag data that we don't need any more
// (but keep the rest of the tag around).
void UnimplementedTag::free_bulk_data(void)
{
  if (loaded_bulk_data) {
    loaded_bulk_data = FALSE;

    // No bulk data to deallocate.
  }
}



