// PriQueue.cpp : Routines for class PriQueue
//
// Manages a priority-queue for the Bintree; is either for splittable
// triangles or mergeable diamonds, ordered by either their priority or their
// recomputation distance.

#include "stdafx.h"

// Create a new (blank) priority queue.
PriQueue::PriQueue(int sort_order)
{
  num_elements = queue_bufsize = 0;
  queue_buf = NULL;

  this->sort_order = sort_order;
}

// Clear the priority queue.
void PriQueue::clear(void)
{
  num_elements = 0;
}

// Insert an element into the priority queue and return its priority-queue location.
void PriQueue::insert_element(priq_info *insert_element)
{
  int nextval = ++num_elements;

  if (nextval >= queue_bufsize) {
    queue_bufsize += queue_growsize;
    queue_buf = (priq_info **) realloc((void *) queue_buf, queue_bufsize * sizeof(int));
    if (!queue_buf) mem_fail(queue_bufsize * sizeof(int),
                             "Priority Queue queue buffer grow");
  }

  insert_element->priq_location = nextval;
  bubble_up(insert_element);
}

// Get the highest-priority element off the queue (with/without deleting it).
priq_info *PriQueue::get_element(void)
{
  if (num_elements == 0)
    return NULL;

  // We will return the element from the top of the heap.
  return queue_buf[1];
}

// An element must be prematurely removed from the priority queue.
void PriQueue::remove_element(priq_info *remove_me)
{
  priq_info *bubbling_element;

  assert((remove_me->priq_location >= 1) &&
         (remove_me->priq_location <= num_elements));

  // Decrement the number of elements.
  num_elements--;

  if (remove_me->priq_location <= num_elements) {
    // If this element wasn't the last one in the heap, then get
    // the element from the far bottom of the heap and put it here.
    bubbling_element = queue_buf[num_elements + 1];
    assert(bubbling_element->priq_location == num_elements + 1);
    bubbling_element->priq_location = remove_me->priq_location;

    if (sort_order * (bubbling_element->priority - remove_me->priority) > 0) {
      // The new element is higher priority than the element that was here -
      // bubble it up the heap until it reaches the right place.
      bubble_up(bubbling_element);
    } else {
      // The new element is lower priority than the element that was here -
      // bubble it down the heap until it reaches the right place.
      bubble_down(bubbling_element);
    }
  }

  remove_me->priq_location = -1;
}

// Bubble an element up the heap from a starting location until it reaches the correct location.
void PriQueue::bubble_up(priq_info *bubbling_element)
{
  int location, parent_location;
  priq_info *parent_element;

  location = bubbling_element->priq_location;
  while (1) {
    if ((parent_location = location >> 1) == 0) {
      // We have bubbled to the top of the heap! Stop.
      break;
    }

    parent_element = queue_buf[parent_location];
    assert(parent_element->priq_location == parent_location);
    if (sort_order * (bubbling_element->priority - parent_element->priority) > 0) {
      // The bubbling element should be higher than its parent element.
      // Move the parent element down to make room, and continue bubbling.
      queue_buf[location] = parent_element;
      parent_element->priq_location = location;

      location = parent_location;
    } else {
      // The bubbling element should be lower than its parent element;
      // this is fine, we don't need to continue bubbling. Stop.
      break;
    }
  }
  
  // We have bubbled as far as we're going to go.
  queue_buf[location] = bubbling_element;
  bubbling_element->priq_location = location;
}

// Bubble an element down the heap from a starting location until it reaches the correct location.
void PriQueue::bubble_down(priq_info *bubbling_element)
{
  int location, larger_location, right_location;
  priq_info *larger_element, *right_element;

  location = bubbling_element->priq_location;
  while (1) {
    if ((larger_location = location << 1) > num_elements) {
      // We have bubbled down to a leafnode at the bottom of the heap. Stop.
      break;
    }
    larger_element = queue_buf[larger_location];
    assert(larger_element->priq_location == larger_location);

    right_location = larger_location + 1;
    if (right_location <= num_elements) {
      // We are at a node which has a right child. Look to see which
      // of the children has a higher priority.
      right_element = queue_buf[right_location];
      assert(right_element->priq_location == right_location);

      if (sort_order * (right_element->priority - larger_element->priority) > 0) {
        // The right child has a higher priority than the left child;
        // we should bubble down the right side.
        larger_location = right_location;
        larger_element = right_element;
      }
    }

    // Is the bubbling element of a higher priority than the highest-priority
    // child?
    if (sort_order * (bubbling_element->priority - larger_element->priority) > 0) {
      // Yes. The bubbling element is of a higher priority than both its
      // children and thus the heap property is preserved. Stop here.
      break;
    } else {
      // No - we must continue bubbling down the higher side. Move the higher
      // child into the parent node.
      queue_buf[location] = larger_element;
      larger_element->priq_location = location;

      location = larger_location;
    }
  }

  // We have bubbled as far as we're going to go.
  queue_buf[location] = bubbling_element;
  bubbling_element->priq_location = location;
}

