/*
* quaternion.h
* quaternion stuff for particle-system.
*/

#define SQR(x)      ((x) * (x))

class quaternion {
public:
  double w;
  vec3 v;

  inline quaternion() {
  }

  inline quaternion(double aw, double ax, double ay, double az)
  {
    w = aw;
    v = vec3(ax, ay, az);
  }

  inline quaternion(double aw, vec3 av)
  {
    w = aw;
    v = av;
  }

  inline void unit(void)
  {
    w = 1;
    v.zero();
  }

  inline friend double mag(quaternion &q)
  {
    return sqrt(q * q);
  }

  inline void operator= (double *a)
  {
    read(a);
  }

  inline friend quaternion operator+ (quaternion &a, quaternion &b)
  {
    return quaternion(a.w + b.w, a.v + b.v);
  }
  
  inline quaternion &operator+= (quaternion &a)
  {
    w += a.w;
    v += a.v;
    
    return *this;
  }
  
  inline friend quaternion operator- (quaternion &a, quaternion &b)
  {
    return quaternion(a.w - b.w, a.v - b.v);
  }
  
  inline friend quaternion operator- (quaternion &q)
  {
    return quaternion(-q.w, -q.v);
  }

  inline quaternion &operator-= (quaternion &a)
  {
    w -= a.w;
    v -= a.v;
    
    return *this;
  }
  
  inline friend quaternion operator* (double s, quaternion &q)
  {
    return quaternion(q.w * s, q.v * s);
  }
  
  inline friend quaternion operator* (quaternion &q, double s)
  {
    return quaternion(q.w * s, q.v * s);
  }
  
  inline quaternion &operator*= (double s)
  {
    w *= s;
    v *= s;
    
    return *this;
  }
  
  inline friend quaternion operator/ (quaternion &q, double s)
  {
    return quaternion(q.w / s, q.v / s);
  }
  
  inline quaternion &operator/= (double s)
  {
    w /= s;
    v /= s;
    
    return *this;
  }

  // Dot product
  inline friend double operator* (quaternion &a, quaternion &b)
  {
    return (a.w * b.w + a.v * b.v);
  }
  
  inline friend void normalize(quaternion &q)
  {
    q /= mag(q);
  }

  inline void read(double *array)
  {
    w = array[0];
    v.read(&array[1]);
  }

  inline void store(double *array)
  {
    array[0] = w;
    v.store(&array[1]);
  }

  // Multiplicative inverse
  inline quaternion operator~ (void)
  {
    double m_inv = 1.0 / mag(*this);

    return quaternion (+m_inv * w, -m_inv * v);
  }

  // Quaternion multiplication
  inline friend quaternion operator^ (quaternion &a, quaternion &b)
  {
    return quaternion((a.w*b.w) - (a.v*b.v), (a.v^b.v) + (b.w*a.v) + (a.w*b.v));
  }

  // Quaternion rotation applied to a vector
  inline friend vec3 operator^ (vec3 &vv, quaternion &q)
  {
    quaternion temp_v(0, vv);

    return (q ^ temp_v ^ ~q).v;
  }

  // Convert the quaternion into a rotation matrix
  inline void make_matrix(double *m)
  {
    // We look at the quaternion's effects on the unit basis vectors.
    // This gives us the rotation matrix for q = [W, X, Y, Z]:
    //
    //  (1 - 2Y^2 - 2Z^2     2XY - 2WZ         2XZ + 2WY  )
    //  (   2XY + 2WZ     1 - 2X^2 - 2Z^2      2YZ - 2WX  )
    //  (   2XZ - 2WY        2YZ + 2WX     1 - 2X^2 - 2Y^2)
    //
    //  Writes into a 3x3 matrix organised as (0, 1, 2 = column 0)

    m[0] = 1 - 2*v.y*v.y - 2*v.z*v.z;
    m[1] = 2*v.x*v.y + 2*w*v.z;
    m[2] = 2*v.x*v.z - 2*w*v.y;
    m[3] = 2*v.x*v.y - 2*w*v.z;
    m[4] = 1 - 2*v.x*v.x - 2*v.z*v.z;
    m[5] = 2*v.y*v.z + 2*w*v.x;
    m[6] = 2*v.x*v.z + 2*w*v.y;
    m[7] = 2*v.y*v.z - 2*w*v.x;
    m[8] = 1 - 2*v.x*v.x - 2*v.y*v.y;
  }

  friend quaternion q_slerp(quaternion &a, double frac, quaternion &b)
  {
    double theta, sintheta, fwd, back, costheta;
    quaternion temp_result;
    
    // We first find the angle between the quaternions to be interpolated.
    costheta = a.v * b.v;
    
    if (1.0 + costheta > 1e-06) {
      // cos(theta) is not close to -1.
      
      if (1.0 - costheta > 1e-06) {
        // cos(theta) is not close to +1.
        
        // Thus we can use the normal method of interpolation.
        theta = acos(costheta);
        
        // Interpolate using the formula
        // (sin ((1-t)*theta) / sin (theta))*a + (sin ((t*theta) / sin (theta))*b
        sintheta = sin(theta);
        
        fwd  = sin(     frac  * theta) / sintheta;
        back = sin((1 - frac) * theta) / sintheta;
      } else {
        // cos(theta) is close to +1. Thus the vectors are very close together
        // and we should use linear interpolation.
        fwd  = frac;
        back = 1 - frac;
      }
      
      return (back*a + fwd*b);
    } else {
      // cos(theta) is close to -1; the vectors are nearly opposite.
      // We use an alternative interpolation method.
      
      quaternion temp_q(a.w, a.v.z, -a.v.y, a.v.x);
      
      fwd = sin(frac * PI);
      back = sin((0.5 - frac) * PI);
      
      return (back*a + fwd*temp_q);
    }
  }
};

