// TagDatabase.cpp : Routines for class TagDatabase.
//

#include "stdafx.h"

/*
 * TagHashStructure : manages a hash table for a class of tags.
 */

// Store a tag.
void TagHashStructure::store_tag(tag_id store_id, void *tagdata, MythPlugin *owner)
{
  unsigned char bucket_index = hashfunction(store_id);
  taghashentry *tagentry = bucket[bucket_index], *lastentry = NULL, *newentry;
  long our_type, other_type;

  // Look along the chain for a place to insert this tag.
  while (tagentry != NULL) {
    if (tagentry->our_id == store_id) {
      // This tag entry indexes the same tag as we do. There are three possibilities:
      our_type = owner->header.plugin_type;
      other_type = tagentry->plugin->header.plugin_type;

      //   1) we are made irrelevant by this tag - if we have a lower precedence
      //      [higher type value] than the existing tag. plugins can't make anything
      //      irrelevant, as they may be conditionally not installed... if there is a
      //      plugin here that makes us irrelevant, just keep on 
      if (our_type > other_type) {
        if (other_type == PLUGINTYPE_PLUGIN) {
          // This tag makes us irrelevant, but we might have to overwrite the
          // underlying tags below it (as it is a plugin and might not be activated).
          // Keep looking along the chain.
          lastentry = tagentry;
          tagentry = tagentry->next;
          continue;
        } else {
          // We are irrelevant compared to this already-stored tag.
          return;
        }
      }

      //   2) we overwrite this tag  - if we have a higher or equal precedence
      //      [lower type value] than the existing tag. only overwrite if we are not a plugin
      //      because plugins' activation status may change.
      //
      //      it is important that we overwrite if we have
      //      equal precedence, because multiple patches will be ordered in alphabetical
      //      order (there may be another flag which controls patch precedence but I couldn't
      //      find it).
      if (our_type != PLUGINTYPE_PLUGIN) {
        // Overwrite this tag.
        tagentry->our_id = store_id;
        tagentry->plugin = owner;
        tagentry->tagdata = tagdata;
        return;
      }

      //   3) we are PLUGINTYPE_PLUGIN and we have a higher or equal precedence to the existing
      //      tag. insert ourselves into the chain here; we may or may not be used as a plugin
      //      depending on our activation status.
      break;
    }

    // Keep looking along the chain.
    lastentry = tagentry;
    tagentry = tagentry->next;
  }

  if (tagentry == NULL) {
    // There are no other tags with this ID in the hash database. Increment num_tags.
    num_tags++;
  }

  // We have reached the end of the chain, or a point where we should insert ourselves
  // ahead of other tags.
  newentry = (taghashentry *) malloc(sizeof(taghashentry));
  if (!newentry) mem_fail(sizeof(taghashentry),
                          "TagHashStructure entry");
  newentry->our_id = store_id;
  newentry->plugin = owner;
  newentry->tagdata = tagdata;
  newentry->last_iterate_id = 0;
  newentry->next = tagentry;

  // Link up the list to us.
  if (lastentry == NULL)
    bucket[bucket_index] = newentry;
  else
    lastentry->next = newentry;
}

// Retrieve a tag.
void *TagHashStructure::retrieve_tag(tag_id wanted_id, int ignore_plugin_activation)
{
  unsigned char bucket_index = hashfunction(wanted_id);
  taghashentry *tagentry;

  for (tagentry = bucket[bucket_index]; tagentry != NULL; tagentry = tagentry->next) {
    // DEBUGGING... if null_tag_id is passed in, just return the first tag of any kind.
    if ((wanted_id != null_tag_id) && (tagentry->our_id != wanted_id))
      continue;

    // Ignore tags from plugins that aren't currently activated... EXCEPT
    // if we are currently looking for entrypoint and so have the ignore flag set.
    if ((tagentry->plugin->header.plugin_type == PLUGINTYPE_PLUGIN) &&
        (!tagentry->plugin->plugin_activated) && (!ignore_plugin_activation))
      continue;

    return tagentry->tagdata;
  }

  // We could not find a tag that matched the ID supplied and was activated!
  return NULL;
}

// Iterate all tags in the structure, retrieving the highest-precedence one.
// Just used for entrypoints usually.
int TagHashStructure::iterate_tags(iteratestruct **iterator_ptr)
{
  iteratestruct *iterator = *iterator_ptr;
  taghashentry *t;

  if (iterator == NULL) {
    iterator = (iteratestruct *) malloc(sizeof(iteratestruct));
    if (!iterator) mem_fail(sizeof(iteratestruct),
                            "TagHashStructure iterator");
    *iterator_ptr = iterator;

    iterator->bucket_index = -1;
    iterator->iterate_id = next_iterate_id++;
    iterator->tagentry = NULL;
  }

  while (1) {
    // Find the next tag that hasn't been iterated.
    if (iterator->tagentry == NULL) {
      // Go to the next bucket.
      if (iterator->bucket_index == 255)
        return FALSE;

      iterator->tagentry = bucket[++iterator->bucket_index];
    } else {
      // Go to the next tag.
      iterator->tagentry = iterator->tagentry->next;
    }

    if (iterator->tagentry == NULL)
      continue;
    
    // If we've found a tag, we want to stop here.
    if (iterator->tagentry->last_iterate_id != iterator->iterate_id) {
      // We must color this tag and any further occurrences of this tag
      // in the has bucket as having already been returned.
      for (t = iterator->tagentry; t != NULL; t = t->next)
        if (t->our_id == iterator->tagentry->our_id)
          t->last_iterate_id = iterator->iterate_id;
      
      assert(iterator->tagentry != NULL);
      return TRUE;
    }
  }
}

// The hash function. NB: this is a BRAIN-DEAD hash function but will
// do for now.
unsigned char TagHashStructure::hashfunction(tag_id hash_id)
{
  unsigned char returnval = 0;

  returnval = hash_id & 0xFF;
  hash_id >>= 7;
  returnval ^= (hash_id & 0xFF);
  hash_id >>= 7;
  returnval ^= (hash_id & 0xFF);
  hash_id >>= 7;
  returnval ^= (hash_id & 0xFF);
  hash_id >>= 7;
  returnval ^= (hash_id & 0xFF);

  return returnval;
}

// Create a new hash structure for a given kind of tag.
TagHashStructure::TagHashStructure(tag_id tagtype)
{
  this->tagtype = tagtype;

  next_iterate_id = 1;
  num_tags = 0;

  memset(bucket, 0, sizeof(bucket));
}

TagHashStructure::~TagHashStructure(void)
{
  int i;
  taghashentry *t;

  for (i = 0; i < 256; i++)
    for (t = bucket[i]; t != NULL; t = t->next)
      free(t);

  free(bucket);
}

/*
 * TagDatabase : manages all the different classes of tags, and entry points.
 */

// Create an empty tag database.
TagDatabase::TagDatabase(void)
{
  num_plugins = pluginarraysize = 0;
  pluginarray = NULL;

  num_hashes = hasharraysize = 0;
  hasharray = NULL;

  entrypoint_hash = new TagHashStructure(null_tag_id);   // entrypoints aren't actually a type of tag...
}

TagDatabase::~TagDatabase(void)
{
  for (int i = 0; i < num_plugins; i++)
    delete pluginarray[i];

  if (pluginarray != NULL)
    free(pluginarray);
}

// Read all the plugins in a given directory.
void TagDatabase::read_plugins(char *directory)
{
  void *directorystructure = NULL;
  char filenamebuf[256];
  int is_directory;
  MythPlugin *new_plugin;

  while (iterate_directory(&directorystructure, directory, filenamebuf, &is_directory))
    if (!is_directory)
      if ((new_plugin = read_plugin(filenamebuf)) != NULL)
        register_plugin(new_plugin);
}

// Register a plugin.
void TagDatabase::register_plugin(MythPlugin *plugin)
{
  int i;
  Tag *tag;
  TagHashStructure *hash;

  if (num_plugins >= pluginarraysize) {
    pluginarraysize += pluginarray_growsize;
    pluginarray = (MythPlugin **) realloc((void *) pluginarray, pluginarraysize * sizeof(MythPlugin *));
    if (!pluginarray) mem_fail(pluginarraysize * sizeof(MythPlugin *),
                               "TagDatabase plugin array");
  }

  pluginarray[num_plugins++] = plugin;

  for (i = 0; i < plugin->header.num_tags; i++) {
    tag = plugin->tags[i];
    hash = hash_structure(tag->header->tag_type);

    if (hash == NULL) {
      if (num_hashes >= hasharraysize) {
        hasharraysize += hasharray_growsize;
        hasharray = (TagHashStructure **) realloc((void *) hasharray,
                  hasharraysize * sizeof(TagHashStructure *));
        if (!hasharray) mem_fail(hasharraysize * sizeof(TagHashStructure *),
                                 "TagDatabase hash-structure array");
      }
      hash = new TagHashStructure(tag->header->tag_type);
      hasharray[num_hashes++] = hash;
    }

    hash->store_tag(tag->header->our_id, (void *) tag, plugin);
  }

  // Store the entrypoints in the special entrypoint hash.
  for (i = 0; i < plugin->header.num_entrypoints; i++)
    entrypoint_hash->store_tag(plugin->entrypoint_array[i].mesh, (void *) &plugin->entrypoint_array[i], plugin);
}

// Get the hash structure for a particular kind of tag.
TagHashStructure *TagDatabase::hash_structure(long tagtype)
{
  int i;

  for (i = 0; i < num_hashes; i++)
    if (hasharray[i]->tagtype == tagtype)
      return hasharray[i];

  return NULL;
}

